/*
*****************************************************************************
* 
*   FILE NAME       :   INF.H
*   PROGRAMMER(S)   :   Robert Arnold
*   DATE CREATED    :   June 30, 1993
*
*   LAST EDITED : Tue Mar 29 16:30:23 1994 by arnold (Robert Arnold) on grep
*
*   DESCRIPTION:
*
*      This file contains all the constants with respect to the inf file
*      produced by the VPO compiler.
*
*   ASSUMPTIONS:
*      1. Tree.c (function Create_Tnode) assumes that 0 is an invalid
*         block number. <rda 02/25/94>
*
*      2. View each function as a loop that iterates 1 time.  Therefore,
*         each function should have at least 1 loop structure.
*         The function loop should have:
*                loop number = 0
*                nesting level = 0
*                minimum iterations = 1
*                maximum iterations = 1
*                list of all the blocks in the function
*                list of all the exit blocks out of the function
******************************************************************************
*
*   MODIFICATION HISTORY:
*
*   Date       Change				
*   --------   ----------------------------------------------------------------
*   02/10/94   Added "define INF.H" for conditional compilation
*
******************************************************************************
*/

#ifndef __INF_H__
#define __INF_H__

#define  FUNC_START_BLOCK       1       /* Number assigned to the 1st block
                                           of every func. */
#define  MIN_LOOP               0       /* Minimum possible value assigned
                                           to a loop by VPO. */
#define  LIST_TERMINATOR       -1        /* terminate list of related numbers */

#define  File_EXT              ".inf"    /* filename extension for inf files */
#define  FUNC_IDENTIFIER       -3        /* Number that preceeds a function */
#define  Func_IDENTIFIER       "-3"      /* String representation for strcmp */
#define  STRUCT_DEF            "!"       /* beginning of a struct definition */
#define  LOOP                  "loop"    /* start of a loop structure */
#define  BLOCK                 "block"   /* start of a block structure */
#define  DOMS                  "doms"    /* start of a list of dom blocks */
#define  PREDS                 "preds"   /* start of a list of pred blocks */
#define  SUCCS                 "succs"   /* start of a list of succ blocks */
#define  REG_LIST_START        '('       /* start of list of regs for an inst */
#define  REG_LIST_TERMINATOR   ')'       /* end of list of regs for an inst */

#define ASK_USER             -2
#define COMPUTE_ITERATIONS   -3
#define INDUCTION_VAR_KNOWN  -4
#define MULTIPLE_EXIT        -5

#define REL_EQ     1        /* Equal */
#define REL_NE     2        /* Not equal */
#define REL_GT     3        /* Greater-than */
#define REL_LT     4        /* Less-than */
#define REL_GE     5        /* Greater-than-or-equal */
#define REL_LE     6        /* Less-than-or-equal */

#endif // __INF_H__
