/*
*****************************************************************************
*
* 
*   FILE NAME       :   TIME.H
*   PROGRAMMER(S)   :   Robert Arnold
*   DATE CREATED    :   July 3, 1993
*
******************************************************************************
*
*   MODIFICATION HISTORY:
*
*   Date       Change				
*   --------   ----------------------------------------------------------------
*   07/03/93   Created timing_list structure
*   03/05/94   Added time_case enumeration
*   03/05/94   Added cycle_times struct
*   03/06/94   Added cache_access enumeration
******************************************************************************
*/

#ifndef __TIME_H_INCL__
#define __TIME_H_INCL__

#ifndef INST_H
#include "inst.h"
#endif

#define MAX_STRING_LENGTH 1024
//#define DC_MISS_PENALTY 0
#define MAXSETSIZE 8
#define MAXNUMSETS 8192

#define YES 1
#define NO 0
#define INSERT_FUNC 2
#define FUNC_WAIT 1
#define NO_FUNC 0

#define MAXINT   0x7fffffff    /* Maximum integer value on the machine */
#define UNUSED -1
typedef  int  row[NUM_STAGES + 2];
typedef  struct instruct * instruct_row[NUM_STAGES];
typedef unsigned int *dvector;    /* basic block list type */
#define dinit() ((dvector) NULL)
#define DALLOC (unsigned int *) salloc(dveclen << int2char)
#define dfree(ptr) if (ptr) free((char *) ptr)

struct delay_after_fm_node
{
  int inst;
  int cycles;
  struct delay_after_fm_node *next;
};

struct cycle_jump_node
{
  int from;
  int to;
  struct cycle_jump_node *next;
};

struct union_list_node
{
  struct instruct *occupant;
  int cycles;
  struct union_list_node *next;
};

struct inst_num_node
{
  int num;
  struct inst_num_node *next;
  struct inst_num_node *prev;
};

typedef  char   string[MAX_STRING_LENGTH];


enum time_case {WORST = 0, BEST = 1};
enum cache_access {MISS = 0, HIT = 1};
enum loop_path_type {NONE=0, BACKEDGE, EXIT};
   

struct block_list
{
  int  block_number;            /* block number assigned by VPO */
  struct block_list  *next;     /* pointer to next block in list */
};



                             /* STRUCT LOOP_BLOCK_LIST                       */
                             /*  This struct is used to represent each block */
                             /* in a loop path.  If the node represents a    */
                             /* loop, then the boolean flag loop_node will   */
                             /* be true and the block number will contain    */
                             /* the block number for the header of the loop. */
struct loop_block_list
{
  int  block_number;             /* block number in the loop path */
  bool  is_loop_node;    /* indicates block represents a nested loop */

  struct loop_node  *loop_node_ptr;      /* ptr to the loop_node representing
                                            that loop                        */

  long wc_time;

  struct loop_block_list    *next;
};

struct range_node
{
  int number;			/* identifies which range this is */
  int min;
  int max;
  int max_iter;
  int path_mask;
  struct range_node *next;
  struct range_node *prev;
};

struct path_num_node		/* list of path numbers for next iteration */
{
  int number;
  struct path_num_node *next;
  struct path_num_node *prev;
};




                      /* STRUCT LOOP_PATH_NODE                         */
                      /* EXIT_LIST: Is a list of all the exit blocks   */
                      /* this path can exit to.  Note that for loops,  */
                      /* the exit block is a block outside of the loop,*/
                      /* so that exit block will not be contained in  */
                      /* the path_list.  For loop_path_nodes           */
                      /* representing functions, due to limitations in */
                      /* the algorithm, the exit block IS contained in */
                      /* the path list.                                */
                      /* NOTE:                                         */
                      /*   The INF file lists exit blocks for a loop   */
                      /*   as blocks that the loop can exit out of.    */
                      /*   These being blocks that are WITHIN the loop.*/
                      /*   The blocks contained in the exit_list are   */
                      /*   the blocks OUTSIDE the loop, i.e. the blocks*/
                      /*   that the loop exits to.                     */

struct loop_path_node 
{
  int                     number;
  struct path_num_node    *cant_next;        /* paths that can't go next */
  int min;				      /* min & max iter numbers */
  int max;
  int			  min_iter;	      /* expected # iter */
  int                     max_iter;
  int                     unique_min;
  int                     unique_max;
  int                     take_away;	      /* iters within a range */
  bool            already_took_away;
  int                     max_take_away;
  int                     required_iters_left;
  int                     nonrequired_iters_left;
  int                     range_index;
  enum cache_access       prev_fm_status;    /* see if we need recalc */
  enum cache_access       prev_fh_status;
  struct cache_block      *prev_fm_list;
  struct cache_block      *prev_fh_list;
  int                     header_block;      /* header block to this path */
  int                     exit_block;        /* last block in this path */
  struct loop_block_list  *path_list;        /* List of all blocks in path
                                                             thru the loop */
  struct block_list       *exit_list;        /* List of exit blocks 
                                                             from this path */
  struct cache_block      *best_fm_list;     /* List 1st miss blocks in cache
                                                             for best case */
  struct cache_block      *worst_fm_list;    /* List 1st miss blocks in cache 
                                                             for worst case */
  struct block_list       *worst_fh_list;    /* List first hits 
                                                     for worst case timing */
  struct block_list       *best_fh_list;     /* List first hits 
                                                     for best case timing*/
  int path_type_mask;
#ifdef INTERFACE
  int  min_base_fh;
  int  max_cycle_fmiss;   /* Max number cycles where each first miss = miss */
#endif
  int  extra_time;	
  int 	extra_fh_time;
  int  extra_fm_time;
  int  shared_fm_time;		   /* extra time from shared fm's 2/9/99*/
  int  bc_extra_time;

  struct instruct     *beg_occupant[NUM_STAGES];
  struct instruct     *end_occupant[NUM_STAGES];
  int  beg_cycles[NUM_STAGES];
  int  end_cycles[NUM_STAGES];
  int  path_cycles;
  int  inner_ff_trans;
  int  fm_encountered;
  int  fh_encountered;
  int  release_time[NUMREGS];
  int  src_needed[NUMREGS];
  int  src_stage[NUMREGS];
  int  beg_cache_line;
  int  first_word_num;
  int	end_cache_line;
  int	*beg_avail_time;
  int  *end_avail_time;
  int	*fm_end_avail_time;
  int	first_line_delay;
  int  last_line_delay;
  int	delay_after_fm;
  int  first_prog_offset;
  int  *beg_word_num;
  int  last_fm;
  int	fm_dead_cycles;
  char last_miss_cat;		/* either a 'm' or a 'f' */
  int	finish_load;		/* how many cycles after IF done 'til loaded */
  struct delay_after_fm_node *delay_after_fm_head;
  struct loop_path_node   *next;

  bool  already_timed_fmiss;  /* just time path once */
  int		occ_cycles[NUM_STAGES];
  int		bc_pc_nai_time;
  int		wc_pc_nai_time;
  int           bc_pipe_nai_time;
  int		wc_pipe_nai_time;

  /*BOOKMARK- adding variables required for using frequency model */

  /*The frequency model follows the equation C = M*L + I 
   I is same as path_cycles, need to keep track of the M (misses)*/
  int icache_misses;
  int dcache_misses;  
  int extra_latency;
  int latency_considered;/*this will contain the latency after considering 
			   all iterations of the path */
  int total_latency;/*this is sum of i, d extra_latency and extra_time */
  /*some statistics, incase I have to take care of overlapping instr later*/

  int diff;
  int overlap;
  /*extra_time and family members are not added... ASSUMED that if 
    calculating parameterized timing, they will hold some multiple of 
    icache latency */



};

                        /* STRUCT LOOP_NODE                               */  
                        /* Assumption is made that loop has one unique    */
                        /* header block but can have multiple exit blocks.*/
                        /* Exit blocks saved at each path.                */
                        /* LIST_TERMINATOR is a define specified in the   */
                        /* "inf.h" file and used to indicate that a value */
                        /* is unknown.                                    */
struct loop_node                                        
{                                                       
  int    number;              /* Loop number assigned by vpo compiler */
  int    header_block;              
  struct loop_path_node  *path;  /* list of all possible paths thru the loop */
  int    nest_level;          /* nesting level of the loop */
  int    min_iterations;      /* min iter of loop--LIST_TERMINATOR = unknown */
  int    max_iterations;      /* max iter of loop--LIST_TERMINATOR = unknown */
  int    abs_min_iterations;	/* not the average in cases like sort */
  int    abs_max_iterations;	  /* used for reporting local times */
  struct loop_node          *next;
};

struct block_element
{
  int            block_number;        /* block number assigned by VPO */
  struct instruct   *instruct_list;   /* list of instructions in block */
  struct block_list  *pred;       /* list of predicate blocks */
  struct block_list  *succ;       /* list of successor blocks */
  struct block_list  *infeasible_succ;
  struct block_list  *unknown;
  struct block_list  *branch;
  struct block_list  *fallthru;
  struct block_list  *unfellthru;  /* unknown if not fell thru last time */
  struct block_list  *unbranched;  /* unknown if not branched last time */
  struct block_list  *ft_cause_b;
  struct block_list  *ft_cause_ft;
  struct block_list  *b_cause_b;
  struct block_list  *b_cause_ft;
  dvector in_effect;		    /* I have information about the branch */
  dvector in_unknown;		    /* The state of the branch is unknown */
  dvector in_jump;		    /* is it jump state or fall thru state */
  dvector out_effect;
  dvector out_unknown;
  dvector out_jump;
  int left_min;
  int left_max;
  int right_min;
  int right_max;
  char dir;
  int dir_iter;
  struct block_list  *dom;        /* list of blocks that dominate this block */
  struct block_element  *next;    /* pointer to next block in list */
  /* Added by Sibin M on Mar 31, 2005.
   * For Dr. Healy's Parametric TA.
   */
  int    src_start;             /* starting source line number */
  int    src_end;               /* ending source line number */
  /* Added by Sibin M on Mar 31, 2005.
   * For Dr. Healy's Parametric TA.
   */
};

struct timing_label_element
{
  int number;			/* identifies which @ l.l., as in @1 or @2... */
  int earliest_line;		/* first line containing this @ */
  int lower_bound;		/* user's bc expected time for list of inst */
  int upper_bound;		/* user's wc expected time */
  struct line_list_element *line_list;
  struct timing_label_element *next;
};

struct line_list_element
{
  int lineno;
  struct line_list_element *next;
};

struct loop_element
{
  int    number;                /* loop number assigned by vpo compiler */   
  int    nest_level;            /* loop nesting level as determined by VPO */
  int    min_iterations;        /* min iterations of loop */
  int    max_iterations;        /* max iterations of loop */
  int    abs_min_iterations;
  int    abs_max_iterations;
  int    total_iters;
  int    induction_reg;
  int    induction_init;
  int    induction_limit;
  int    induction_incr;
  bool early_exit;
  bool constant_limit;
  bool need_to_average;
  int    begin_value;
  int    begin_offset;
  int    limit_value;
  int    limit_offset;
  int    increment;
  int    outer_number;
  int    lower_bound;           /* user's expected bc time (10/1) */
  int    upper_bound;           /* user's expected wc time (10/1) */
  int    header;                /* block number that is the loop header */
  int    src_start;             /* starting source line number */
  int    src_end;               /* ending source line number */
  struct block_list   *block_list;   /* list of blocks contained in loop */
  struct block_list   *exit_list;    /* list of blocks that can exit loop */
  struct loop_element *next;         /* pointer to next loop */

  /* Added by Sibin M on Mar 31, 2005.
   * For Dr. Healy's Parametric TA.
   */
  int 	 need_polynomial; /* set this to 1 if we are using polynomial analysis for this loop. */
  char   *loop_variable; /* this is not null whenever the loop requires polynomial analysis and
                            will hold the variable representing the number of times to perform the loop */
  struct polynomial *polynomial_repn; /* this holds a list of the polynomial strucutres for the time of this loop */

  /* Added by Sibin M on Mar 31, 2005.
   * For Dr. Healy's Parametric TA.
   */
};


struct path_element
{
  int                      block_number;
  bool             is_loop_node;
  struct loop_node         *loop_node_ptr;
  struct path_element      *succ;
  struct path_element      *next;
};

/*********************
*
*   NAME:   INF_FUNC_ELEMENT
*
*                            * these 2 variables will be used during         *
*                            * calculation of all possible paths across      *
*                            * function calls.  If a block flows into a      *
*                            * second block outside the function, comparison *
*                            * on this range will determine if the desired   *
*                            * block falls within this particular function.  *
*       int     min_block_number;
*       int     max_block_number;
*************************/
struct inf_func_element
{
  string     func;                /* function name                           */
  int        number;              /* number assigned to func when read
                                                               from inf file  */
  string     inf_filename;        /* inf filename function read from         */
  int        min_block_number;    /* minimum block # contained within func   */
  int        max_block_number;    /* maximum block # contained within func   */

  struct timing_label_element *timing_label;  /* l.l. of timing labels */

  struct loop_element       *loop;        /* ptr to list of loops
                                                             inside function */
  struct block_element      *block;       /* ptr to list of blocks
                                                             inside function */
  struct path_element       *path;        /* ptr all poss paths of 
                                                      control flow thru func */
  struct loop_node          *loop_nodes;  /* ptr to nodes representing
                                                        paths thru each loop */
  struct inf_func_element   *next;        /* ptr to next func in inf file    */
};


/* New structure that will store the subpath information. The subpath indicated *
   in this structure will be timed. 1/29/03 
*****************/
   


struct instr_history {
  int icache_miss;
  int dcache_miss;
  int is_mem_op;
  int is_overlap;
  int overlap_cycles;
  int is_stalled;  //by RAW
  int negate;
};


/*D$ hit time*/
#define DC_HIT_LATENCY   1
#define IC_HIT_LATENCY   1
#define  IC_MISS_LATENCY  1
#define MIN_LATENCY_RANGE 10
#define MAX_LATENCY_RANGE 100

#endif // __TIME_H_INCL__
