/*
 *****************************************************************************
 * 
 *   FILE NAME       :   TREE.H
 *   PROGRAMMER(S)   :   Robert Arnold
 *   DATE CREATED    :   February 10, 1994
 *
 ******************************************************************************
 *
 *   MODIFICATION HISTORY:
 *
 *   Date       Change				
 *   --------   ----------------------------------------------------------------
 *   02/10/94   Created Tree node
 *   02/25/94   Added tnode_type
 ******************************************************************************
 */

#ifndef __TREE_H_INCL__
#define __TREE_H_INCL__

#include "instdef.h"
#include "timeds.h"

#define  Func_MAIN       "_main"   /* string used in search of main func
				      in stats file */

#define  Loop_PREFIX     "LOOP"    /* string used in creating names for nodes
				      representing loops in timing tree */

enum tnode_type {FNODE=0, LNODE=1};

#define UNKNOWN 0
#define BRANCH 1
#define FALLTHRU 2


struct cache_block
{
	int     line_number;        /* cache line number  */
	int     inst_number;        /* program inst number */
	int     instance;
	bool  trans_occurred;
	char    best_case_cat;      /* last category used for this cache line # */
	char    worst_case_cat;

	char    prev_bc_cat;
	char    prev_wc_cat;		/* previous worst_case_cat, at lower level */

	struct category_list  *cat_list;  /* list of inst categories for that block*/
	struct cache_block    *next;
};


struct best_case_timing
{
	struct cache_block    *fm_list;
	struct cache_block    *fh_list;

	int    beg_cycles[NUM_STAGES];
	struct instruct  *beg_occupant[NUM_STAGES];
	int    end_cycles[NUM_STAGES];
	struct instruct  *end_occupant[NUM_STAGES];

	int    pipeline_time;
	int    extra_time;
	int	  extra_fh_time;
	int	  extra_fm_time;
	int 	  num_paths;
	struct union_list_node  *beg_head[NUM_STAGES];
	struct union_list_node  *end_head[NUM_STAGES];
	int  src_needed[NUMREGS];  
	int  src_stage[NUMREGS];
	int  release_time[NUMREGS];
	int  occ_cycles[NUM_STAGES];	   	/* for compressibility (6/13) */
	int  pc_nai_time;			/* bc naive time for pipeline & cache */
	int  pipe_nai_time;
	int	beg_cache_line;
	int  end_cache_line;
	int	*beg_avail_time;
	int	*end_avail_time;
	int	first_line_delay;
	int	delay_after_fm;
	int	last_line_delay;
	int	first_prog_offset;
	int  *beg_word_num;
	int	*fm_end_avail_time;
	int  last_fm;
	int	fm_dead_cycles;
	int	finish_load;
	struct delay_after_fm_node *delay_after_fm_head;

	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */

	int need_polynomial ;
	struct polynomial* polynomial ;

	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */
};


struct worst_case_timing
{
	struct cache_block  *fm_list;     /* fm list for base path */
	struct cache_block  *fh_list;     /* list of first hits */

	int  pipeline_time;
	int  abs_time;
	int  extra_time;		/* for f->f transitions (4/23) */
	int  extra_fh_time;
	int  extra_fm_time;
	int  shared_fm_time;
	int  num_paths;
	struct union_list_node  *beg_head[NUM_STAGES];
	struct union_list_node  *end_head[NUM_STAGES];
	int  src_needed[NUMREGS];
	int  src_stage[NUMREGS];
	int  release_time[NUMREGS];
	int  pc_nai_time;
	int  pipe_nai_time;
	int  beg_cache_line;
	int  end_cache_line;
	int  *beg_avail_time;
	int  *end_avail_time;
	int  first_line_delay;
	int  delay_after_fm;
	int  last_line_delay;
	int  first_prog_offset;
	int  *beg_word_num;
	int  *fm_end_avail_time;
	int  last_fm;
	int  fm_dead_cycles;
	int  finish_load;
	struct delay_after_fm_node *delay_after_fm_head;
	/*BOOKMARK - Adding variables to add the frequency model */
	int icache_misses;
	int dcache_misses;
	int extra_latency;
	int latency_considered;
	int total_latency;/*this is sum of i, d extra_latency and extra_time */
	/*some statistics, incase I have to take care of overlapping instr later*/
	int abs_total_latency;
	int overlap;
	int diff;

	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */

	int need_polynomial ;
	struct polynomial* polynomial ;

	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */
};


struct exit_block
{
	int                        block_number;
	struct best_case_timing    *best_case;
	struct worst_case_timing   *worst_case;
	struct exit_block          *next;
};


struct category_list
{
	char                   best_case;
	char                   worst_case;
	struct category_list   *next;
};


struct inst_cat
{
	int     inst_number;     /* instruction number */
	int     line_number;     /* program line number used for first misses */
	int     loop_number;     /* loop number inst is located */
	struct category_list  *cat_list;
	string      func;        /* name of function being called by this inst */
	int         instance;    /* instance of called function */
	struct inst_cat       *next;  /* ptr to next instruction category */
};


struct branch_node
{
	int number;		/* the block containing this branch */
	int state;		/* state of branch:  UNKNOWN, BRANCH, FALLTHRU */
	struct branch_node *next;
	struct branch_node *prev;
};


struct tnode
{
	enum tnode_type   node_type;     /* Nodes in the tree represent either
					    functions or loops   */
	string   name;                   /* function or loop name              */
	int      instance;               /* instance of that function          */
	int      loop_number;            /* loop # to which node is a child to */
	int      loop_header_block;      /* if this is loop node,
					    this represents loop header block */
	bool      timed;         /* indicates if timing is
				    complete for this node */
	int      wc_pipeline_time;
	int      bc_pipeline_time;
	int      wc_abs_time;
	int      bc_abs_time;
	struct union_list_node *wc_beg_head[NUM_STAGES];
	struct union_list_node *wc_end_head[NUM_STAGES];
	struct union_list_node *bc_beg_head[NUM_STAGES];
	struct union_list_node *bc_end_head[NUM_STAGES];
	int      extra_time;
	int      extra_fh_time;
	int      extra_fm_time;
	int      shared_fm_time;
	int	   bc_extra_time;
	int	   bc_pc_nai_time;
	int	   wc_pc_nai_time;
	int      bc_pipe_nai_time;
	int	   wc_pipe_nai_time;
	int	   wc_release_time[NUMREGS];
	int      wc_src_needed[NUMREGS];
	int	   wc_src_stage[NUMREGS];
	int	   bc_release_time[NUMREGS];
	int	   bc_src_needed[NUMREGS];
	int	   bc_src_stage[NUMREGS];
	int	   occ_cycles[NUM_STAGES];
	int	   wc_beg_cache_line;
	int	   wc_end_cache_line;
	int	   bc_beg_cache_line;
	int	   bc_end_cache_line;
	int	   *wc_beg_avail_time;
	int	   *bc_beg_avail_time;
	int	   *wc_end_avail_time;
	int	   *bc_end_avail_time;
	int	   first_line_delay;
	int	   delay_after_fm;
	int	   last_line_delay;
	int	   wc_first_prog_offset;
	int	   *wc_beg_word_num;
	int	   *wc_fm_end_avail_time;
	int      last_fm;
	int	   fm_dead_cycles;
	int	   wc_finish_load;
	struct   delay_after_fm_node *wc_delay_after_fm_head;
	struct   path_set_node *path_set;	/* P0, P1, etc. */
	int      num_paths;
	int      num_blocks;
	int      **imm_follow;
	int      **path_reach;
	struct range_node *range_list;

	struct exit_block  *exit_list;   /* list exit blocks from node,
					    each exit block has timing values */
	struct inst_cat    *instruct_list;    /* ptr to list of insts for this node*/
	struct inf_func_element   *inf_ptr;   /* ptr to inf information          */     
	struct loop_node   *loop_paths;       /* ptr to all paths thru this loop */
	struct branch_node *branch;      /* ptr to branches in node */
	struct tnode  *parent;           /* ptr to the parent node               */
	struct tnode  *sibling;          /* ptr to the sibling nodes             */
	struct tnode  *child;            /* ptr to child nodes                   */
	struct tnode  *next;             /* ptr to next tnode                    */

	/*BOOKMARK - Adding variables to add the frequency model */
	int icache_misses;
	int dcache_misses;
	int extra_latency;
	int latency_considered;
	int total_latency;
	int abs_total_latency;
	int overlap;
	int diff;

	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */

	int need_polynomial;
	struct polynomial *polynomial_worst_case;
	
	/* Added by Sibin M on Mar 31, 2005. 
	 * For Parametric TA from Dr. Healy.
	 */
	long visa_worst_case;
};

#endif // __TREE_H_INCL__
