#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mtadel.h"
#include "debug.h"

extern char *replace(char *, char *, char *);
extern char *mtadel_itoa(int );
extern int search (char *, char *);
extern void mark_iterator_variable(char *);
extern int dependents(char *);

/* ============================
calculate(to_calculate)

This function will calculate the answer to _to_calculate
	given its position relative to the other summations
	within the linked list returned by parse_command. 
	In other words, calculate is aware of the nesting
	level of _to_calculate. For "niceness", calculate()
	is called for each summation in the aforementioned
	linked list. Therefore, this function does not 
	invoke recursion. 

See calculate.txt for the actual calculation algorithm. It is
	too messy to put here.
   ==========================*/
void calculate(struct summation *to_calculate)
{
	CALC_VERBOSE("we are starting calculate!\n");

	if (to_calculate->lo_is_sym == 0 && to_calculate->hi_is_sym == 0)
	/* this means that our bounds are constant and only means something when the previous is completely numeric*/
	{
		/* lo is constant:
			sum = (hi-lo+1)*prev_sum
			if there is no prev_sum, then just pretend its a 1!
		*/

		/* if the previous sum is symbolic, then please just mark the induction 
			variable as iterate, move the equation to the sum of the current
			summation structure, and then move on. */
		int iterations = (to_calculate->hi - to_calculate->lo + 1);
		if (to_calculate->next != NULL && to_calculate->next->sum_is_sym == 1)
		{
			/* the previous sum is symbolic */
			/* let's do an actual copy, just to make sure we don't screw anything up! */
			/* if the variable is already in the summation, then we mark it as an iterator, otherwise
				we can just pretend that it is a constant multiplier! */
			CALC_VERBOSE("to_calculate->next->sum_sym: %s\n", to_calculate->next->sum_sym);
			CALC_VERBOSE("to_calculate->induction: %s\n", to_calculate->induction);

			if (search(to_calculate->next->sum_sym, to_calculate->induction) 
				|| dependents(to_calculate->induction))
			{
				char *tmp_sum = (char*)malloc(strlen(to_calculate->next->sum_sym) + 1);
				CALC_VERBOSE("search() or dependents() succeeded!\n");
				memset(tmp_sum, 0, strlen(to_calculate->next->sum_sym) + 1);
			
				mark_iterator_variable(to_calculate->induction);
				strcpy(tmp_sum, to_calculate->next->sum_sym);
				to_calculate->sum_sym = tmp_sum;
				to_calculate->sum_is_sym = 1;
			}
			else
			{
				char *iterations_c = mtadel_itoa(iterations);

				int iterations_length = strlen(iterations_c);
				int previous_length = strlen(to_calculate->next->sum_sym);

				int tot_len = previous_length + iterations_length + 4;

				char *sum = (char*)malloc(tot_len);
				CALC_VERBOSE("we are here! \n");
				memset(sum, 0, tot_len);

				strcpy(sum, to_calculate->next->sum_sym);
				strcat(sum, "*(");
				strcat(sum, iterations_c);
				strcat(sum, ")");

				to_calculate->sum_sym = sum;
				to_calculate->sum_is_sym = 1;
			}
		}
		else
		{
			/* either the previous sum is NULL (ie 1) or it is a constant! */
			int prev_sum = (to_calculate->next) ? to_calculate->next->sum : 1;
			to_calculate->sum = prev_sum * iterations;
			to_calculate->sum_is_sym = 0;
		}
	}
	else
	{	
		if (to_calculate->next && to_calculate->next->sum_is_sym)
		{
			int length = strlen(to_calculate->induction) + strlen(to_calculate->next->sum_sym);
			char *replacement = (char*)malloc(length);
			char *replaced = NULL;
			memset(replacement, 0, length);

			strcpy(replacement, to_calculate->induction);
			if (to_calculate->lo_is_sym)
				strcat(replacement, to_calculate->lo_sym);
			else
				strcat(replacement, mtadel_itoa(to_calculate->lo));

			replaced = replace(to_calculate->next->sum_sym, to_calculate->induction, replacement);

			if (!strcmp(replaced, to_calculate->next->sum_sym))
			{
				mark_iterator_variable(replacement);
			}

			to_calculate->sum_sym = replaced;
			to_calculate->sum_is_sym = 1;
		}
		if ((to_calculate->next && to_calculate->next->sum_is_sym == 0) || !to_calculate->next)
		{
			char *previous = NULL;
			int previous_length = 0;
			char *hi = NULL;
			char *lo = NULL;
			char *lo_shifted = NULL;
			char *hi_shifted = NULL;

			int hi_length, lo_length, tot_len;
			char *sum = NULL;

			if (to_calculate->next)
				previous = mtadel_itoa(to_calculate->next->sum);
			else
				previous = mtadel_itoa(1);
			
			/* there must be 2 different cases here: if the to_calculate->lo_is_sym == 1
								if the to_calculate->lo_is_sym == 0. */

			previous_length = strlen(previous);

			if (to_calculate->hi_is_sym == 1)
				hi = to_calculate->hi_sym;
			else
				hi = mtadel_itoa(to_calculate->hi);
			
			if (to_calculate->lo_is_sym == 1)
				lo = to_calculate->lo_sym;
			else
				lo = mtadel_itoa(to_calculate->lo);

			if (to_calculate->is_lo_shifted)
			{
				/* do something about a lower shift that is present! */
				int len = strlen(to_calculate->lo_shift) + strlen(lo) + 4;
				lo_shifted = (char*)malloc(len);
				memset(lo_shifted, 0, len);

				strcpy(lo_shifted, "(");
				strcat(lo_shifted, lo);
				strcat(lo_shifted, to_calculate->lo_shift);
				strcat(lo_shifted, ")");
			}

			if (to_calculate->is_hi_shifted)
			{
				/* do something about an upper shift that is present! */
				int len = strlen(to_calculate->hi_shift) + strlen(hi) + 4;
				hi_shifted = (char*)malloc(len);
				memset(hi_shifted, 0, len);

				strcpy(hi_shifted, "(");
				strcat(hi_shifted, hi);
				strcat(hi_shifted, to_calculate->hi_shift);
				strcat(hi_shifted, ")");
			}

			hi_length = strlen(hi);
			lo_length = strlen(lo);

			tot_len = previous_length + ((hi_shifted!=NULL) ? strlen(hi_shifted) : hi_length) + 
					((lo_shifted!=NULL) ? strlen(lo_shifted) : lo_length) + 6;
			sum = (char*)malloc(tot_len);
			memset(sum, 0, tot_len);

			strcpy(sum, previous);
			strcat(sum, "*(");

			if (hi_shifted != NULL)
				strcat(sum, hi_shifted);
			else
				strcat(sum, hi);

			strcat(sum, "-");
			if (lo_shifted != NULL)
				strcat(sum, lo_shifted);
			else
				strcat(sum, lo);
			strcat(sum, "+1");
			strcat(sum, ")");

			to_calculate->sum_sym = sum;
			to_calculate->sum_is_sym = 1;
			CALC_VERBOSE("to_calculate->sum_sym: %s\n", to_calculate->sum_sym);
		}
	}
	return;
}
