#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../debug.h"

#define PLUS 257
#define MINUS 258
#define MULTIPLICATION 259
#define INTEGER 260
#define OPEN_PAREN 261
#define CLOSE_PAREN 262



char *yytext;
extern int yylval;

extern char *expression;
int where_was_i = 0;

/* ============================
yylex()

This function tokenizes the string pointed to by the global
	variable _expression. All spaces are ignored and the
	possible tokens are PLUS, MINUS, MULTIPLICATION, DIVISION,
	INTEGER, OPEN_PAREN, and CLOSE_PAREN. 

	Algorithm: 
	1. Index the string _expression based on the where_was_i variable.
	2. If the character in this position is part of an integer
		a. add the character to _current_token.
		b. set _found_part_of_int to 1.
		c. increment where_was_i.
		d. if the next character in _expression is the end of string,
			then return INTEGER.
		e. Start again at 2.
	   else
	   	a. check to see if _found_part_of_int == 1
			if so, then return INTEGER.
		   else
		   	return based upon the type of operator
			encountered and increment _where_was_i
			by 1.
   ==========================*/
int yylex()
{
	int i = where_was_i, found_part_of_int = 0;
	int bound = strlen(expression);
	char *current_token = NULL;
	for (; i < bound; i++)
	{
		if (expression[i] == ' ')
		{
			where_was_i++;
			continue;
		}

		if ( ((int)(expression[i] - '0')) < 10 && ((int)(expression[i] - '0')) >= 0)
		{
			/* we have an integer */
			found_part_of_int = 1;
			if (current_token == NULL)
			{
				current_token = (char*)malloc(2);
				memset(current_token, 0, 2);
				current_token[0] = expression[i];
			}
			else
			{
				int len = strlen(current_token);
				current_token = realloc(current_token, len + 1);
				current_token[len] = expression[i];
				current_token[len + 1] = '\0';
			}
			where_was_i++;

			/* this handles the case where we have an integer immediately preceeding the end of the string! */
			if (expression[i + 1] == '\0')
			{
				/* we found a + and it will delimit and integer! */
				yytext = current_token;
				yylval = atoi(yytext);
				LEX_VERBOSE("yytext: -%s-\n", yytext);
				return INTEGER;
			}
				
		}
		else
		{
			/* we do not have an integer! */
			if (	(expression[i] == '+' || 
				expression[i] == '*' || 
				expression[i] == '-' || 
				expression[i] == ')' || 
				expression[i] == '(' || 
				expression[i+1] == '\0')
				&& found_part_of_int == 1
			)
			{
				/* we found a + and it will delimit and integer! */
				yytext = current_token;
				yylval = atoi(current_token);
				LEX_VERBOSE("yytext: -%s-\n", yytext);
				return INTEGER;
			}
			where_was_i++;
			yytext = (char*)malloc(2);
			memset(yytext, 0, 2);
			yytext[0] = expression[i];
			if (expression[i] == '+')
				return PLUS;
			if (expression[i] == '-')
				return MINUS;
			if (expression[i] == '*')
				return MULTIPLICATION;
			if (expression[i] == '(')
				return OPEN_PAREN;
			if (expression[i] == ')')
				return CLOSE_PAREN;

			//return (expression[i] == '+') ? PLUS : MULTIPLICATION;
		}
	}
	return 0;
}

/*
// uncomment this when you are running make lexer_test
int main()
{
	LEX_VERBOSE("expression: %s\n", expression);
	while (yylex())
	{
		LEX_VERBOSE("-%s-\n", yytext);
	}
	return 0;
}
*/
