#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "mtadel.h"

extern int is_numeric( char *);
extern void mark_iterator_variable(char*);

/* ============================
part_of_variables(n)

if _c is a mathematical operator 
	(including parenthesis),
	then return 1.
	Else, return 0.
   ==========================*/
int part_of_variable(char c)
{
	if (	c == '-' ||
		c == ')' ||	
		c == '(' ||	
		c == '*' ||	
		c == ' ' ||	
		c == '+' ||	
		c == '-' )	
		return 0;

	return 1;
}

/* ============================
search(subject, pattern)

Return true (1) if _pattern is found within
	_subject. Not only does it have to be
	found, it must be found in a position
	where it is delimited by a space or
	mathematical operator on either side.

	Algorithm: Standard search algorithm. 
	Look for the beginning of a match. Loop
	through the pattern and the subject 
	simultaneously. If the loop iterates
	entirely through the pattern without
	finding a difference between it and 
	the _subject, then we check the characters
	around the match within the subject.
	If the characters on either side of the
	match are either a space or a mathe-
	matical operator, then the match is real.
	Otherwise, reset variables and 
	continue to look for the beginning
	of a match.
   ==========================*/
int search(char *subject, char *pattern)
{
	int subject_length = strlen(subject);
	int pattern_length = strlen(pattern);
	int i = 0;

	for (; i < subject_length; i++)
	{
		if (subject[i] == pattern[0])
		{
			/* start of the match! */
			int inner_i = i;
			int j = 0;
			int not_found = 0;
			for (; inner_i < subject_length && j < pattern_length; inner_i++, j++)
			{
				if (subject[inner_i] != pattern[j])
				{
					not_found = 1;
					break;
				}
			}
			if (not_found == 0 && j == pattern_length)
			{
				/* we found a match! */
				if   (! (
					( subject[inner_i] == ' ') ||
					( subject[inner_i] == '*') ||
					( subject[inner_i] == '\0') ||
					( subject[inner_i] == '(') ||
					( subject[inner_i] == ')') ||
					( subject[inner_i] == '-') ||
					( subject[inner_i] == '+' )
					) )
				{
					MARK_VERBOSE("continue'ing from the first if statement within an if inside search()\n"); 
					MARK_VERBOSE("\t subject: %s pattern: %s subject[inner_i]: %c\n", subject, pattern, subject[inner_i]);
					continue;
				}
			
				if      ( ((i-1) > 0) && !(
					( subject[i-1] == ' ') ||
					( subject[i-1] == '*') ||
					( subject[i-1] == '-') ||
					( subject[i-1] == '(') ||
					( subject[i-1] == ')') ||
					( subject[i-1] == '+' )
					) )
				{
					MARK_VERBOSE("continue'ing from the second if statement within an if inside search()\n"); 
					MARK_VERBOSE("\t subject: %s pattern: %s\n", subject, pattern);
					continue;
				}
				return 1;
			}
		}
	}
	return 0;
}

/* ============================
find_variables(to_search)

Look through _to_search and find all
	the variables. For each of the
	variables, mark them for
	iteration within the symbol table.
	This is useful in calculate().
   ==========================*/
int find_variables(char *to_search)
{
	int i = 0, len = 0, start, end;
	if (to_search == NULL) return 0;
	/* if _to_search is entirely numeric, then we should not
		even waste our time looking for variables. */
	if (is_numeric(to_search))
	{
		return 2;
	}

	i=0; len = strlen(to_search);
	for (; i<len; i++)
	{
		char *symbol;
		if (part_of_variable(to_search[i]))
		{
			start = i;
			while (part_of_variable(to_search[i]) && i<len)
			{
				i++;
			}
			end = i;
			/* set i back 1, because hitting the end of the for loop will cause it to increase by 1
				and we want to start exactly at the i we have here. */
			i--;

			/* the characters between to_search[start] to to_search[end] is a variable. */
			symbol = (char*)malloc(end-start + 1);
			memset(symbol, 0, end-start+1);
			strncpy(symbol, to_search + start, end-start);
			FIND_VERBOSE("symbol: %s\n", symbol);
			mark_iterator_variable(symbol);
			free(symbol);
		}
	}
	return 1;
}
