//=================================================================
//Replace the Prolog summation interpreter with this code.
//
//Author: Will Hawkins
//Date: March 30, 2004.
//=================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "mtadel.h"

#define _GNU_SOURCE // to use getline()

extern char*replace(char*, char*, char*);
extern int mtadel_eval(char *);
extern char *mtadel_itoa(int);
extern int where_was_i;
extern void calculate(struct summation *);
extern int find_variables(char *);
extern void evaluate(struct summation *);
extern void update_upper(char *, char *);
extern int iterations;

int r_answer;

/* symbol table stuff */
extern void init_symbol_table(void);
extern void add_symbol(char *, int, int, int);
extern struct symbol *get_symbol(char *);
extern void print_symbol_table(void);
extern void normalize(void);

/* ============================
is_numeric("string" test)

returns true(1) or false(0) based on whether or
	not the _test parameter is an equation
	without variables. 
   ==========================*/
int is_numeric(char *test)
{
	/* iterate and make sure the entire equation is numerical! */
	int length;
	int i = 0;
	length = strlen(test);
	for (; i<length; i++)
	{
		if (	test[i] == '-' ||
			test[i] == ')' ||	
			test[i] == '(' ||	
			test[i] == '*' ||	
			test[i] == ' ' ||	
			test[i] == '+' ||	
			test[i] == '-' )	
		{
			continue;
		}

		M_VERBOSE("MARK test[i] - '0' = -%d-\n", test[i]-'0');
		if ( ((test[i]-'0')<10 ) &&  ((test[i]-'0') >= 0))
			continue;
		else
			break;
	}
	if (i == (length))
		return 1;
	else
		return 0;
}

/* ============================
is_digit()

returns true(1) or false(0) based on whether or 
	not parameter _c is a digit. 
   ==========================*/
int is_number(char *test)
{
	int length;
	int i = 0;
	length = strlen(test);
	for (; i<length; i++)
	{
		if (test[i] == '-' && i == 0)
			continue;

		M_VERBOSE("MARK test[i] - '0' = -%d-\n", test[i]-'0');
		if ( ((test[i]-'0')<10 ) &&  ((test[i]-'0') >= 0))
			continue;
		else
		{
			/* break out of the loop if the character is not a number. */
			break;
		}
	}
	/* if we looked at each of the characters, and they are all numbers (or a negative sign)
		then we can say the _test was a number */
	if (i == (length))
		return 1;
	else
		return 0;
}

/* ============================
parse_command()

This function parses a command string like
sum(1, induction=bottom .. top [by induction2=bottom2 
.. top2 [by ...]]). Upon successfully parsing
a command, the function will return the pointer to a 
summation structure. Within this structure is
a pointer to the "next" summation structure. Therefore,
the return value of summation is actually the first
structure in a linked list representing all the 
summations to be calculated.
   ==========================*/

struct summation * parse_command(char *command)
{
	unsigned again = 0;
	int iteration = 0;
	char *lower, *upper, *induction;
	struct summation *prev_sum;
	struct summation *iterator;
	struct summation *head;

	prev_sum = NULL;

	upper = NULL; lower = NULL;

	command += 6;

	M_VERBOSE("command:%s\n", command);

	/* parsing algorithm:
		assumptions: iteration is an integer representing the current position within the string
			to be parsed.
		1. loop through the string until a "=" sign is read. For each of these non "=" characters
			that are processed, add them to the _induction string.
		2. skip past the = sign.
		3. loop through the string until a "." is read. For each of these non "." characters
			that are processed, add them to the _lower string.
			a. Of course, if we encounter any mathematical operators before this "." sentinel
				character, then we have lower bound that is mathematically shifted.
			b. Add the detected mathematical operator and all characters after that (until a "." 
				is read) to the _lower_shift string.
		4. now that a "." has been read, we expect another ".". Read the next character to decide if 
			the expectation is true. If the next character read is not a ".", print an error and 
			return NULL.
		5. Repeat step 3 to read the upper bound of the loop. Of course, this step will
			manipulate the _upper and _upper_shift variables instead of the _lower and
			_lower_shift variables.
		6. Put _induction, _lower, _lower_shift, _upper and _upper_shift into the correct places
			in a summation structure. Then, add this new structure representing the most recently
			parsed portion of the command to the end of the linked list. 
		7. Set the again variable to 1 if there is another summation to read. We can decide if another summation
			is to be read by processing the characters following 0 or more spaces immediately after
			the upper command in the command string.
		8. Add the current summation's induction variable to the symbol table. 
		9. If the lower or upper bound are variable, then a hybrid symbol must be created and added to the
			symbol table. 
		10. If again == 1, then start again at step 1. Otherwise, finish.
	*/
	do 
	{
		char *lower_shift, *upper_shift;
		struct summation *to_add = NULL;
		again = 0;

		induction = (char*)malloc(1);
		induction[0] = '\0';
		while (command[iteration] != '=')
		{
			int induction_length = 0;
			if (command[iteration] == ' ')
			{
				iteration++;
				continue;
			}

			induction_length = strlen(induction);
			M_VERBOSE("strlen(induction): %d\n", induction_length);

			M_VERBOSE("before realloc on induction\n");
			induction = realloc(induction, strlen(induction)+2);
			M_VERBOSE("after realloc on induction\n");

			induction[induction_length] = command[iteration];
			induction[induction_length + 1] = '\0';

			M_VERBOSE("new strlen(induction): %d\n", strlen(induction));

			iteration ++;
		}

		/*we are already pointing to the "=" sign, so, let's move one more past it.*/
		iteration++;

		lower = (char*)malloc(1);
		lower[0] = '\0';
		while (command[iteration] != '.' &&
			command[iteration] != '+' &&
			command[iteration] != '-' &&
			command[iteration] != '*' &&
			command[iteration] != '/'
			)
		{
			if (command[iteration] == ' ')
			{
				iteration++;
				continue;
			}
			M_VERBOSE("before realloc on lower\n");
			lower = realloc(lower, strlen(lower) + 2);
			M_VERBOSE("after realloc on lower\n");
			strncat(lower, &command[iteration], 1);
			iteration++;
		}

		/* use this to lick up a linear shift, if necessary */
		lower_shift = (char*)malloc(1);
		lower_shift[0] = '\0';

		if (command[iteration] != '.')
		{
			while (command[iteration] != '.')
			{
				if (command[iteration] == ' ')
				{
					iteration++;
					continue;
				}
				lower_shift = realloc(lower_shift, strlen(lower_shift) + 2);
				strncat(lower_shift, &command[iteration], 1);
				iteration++;
			}
		}
		M_VERBOSE("lower_shift: %s\n", lower_shift);

		if (command[++iteration] != '.')
		{
			ERROR("error parsing command string: .. expected!\n");
			break;
		}

		iteration++;
		while (command[iteration] == ' ')
			iteration++;
		
		upper = (char*)malloc(1);
		upper[0] = '\0';
		/* this conditional expression is complicated because the upper bound 
			could be terminated by the end of a command (')'), a space or another
			mathematical operator. */
		while (command[iteration] != ' ' && command[iteration] != ')'
			&& command[iteration] != '*'
			&& command[iteration] != '+'
			&& command[iteration] != '-'
			&& command[iteration] != '/'
			)
		{
			int upper_length = strlen(upper);
			M_VERBOSE("before realloc on upper\n");
			upper = realloc(upper, upper_length + 2);
			M_VERBOSE("after realloc on upper\n");

			upper[upper_length] = command[iteration];
			upper[upper_length+1] = '\0';

			iteration++;
		}

		upper_shift = (char*)malloc(1);
		upper_shift[0] = '\0';
		if (command[iteration] != ' ' && command[iteration] != ')')
		{
			while (command[iteration] != ' ' && command[iteration] != ')')
			{
				upper_shift = realloc(upper_shift, strlen(lower_shift) + 2);
				strncat(upper_shift, &command[iteration], 1);
				iteration++;
			}
		}
		M_VERBOSE("upper_shift: %s\n", upper_shift);

		M_VERBOSE("lower:-%s-\n", lower);
		M_VERBOSE("upper:-%s-\n", upper);
		M_VERBOSE("induction:-%s-\n", induction);

		/* eat up possible spaces */
		while (command[iteration] == ' ')
			iteration++;

		/* check to see if the length of _command will permit additional indexes safely. If so, check
			to see if the next 2 characters are 'b' and 'y'. If 'b' and 'y' are read, there is 
			at least 1 more summation to read. So, set the loop variable to continue. */
		if (strlen(command) > iteration && command[iteration] == 'b' && command[iteration+1] == 'y')
		{
			M_VERBOSE("Execute the do ... while() loop to read another sumation.\n");
			again = 1;
			iteration += 2;
		}
		else
			M_VERBOSE("Breaking the do ... while() loop.\n");

		/* create a new summation structure, and put the proper variables in the
			proper places. This is done according to the algorithm described 
			before this function. */
		to_add = (struct summation *)malloc(sizeof(struct summation));
		memset(to_add, 0, sizeof(struct summation));
		to_add->induction = induction;

		if (is_number(lower))
		{
			to_add->lo = atoi(lower);
			to_add->lo_is_sym = 0;
		}
		else
		{
			to_add->lo_sym = lower;
			to_add->lo_is_sym = 1;
		}

		if (is_number(upper))
		{
			to_add->hi = atoi(upper);
			to_add->hi_is_sym = 0;
		}
		else
		{
			to_add->hi_sym = upper;
			to_add->hi_is_sym = 1;
		}

		if (prev_sum == NULL)
		{
			M_VERBOSE("prev_sum is NULL\n");
			/* first node in the list */
			to_add->next = NULL;
			prev_sum = to_add;
		}
		else
		{
			/* another node is already present in the list */
			M_VERBOSE("prev_sum is not NULL\n");
			to_add->next = prev_sum;
			prev_sum  = to_add;
		}

		if (strlen(upper_shift)>1)
		{
			to_add->hi_shift = upper_shift;
			to_add->is_hi_shifted = 1;
		}
		if (strlen(lower_shift)>1)
		{
			to_add->lo_shift = lower_shift;
			to_add->is_lo_shifted = 1;
		}

		/* Now, take the induction variable and add it to the symbol table. */
		if (to_add->lo_is_sym == 0 && to_add->hi_is_sym == 0)
		{
			add_symbol(induction, to_add->lo, to_add->hi, 0);

			if (strlen(upper_shift)>1)
				update_hi_shift(induction, upper_shift);
			if (strlen(lower_shift)>1)
				update_lo_shift(induction, lower_shift);
		}
		else if (to_add->hi_is_sym == 0)
		{
			add_symbol(induction, 0, to_add->hi, 0);

			if (strlen(upper_shift)>1)
				update_hi_shift(induction, upper_shift);
			if (strlen(lower_shift)>1)
				update_lo_shift(induction, lower_shift);
		}


		/* if either the lower bound or the upper bound are variable, 
			then a hybrid symbol is created to be added to the table. */
		if (to_add->lo_is_sym == 1 || to_add->hi_is_sym == 1)
		{
			/* add a hybrid symbol to the table. */

			char *lo, *hi, *hybrid;
			struct symbol *tmp = NULL;

			lo = NULL;

			if (to_add->lo_is_sym == 0)
				lo = mtadel_itoa(to_add->lo);
			else
				lo = to_add->lo_sym;
			
			hybrid = (char *)malloc(strlen(lo) + strlen(induction) - 1);
			memset(hybrid, 0, strlen(lo) + strlen(induction) - 1);
			strcpy(hybrid, induction);
			strcat(hybrid, lo);

			tmp = get_symbol(lo);
			if (tmp)
			{
				/* we can calculate the lows and highs right now! */
				int b_lo = tmp->lo;
				int b_hi = tmp->hi;
				add_symbol(hybrid, b_lo, (b_hi <= to_add->hi) ? b_hi : to_add->hi, 1);
			}
			else
			{
				add_symbol(hybrid, 0, 0, 1);
			}
			update_halves(hybrid, induction, lo);

			if (to_add->hi_is_sym)
				update_upper(hybrid, to_add->hi_sym);

			if (strlen(upper_shift)>1)
				update_hi_shift(hybrid, upper_shift);
			if (strlen(lower_shift)>1)
				update_lo_shift(hybrid, lower_shift);
				
		}
	} while (again);
	head = prev_sum;

	return head;
}
/* ============================
print_summation(iterator)
	
	Print all the fields of a summation structure.
   ==========================*/
void print_summation(struct summation *iterator)
{
	printf("=====summation output=====\n");
	printf("lo: %d\n", iterator->lo);
	printf("hi: %d\n", iterator->hi);
	printf("lo_sym: %s\n", iterator->lo_sym);
	printf("hi_sym: %s\n", iterator->hi_sym);
	printf("lo_is_sym: %d\n", iterator->lo_is_sym);
	printf("hi_is_sym: %d\n", iterator->hi_is_sym);
	printf("sum_is_sym: %d\n", iterator->sum_is_sym);
	printf("induction: %s\n", iterator->induction);
	printf("sum: %d\n", iterator->sum);
	printf("sum_sym: %s\n", iterator->sum_sym);
	printf("lo_shift: %s\n", iterator->lo_shift);
	printf("hi_shift: %s\n", iterator->hi_shift);
	printf("==========\n");
}
/* ============================
solve_summation(ptr)
	
	In order to solve a summation, we need to look 
	through the linked list of summations and make
	it a doubly linked list rather than a singly 
	linked list.
	Then, call calculate() to actually do the work.
   ==========================*/
void solve_summation(struct summation *ptr)
{
	struct summation *iterator = ptr;
	struct summation *tmp = ptr;

	if (ptr == NULL)
		M_VERBOSE("ptr is NULL\n");

	for (; iterator; iterator = iterator->next)
	{
		M_VERBOSE("inside solve_summation loop.\n");
		if (tmp == iterator)
		{
			iterator->previous = NULL;
		}
		else
		{
			iterator->previous = tmp;
		}
		tmp = iterator;
	}

	M_VERBOSE("we now have a doubly linked list!\n");
	for (iterator = tmp; iterator; iterator = iterator->previous)
	{
		calculate(iterator);
	}
	return;
}


/* ============================
main()

This is an interesting thing. Since we
	need most of what is in here for 
	actually calculating the answer
	to a summation, conditional compilation
	will be used. If "TEST" is defined
	(see Makefile), then the code for
	reading a summation command from a 
	file is compiled in and a main()
	function is generated. In so doing,
	the generated program can be run
	for testing whether our summation
	solver works correctly. Otherwise,
	a function named mtadel(char*) is
	compiled. This is the function that
	is called by SPARTA to invoke the
	summation solver. The summation
	to be calculated is passed as the 
	parameter into mtadel().
   ==========================*/
#ifdef TEST

int mtadel_getline(char **where_to_store, int *length, FILE *fp)
{
	/* since this is a hacked version, forget about the length variable 
		and assume that it is a 0. Ie. try to read an entire line 
		from the file. 
	*/
	char input;
	*where_to_store = (char*)malloc(1);
	(*where_to_store)[0] = '\0';

	while (fread(&input, sizeof(char), 1, fp))
	{
		int old_length = strlen(*where_to_store);
		if (input == '\n')
			break;
		*where_to_store = realloc(*where_to_store, old_length + 2);
		(*where_to_store)[old_length] = input;
		(*where_to_store)[old_length+1] = '\0';
	}

	if (feof(fp) != 0)
		return -1;
	else
		return 1;
}

int main()
{
	FILE *fp = fopen("./regression", "r");
	char *line = NULL;
	int len = 0;
#endif
	struct summation *iterator;
	struct summation *head;
#if TEST
	if (fp == NULL)
	{
		M_VERBOSE("could not open ./regression.\n");
		return -1;
	}

	if (-1 == mtadel_getline(&line, &len, fp))
	{
		M_VERBOSE("getline() failed!\n");
		return -1;
	}
	head = parse_command(line);
	M_VERBOSE("line: %s\n", line);
#else
int mtadel(char *line)
{
	head = parse_command(line);
	
#endif
	if (line == NULL) return 0;

/* if we have MTADEL debugging enabled, print out each of the
	summation structures in the linked list generated
	by parse_command(). */
#ifdef M_DEBUG	
	for (iterator = head; iterator; iterator = iterator->next)
		print_summation(iterator);
#endif

/* if we have symbol table debugging enabled, print
	the symbol table. */
#ifdef SYMBOL_DEBUG
	print_symbol_table();
#endif
	/* normalize the symbol table. See comments in symbol.c for
		the purpose of this function. */
	normalize();
	M_VERBOSE("second print_symbol_table:\n");

/* if we have symbol table debugging enabled, 
	print out the symbol table again
	to see the effects of the normalize()
	function call. */
#ifdef SYMBOL_DEBUG
	print_symbol_table();
#endif

	iterator = head;

	M_VERBOSE("Now we are calculating the summations :\n");

	solve_summation(head);
	
	M_VERBOSE("calling find_variables()\n");
	find_variables(head->sum_sym);
	M_VERBOSE("done with find_variables()\n");

/* if we have symbol table debugging enabled
	print the symbol table again. We do this
	to visualize the changes calculate() made 
	to the summation structures wrt the
	iterate field. */
#ifdef SYMBOL_DEBUG
		M_VERBOSE("look for iterators.\n");
		print_symbol_table();
		M_VERBOSE("stop looking for iterators.\n");
#endif

/* if we have MTADEL debugging enabled,
	print out the summations to visualize
	the effects of calculate(). */
#ifdef M_DEBUG
		for (iterator = head; iterator; iterator = iterator->next)
			print_summation(iterator);
#endif

	/* since r_answer is a global variable, we should
		reinitialize before it is to be used. */
	r_answer = 0;
	/* this will actually find a numeric answer */
	evaluate(head);

#ifdef TEST
	printf("r_answer: %d\n", r_answer);
	printf("iterations: %d\n", iterations);
	return 0;
#else
	/* if we are generating the mtadel() function, we
		want to actually return the answer rather
		than just print it out. */
	printf("MTADEL was used.\n");
	return r_answer;
#endif
}
