/* NOTE: nesting the pattern within in the replace string is simply not tolerated! */

#include "../debug.h"

#include <stdio.h>
#include <string.h>

extern char *mtadel_itoa(int );

/* ============================
replace(subject, pattern, replace)

This function will replace all portions of _subject that 
	match _pattern with _replace. This "new" string is
	then 'return'ed. 

	Algorithm: The matching algorithm is the same
	one used in the search() function. The source (and
	algorithm documentation) can be found in ../mark.c.

	1. Declare a variable to hold the string to be returned. 
		Denote that nothing has been added to this string
		by setting its value to NULL.
	2. Find the position of the next occurence of _pattern within
		_subject (alpha). 
	3. If the string to be returned is NULL, then copy the
		string starting at subject[0] - subject[alpha]
		to the string to be returned. 
	   else
	   	copy subject[position of last replacement] - subject[alpha]
		to the string to be returned.
	4. Next, add _replace to the string to be returned. 
	5. Update the position of last replacement variable.
	6. Repeat steps 2 - 5 until there are no more occurences of
		_pattern in _subject.
	7. If the last occurence of _pattern in _subject did not 
		appear at exactly the end of the string, there is
		some "tail" of _subject that must be copied to
		the string to be returned. Copy subject[position 
		of last replacement] - subject[strlen(subject) - 1] 
		to the string to be returned.
   ==========================*/
char *replace(char *subject, char *pattern, char *replace)
{
	int subject_length, pattern_length, replace_length, i, j, last_replace_location;
	char *new_string = NULL;
	if (pattern == NULL)
	{
		ERROR("pattern is NULL. Cannot continue.\n");
		return NULL;
	}

	SR_VERBOSE("subject: %s\n", subject);
	SR_VERBOSE("pattern: %s\n", pattern);
	SR_VERBOSE("replace: %s\n", replace);

	subject_length = strlen(subject);
	pattern_length = strlen(pattern);
	replace_length = strlen(replace);
	i = 0; j = 0; last_replace_location = 0;

	for (; i<subject_length;i++)
	{
		/* look for the start of a match*/
		SR_VERBOSE("subject[i] = %c\n", subject[i]);
		SR_VERBOSE("pattern[0] = %c\n", pattern[0]);

		if (subject[i] == pattern[0])
		{
			int inner_i = i, found_match = 1;
			for (j = 0; 
				j<pattern_length && inner_i<subject_length;
				j++, inner_i++)
			{
				if (pattern[j] != subject[inner_i])
				{
					found_match = 0;
					break;
				}
			}
			/* if we find a match, advance the iterator to past the
			matched portion of the string */
			/* we are here because we matched and i is the start of the matched string! */

			SR_VERBOSE("j = %d \t inner_i-i = %d\n", 
				j, inner_i-i);
		
			/*j must ==  pattern_length or else we know the for loop did not scan the 
				entire contents of the pattern and therefore we ended with only a partial match! */

			/* NOTE: pattern_length is actually 1 more than our iterator, but since we use the strictly
				less than to limit our value of j, we should escape the loop when j == pattern_length */
			/* NOTE (also): j may not = pattern_length in the case that a possible pattern match starts 
				near the end of the string and we are in danger of reading past the end of the string
				before checking to make sure the string matches exactly! */
			if (found_match == 0 || j != pattern_length)
				continue;

			/* NOTE: we can ALWAYS index the string subject at inner_i because of loop constraints above! */
			if   (! (
				( subject[inner_i] == ' ') ||
				( subject[inner_i] == '*') ||
				( subject[inner_i] == '\0') ||
				( subject[inner_i] == '(') ||
				( subject[inner_i] == ')') ||
				( subject[inner_i] == '-') ||
				( subject[inner_i] == '+' )
				) )
				continue;
			
			if      ( ((i-1) > 0) && !(
				( subject[i-1] == ' ') ||
				( subject[i-1] == '*') ||
				( subject[i-1] == '-') ||
				( subject[i-1] == '(') ||
				( subject[i-1] == ')') ||
				( subject[i-1] == '+' )
				) )
				continue;

			if (new_string == NULL)
			{
				/* malloc space for the i-1 portion of the string! */
				new_string = (char*)malloc(i+1);
				memset(new_string, 0, i+1);
				strncpy(new_string, subject, i);
				SR_VERBOSE("new_string: -%s- (within new_string == NULL inside replace()\n", new_string);
			}
			else
			{
				new_string = (char*)realloc(new_string, i - last_replace_location + 1);
				strncat(new_string, subject + last_replace_location, i - last_replace_location);
			}
			SR_VERBOSE("strlen(new_string): %d\n", strlen(new_string));
			SR_VERBOSE("replace_length: %d\n", replace_length);
			new_string = (char*)realloc(new_string, strlen(new_string) + replace_length + 1);
			strcat(new_string, replace);
			SR_VERBOSE("strlen(new_string): %d (after strcat)\n", strlen(new_string));

			last_replace_location = i + pattern_length;
		}
	}
	if (last_replace_location < subject_length)
	{
		int new_string_old_length;
		if (new_string == NULL)
		{
			new_string = (char*)malloc(strlen(subject) + 1);
			memset(new_string, 0, strlen(subject) + 1);
			strcpy(new_string, subject);
			SR_VERBOSE("new_string: %s\n", new_string);
			return new_string;
		}

		new_string_old_length = strlen(new_string);

		new_string = (char*)realloc(new_string, new_string_old_length + i - last_replace_location + 1);
		new_string[new_string_old_length] = '\0';

		strncat(new_string, subject + last_replace_location, i - last_replace_location);
	}
		
	return new_string;
}
/*
//test driver
int main()
{
	char *ns = replace("10 * 1*(i-0+1)", "i", mtadel_itoa(9));
	//char *ns = replace("i+", "i", "1");
	ERROR("ns: -%s-\n", ns);
	return 0;
}
*/
