#include "tracer.h"

#include <string>
#include <iostream>
#include <sstream>
#include <vector>

int Tracer::_indent_level(0) ;
int Tracer::_num_spaces(1) ;
std::ostream* Tracer::_out(0) ;

Tracer::Tracer( const std::string& to_print, bool new_scope )
		: _new_scope( new_scope ) 
{
	if( _out )
	{
		if( _new_scope )
			++_indent_level ;	

		*_out << "--------------------------" << std::endl
		      << std::string( ( _indent_level - 1 ) * _num_spaces, ' ' ) 
		      << "Start::" << to_print << std::endl ;		

		_printed << to_print ;
	}
}

Tracer::Tracer( const char* to_print, bool new_scope )
		: _new_scope( new_scope )
{
	if( _out )
	{
		if( _new_scope )
			++_indent_level ;	

		*_out << "--------------------------" << std::endl
		      << std::string( ( _indent_level - 1 ) * _num_spaces, ' ' ) 
		      << "Start::" << to_print << std::endl ;		

		_printed << to_print ;
	}
}

Tracer::~Tracer()
{
	if( _out )
	{
		*_out << std::string( ( _indent_level - 1 ) * _num_spaces, ' ' )
		      <<  "Stop::" << _printed.str() << std::endl 
		      <<  "--------------------------" << std::endl ;

		if( _new_scope )
			--_indent_level ;
	}
}

Tracer& Tracer::operator << ( int to_print ) 
{
	if( _out )
		*_out << std::string( _indent_level * _num_spaces, ' ' ) 
		      << to_print ;

	return *this ;
}

Tracer& Tracer::operator << ( const std::string& to_print )
{
	if( _out )
		*_out << std::string( _indent_level * _num_spaces, ' ' ) 
		      << to_print ;

	return *this ;
}

Tracer& Tracer::operator << ( const char* to_print )
{
	if( _out )
		*_out << std::string( _indent_level * _num_spaces, ' ' ) 
		      << to_print ;

	return *this ;
}

Tracer& Tracer::operator << ( double to_print )
{
	if( _out )
		*_out << std::string( _indent_level * _num_spaces, ' ' ) 
		      << to_print ;

	return *this ;
}

void Tracer::stream( std::ostream* out ) 
{
	_out = out ;
}

int Tracer::numSpaces( int num_spaces )
{
	int return_val = _num_spaces ;
	_num_spaces = num_spaces ;

	return _num_spaces ;
}
