#include "timeds.h"
#include "tree.h"
#include "instdef.h"

/* Added by Sibin M on Oct 25, 2004. 
 * Process of cleaning up the TA for C++
 */

#include "depend.h"
#include "lib.h"
#include "isa.h"
#include "timedecl.h"

/* Added by Sibin M on Oct 25, 2004. 
 * Process of cleaning up the TA for C++
 */

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#ifdef DISPLAY
#define printc printf
#else
#define printc  
#endif

extern int dveclen, int2char, intscale, bitsperint;

void host_info()
{
  unsigned int ui;
  int rt, index;

  ui = bitsperint = 1;
  while (ui <<= 1)
    bitsperint++;
  for (intscale = 0; bitsperint > 1; bitsperint >>= 1)
    intscale++;
  bitsperint = (bitsperint << intscale) - 1;
  index = rt = sizeof(unsigned int);
  for (int2char = 0; index > 1; index >>= 1)
    int2char++;
  if (rt > (((unsigned int) 1) << int2char))
    int2char++;
}

/*
 * set_vector_length - To determine the vector length for a function,
 * we use the same idea as in renumblks() in controlflow.c  1/31/99
 * Also, allocate space for the 4 vectors per block in the function.
 */
void set_vector_length( struct inf_func_element* inf )
{
  struct inf_func_element *func;
  struct block_element *block;
  int num_blocks = 0;
  extern bool verbose;

  for (func = inf; func; func = func->next)	/* for each function */
  {
    for (block = func->block; block; block = block->next)
      ++num_blocks;

    dveclen = num_blocks >> intscale;
    if (num_blocks & bitsperint)
      ++dveclen;				/* find dveclen for this func */

    if (verbose)
      printc ("for this function, dveclen = %d\n", dveclen);

    for (block = func->block; block; block = block->next)
    {
      /*
       * Allocate space for vectors and initialize them to null.
       */
      block->in_effect   = (dvector) DALLOC;
      block->in_unknown  = (dvector) DALLOC;
      block->in_jump     = (dvector) DALLOC;
      block->out_effect  = (dvector) DALLOC;
      block->out_unknown = (dvector) DALLOC;
      block->out_jump    = (dvector) DALLOC;

      block->in_effect = dinit();
      block->in_unknown = dinit();
      block->in_jump = dinit();
      block->out_effect = dinit();
      block->out_unknown = dinit();
      block->out_jump = dinit();
    }
  }
}

/*
 * The routines salloc(), dinsert(), din(), dinter() ... were copied
 * from VPO.  showdvect() is from ~whalley/new/vpo/lib/misc.c.
 *
 * salloc - allocate space and check for error
 */
char *salloc( int siz )
{
  char *ptr;

  if (!(ptr = ( char* )malloc((unsigned) siz))) {
    fprintf(stderr, "salloc(): out of memory\n");
    exit(1);
    }
  return(ptr);
}

/*
 * dinsert - adds a block reference to a dvector list
 */
void dinsert( dvector* sptr, unsigned int r )
{
  register unsigned int *ptr, *end;

  if (!*sptr) {
    *sptr = (dvector) (ptr = DALLOC);
    end = ptr + dveclen;
    while (ptr != end)
      *ptr++ = 0;
    }
  (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
}

/*
 * din - indicate if a basic block is in a dvector
 */
int din( dvector ptr, unsigned int r )
{
  return(ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}

/*
 * dinter - calculate the intersection of two dvector lists
 */
void dinter( dvector* res, dvector ptr )
{
  register unsigned int *ptr1, *ptr2, *end;

  if (*res && ptr) {
    end = (ptr1 = (unsigned int *) *res) + dveclen;
    ptr2 = (unsigned int *) ptr;
    while (ptr1 != end)
      *ptr1++ &= *ptr2++;
    }
  else if (!ptr)
    *res = ptr;
}

/*
 * dunion - calculates union of two dvectors
 */
void dunion( dvector* dresult, dvector dvec )
{
  register unsigned int *ptr1, *ptr2;
  register int end;

  if (ptr2 = (unsigned int *) dvec) {
    if (!*dresult) {
      *dresult = (dvector) (ptr1 = DALLOC);
      for (end = dveclen; end--; *ptr1++ = *ptr2++);
      }
    else {
      ptr1 = (unsigned int *) *dresult;
      for (end = dveclen; end--; *ptr1++ |= *ptr2++);
      }
    }
}

/*
 * dminus - calculates {a} - {b}
 */
void dminus( dvector* a, dvector b )
{
  register unsigned int *ptr1, *ptr2, *end;

  if ((ptr1 = (unsigned int *) *a) && (ptr2 = (unsigned int *) b))
    for (end = ptr1 + dveclen; ptr1 != end; *ptr1++ &= ~(*ptr2++));
}

/*
 * dcpy - copy a dvector list to another one
 */
void dcpy( dvector* dst, dvector src )
{
  register unsigned int *ptr1, *ptr2, *end;

  if (src) {
    if (!*dst)
      *dst = (dvector) DALLOC;
    end = (ptr1 = (unsigned int *) *dst) + dveclen;
    ptr2 = (unsigned int *) src;
    while (ptr1 != end)
      *ptr1++ = *ptr2++;
    }
  else if (*dst) {
    dfree(*dst);
    *dst = dinit();
    }
}

/*
 * dcmp - compare two dvector lists
 */
int dcmp( dvector a, dvector b )
{
  register unsigned int *ptr1, *ptr2, *end;

  if (a && b) {
    end = (ptr1 = (unsigned int *) a) + dveclen;
    ptr2 = (unsigned int *) b;
    while (ptr1 != end)
      if (*ptr1++ != *ptr2++)
        return(false);
    return(true);
    }
  return(a == b);
}

/*
 * showdvect - show the blocks in a dvector
 */
void showdvect( dvector d )
{
   int i, j;
   extern int dveclen;

   for (i = 0; i < dveclen; i++)
     for (j = 0; j < 8*sizeof(unsigned int); j++)
       if (din(d, i*32+j)) 
	   printc(" %d", i*32+j);	 

   printc("\n");
}
 
/*
 * block_constraints - propagate constraint information between the blocks
 * based on calculating ins and outs.  We use this info later when looking
 * at constraints going into and coming out of an inner loop/child function.
 * started 1/30/99
 */
void block_constraints( struct inf_func_element* inf )
{
  struct inf_func_element *func;
  struct block_element *block, *pred_block;
  struct block_list *block_list_ptr;
  int change, pass = 0;
  dvector intersection_effect, intersection_unknown, intersection_jump,
          this_block_effect, this_block_unknown, this_block_jump,
          temp_vector, temp_vector2, temp_vector3,
          new_effect, new_unknown, new_jump;
  extern bool  verbose;

  extern bool verbose;

  intersection_effect = (dvector) DALLOC;
  intersection_unknown = (dvector) DALLOC;
  intersection_jump = (dvector) DALLOC;
  this_block_effect = (dvector) DALLOC;
  this_block_unknown = (dvector) DALLOC;
  this_block_jump = (dvector) DALLOC;
  temp_vector = (dvector) DALLOC;
  temp_vector2 = (dvector) DALLOC;
  temp_vector3 = (dvector) DALLOC;
  new_effect = (dvector) DALLOC;
  new_unknown = (dvector) DALLOC;
  new_jump = (dvector) DALLOC;

  for (func = inf; func; func = func->next)
  {
    /*
     * 1. Only propagate a constraint into the in's of a block
     *    if it was in the outs of every immediate predecessor block.
     * 2. Propagate into the outs of a block from the in's as long as another
     *    constraint was not associated with the same branch in the same block.
     */
    change = true;
    while (change)
    {
      ++pass;
      change = false;

      if (verbose) {
	printc ("====================================\n");
	printc ("pass #%d\n", pass);
      }

      for (block = func->block; block; block = block->next)       {
	if (verbose) {
        printc ("--------------\ncomputing ins and outs for block %d\n",
                block->block_number);
	}
        intersection_effect = dinit();
        intersection_unknown = dinit();
        intersection_jump = dinit();
        new_effect = dinit();
        new_unknown = dinit();
        new_jump = dinit();
        temp_vector = dinit();
        temp_vector2 = dinit();
        temp_vector3 = dinit();

        if (block->pred)		/* if at least one predecessor */
        {
          pred_block = Locate_Inf_Block(func, block->pred->block_number);
          dcpy (&intersection_effect, pred_block->out_effect);
          dcpy (&intersection_unknown, pred_block->out_unknown);
          dcpy (&intersection_jump, pred_block->out_jump);

          /*
           * p1e = predecessor 1 effect  p1u = unknown  p1j = jump
           * p2e = predecessor 2 effect  p2u = unknown  p2j = jump
           *
           *  p1e   p1u   p1j     p2e   p2u   p2j      int-e int-u int-j
           *   0     0     0       0     0     0         0     0     0
           *   0     0     0       1     1     0         1     1     0
           *   0     0     0       1     0     1         1     0     1
           *   0     0     0       1     0     0         1     0     0
           *
           *   1     1     0       0     0     0         1     1     0
           *   1     1     0       1     1     0         1     1     0
           *   1     1     0       1     0     1         1     1     0
           *   1     1     0       1     0     0         1     1     0
           *
           *   1     0     1       0     0     0         1     0     1
           *   1     0     1       1     1     0         1     1     0
           *   1     0     1       1     0     1         1     0     1
           *   1     0     1       1     0     0         1     1     0
           *
           *   1     0     0       0     0     0         1     0     0
           *   1     0     0       1     1     0         1     1     0
           *   1     0     0       1     0     1         1     1     0
           *   1     0     0       1     0     0         1     0     0
           *
           * int-effect = union of 1-effect and 2-effect
           * int-unknown = union of 1-unknown, 2-unknown, and the union of
           *                 two intersections:
           *                 1.  2-effect, 1-jump minus 2-unknown minus 2-jump
           *                 2.  1-effect, 2-jump minus 1-unknown minus 1-jump
           * int-jump = union of
           *              1.  2-jump minus 1-effect
           *              2.  1-jump minus 2-effect
           *              3.  1-jump intersect 2-jump
           *
           * In the code below, pred 1 is "intersection" and pred 2 is 
           * in pred_block's outs.
           */

          for (block_list_ptr = block->pred->next;
               block_list_ptr;
               block_list_ptr = block_list_ptr->next)
          {
            pred_block = Locate_Inf_Block(func, block_list_ptr->block_number);

            /*
             * Don't change any of the "intersection" vectors until
             * the end of this loop.
             * Computing int-effect is simple.
             */
            dcpy (&new_effect, intersection_effect);
            dunion (&new_effect, pred_block->out_effect);

            /*
             * Calculating int-unknown.
             */
            dcpy (&new_unknown, intersection_unknown);
            dunion (&new_unknown, pred_block->out_unknown);

            dcpy (&temp_vector, pred_block->out_effect);
            dinter (&temp_vector, intersection_jump);
            dminus (&temp_vector, pred_block->out_unknown);
            dminus (&temp_vector, pred_block->out_jump);

            dcpy (&temp_vector2, intersection_effect);
            dinter (&temp_vector2, pred_block->out_jump);
            dminus (&temp_vector2, intersection_unknown);
            dminus (&temp_vector2, intersection_jump);

            dunion (&new_unknown, temp_vector);
            dunion (&new_unknown, temp_vector2);

            /*
             * Calculating int-jump.
             */
            dcpy (&temp_vector, pred_block->out_jump);
            dminus (&temp_vector, intersection_effect);
            dcpy (&temp_vector2, intersection_jump);
            dminus (&temp_vector2, pred_block->out_effect);
            dcpy (&temp_vector3, intersection_jump);
            dinter (&temp_vector3, pred_block->out_jump);

            dcpy (&new_jump, temp_vector);
            dunion (&new_jump, temp_vector2);
            dunion (&new_jump, temp_vector3);

            /*
             * Now copy these new vectors into the intersection.
             */
            dcpy (&intersection_effect, new_effect);
            dcpy (&intersection_unknown, new_unknown);
            dcpy (&intersection_jump, new_jump);
          }
        }

        /*
         * Now that we have the intersections, assign them to block's ins.
         */
        if (! dcmp (block->in_effect, intersection_effect) ||
            ! dcmp (block->in_unknown, intersection_unknown) ||
            ! dcmp (block->in_jump, intersection_jump))
        {
	  if (verbose)
	    printc ("change becomes true:  an IN updated to new intersection\n");
          change = true;
        }
        dcpy (&block->in_effect, intersection_effect);
        dcpy (&block->in_unknown, intersection_unknown);
        dcpy (&block->in_jump, intersection_jump);

        /*
         * **************************************************************
         * Now part 2: find the outs
         * First, we find the vector information about the current block.
         */
        this_block_effect = dinit();
        this_block_unknown = dinit();
        this_block_jump = dinit();

        for (block_list_ptr = block->branch;
             block_list_ptr;
             block_list_ptr = block_list_ptr->next)
        {
          dinsert (&this_block_effect, block_list_ptr->block_number);
          dinsert (&this_block_jump, block_list_ptr->block_number);
        }

        for (block_list_ptr = block->fallthru;
             block_list_ptr;
             block_list_ptr = block_list_ptr->next)
        {
          dinsert (&this_block_effect, block_list_ptr->block_number);
        }

        for (block_list_ptr = block->unknown;
             block_list_ptr;
             block_list_ptr = block_list_ptr->next)
        {
          dinsert (&this_block_effect, block_list_ptr->block_number);
          dinsert (&this_block_unknown, block_list_ptr->block_number);
        }

	if (verbose) {
	  printc ("after looking up this block's branch effects, we have\n");
	  printc ("this block effect:\n");
	
	  showdvect (this_block_effect);
	  printc ("this block unknown:\n");
	  showdvect (this_block_unknown);
	  printc ("this block jump:\n");
	  showdvect (this_block_jump);
	}
        /*
         * For each bit:  out = this ? this : in
         *
         * this-e this-u this-j    in-e  in-u  in-j    out-e out-u out-j
         *    0      0      0        0     0     0       0     0     0
         *    0      0      0        1     1     0       1     1     0
         *    0      0      0        1     0     1       1     0     1
         *    0      0      0        1     0     0       1     0     0
         *
         *    1      1      0        0     0     0       1     1     0
         *    1      1      0        1     1     0       1     1     0
         *    1      1      0        1     0     1       1     1     0
         *    1      1      0        1     0     0       1     1     0
         *
         *    1      0      1        0     0     0       1     0     1
         *    1      0      1        1     1     0       1     0     1
         *    1      0      1        1     0     1       1     0     1
         *    1      0      1        1     0     0       1     0     1
         *
         *    1      0      0        0     0     0       1     0     0
         *    1      0      0        1     1     0       1     0     0
         *    1      0      0        1     0     1       1     0     0
         *    1      0      0        1     0     0       1     0     0
         *
         * In cases 1-4 we take the in-values because "this" is unknown.
         * In cases 5-16 "this" is known so we override the in.
         *
         * out-effect = union of this-effect and in-effect
         * out-unknown = union of this-unknown and (in_unknown - this_effect)
         * out_jump = union of this_jump and (in_jump - this_effect)
         */

        /*
         * compute the out_effect
         */
        dcpy (&new_effect, this_block_effect);
        dunion (&new_effect, block->in_effect);

        /*
         * compute the out_unknown
         */
        dcpy (&new_unknown, block->in_unknown);
        dminus (&new_unknown, this_block_effect);
        dunion (&new_unknown, this_block_unknown);

        /*
         * compute the out_jump
         */
        dcpy (&new_jump, block->in_jump);
        dminus (&new_jump, this_block_effect);
        dunion (&new_jump, this_block_jump);

        /*
         * Now, see if any of these bit vectors represents a change in outs.
         */
        if (! dcmp (block->out_effect, new_effect))
        {
	  if (verbose) {
	    printc ("change becomes true:  out_effect reassigned\n");
	    printc ("old out_effect:\n");
	    showdvect (block->out_effect);
	  }
          change = true;
        }

        if (! dcmp (block->out_unknown, new_unknown))
        {
	  if (verbose) {
	    printc ("change becomes true:  out_unknown reassigned\n");
	    printc ("old out_unknown:\n");
	    showdvect (block->out_unknown);
	  }
          change = true;
        }

        if (! dcmp (block->out_jump, new_jump))
        {
	  if (verbose) {
	    printc ("change becomes true:  out_jump reassigned\n");
	    printc ("old out_jump:\n");
	    showdvect (block->out_jump);
	  }
          change = true;
        }


        /*
         * Set the outs to what we have just calculated them to be.
         */
        dcpy (&block->out_effect, new_effect);
        dcpy (&block->out_unknown, new_unknown);
        dcpy (&block->out_jump, new_jump);

	if (verbose) {
	  printc ("here are block %d's ins and outs\n", block->block_number);
	  printc ("in effect:\n");
	  showdvect (block->in_effect);
	  printc ("in unknown:\n");
	  showdvect (block->in_unknown);
	  printc ("in jump:\n");
	  showdvect (block->in_jump);
	  printc ("effect inside this block:\n");
	  showdvect (this_block_effect);
	  printc ("jump inside this block:\n");
	  showdvect (this_block_jump);
	  printc ("out effect:\n");
	  showdvect (block->out_effect);
	  printc ("out unknown:\n");
	  showdvect (block->out_unknown);
	  printc ("out jump:\n");
	  showdvect (block->out_jump);
	}
      }
    }
  }
}


/* print_block_info - This function shows the inf information for
 *                    a block having to do with how the block may
 *                    affect conditional branches in the function. 6/29/98
 */
void print_block_info( struct inf_func_element* inf, struct tnode* node_ptr )
{
   struct block_element  *block_element_ptr;
   struct block_list     *p;
   struct loop_path_node *path_ptr;
   struct loop_block_list *block_ptr;
   int path_num = 0; 

   for (block_element_ptr = inf->block;
        block_element_ptr;
        block_element_ptr = block_element_ptr->next)
   {
     printc ("block %d\n", block_element_ptr->block_number);

     if (block_element_ptr->unknown) {
       printc ("  makes unknown  ");
       for (p = block_element_ptr->unknown;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->branch) {
       printc ("  makes branch  ");
       for (p = block_element_ptr->branch;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->fallthru) {
       printc ("  makes fallthru  ");
       for (p = block_element_ptr->fallthru;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->unfellthru) {
       printc ("  makes unknown if not fell thru before  ");
       for (p = block_element_ptr->unfellthru;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->unbranched) {
       printc ("  makes unknown if not branched before  ");
       for (p = block_element_ptr->unbranched;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->ft_cause_b) {
       printc ("  fall thru causes branch  ");
       for (p = block_element_ptr->ft_cause_b;  p;  p= p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->ft_cause_ft) {
       printc ("  fall thru cause fall thru  ");
       for (p = block_element_ptr->ft_cause_ft;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->b_cause_b) {
       printc ("  branch causes branch  ");
       for (p = block_element_ptr->b_cause_b;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
     if (block_element_ptr->b_cause_ft) {
       printc ("  branch causes fall thru  ");
       for (p = block_element_ptr->b_cause_ft;  p;  p = p->next)
         printc ("%2d  ", p->block_number);
       printc ("\n");
     }
   }
   for (path_ptr = node_ptr->loop_paths->path;
        path_ptr;
        path_ptr = path_ptr->next)
   {
     printc ("blocks in %s%s path %d:  ",
             path_ptr->path_type_mask & BACKEDGE ? "cont" : "",
             path_ptr->path_type_mask & EXIT ? "exit" : "",
             ++path_num);
     
     for (block_ptr = path_ptr->path_list;
          block_ptr;
          block_ptr = block_ptr->next)
     {
       printc ("%2d  ", block_ptr->block_number);
     }
     printc ("\n");
   }
}


struct inst_num_node * find_inst_num( struct inst_num_node *head, int num )
{
  struct inst_num_node *p;

  for (p = head; p; p = p->next)
  {
    if (p->num == num)
      return p;
  }
  return NULL;		/* if inst num given is not found, return NULL */
}



/* init_branch_info - traverse the blocks in the node, looking for
   a conditional branch instruction.  Create a list of these branches.
   There can be at most one branch per block.
   7/23/98
*/

void init_branch_info( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct block_element  *block_element_ptr;
  struct instruct *instruct_list_ptr;
  struct branch_node *temp, *head = NULL, *tail = NULL;
  int num_branches = 0;
  extern bool verbose;


  for (block_element_ptr = inf->block;
       block_element_ptr;
       block_element_ptr = block_element_ptr->next)
  {
    for (instruct_list_ptr = block_element_ptr->instruct_list;
         instruct_list_ptr;
         instruct_list_ptr = instruct_list_ptr->next)
    {
       if (is_branch (instruct_list_ptr->inst_type))
       {
         temp = (struct branch_node *)
                malloc (sizeof (struct branch_node));
         temp->number = block_element_ptr->block_number;
         temp->state = UNKNOWN;

         if (++num_branches == 1)
         {
           tail = temp;
           head = temp;
           temp->prev = NULL;
           temp->next = NULL;
         }
         else
         {
           temp->next = NULL;
           temp->prev = tail;
           tail->next = temp;
           tail = temp;
         }
	 if (verbose)
	   printc ("I see branch in block %d\n", temp->number);

         break;		/* go on to next block in node */
       }
    }
  }
  node_ptr->branch = head;
}

struct branch_node *lookup_branch( struct tnode* node_ptr, int number )
{
  struct branch_node *p;

  for (p = node_ptr->branch; p; p = p->next)
    if (p->number == number)
      return p;
  return NULL;
}

void print_branch_info( struct tnode* node_ptr )
{
  struct branch_node *p;

  for (p = node_ptr->branch; p; p = p->next)
    printc ("  branch %d = %d\t", p->number, p->state);

  printc ("\n");
}    

void clear_branch_info( struct tnode* node_ptr )
{
  struct branch_node *p;

  for (p = node_ptr->branch; p; p = p->next)
    p->state = UNKNOWN;
}

void set_branch_states( struct tnode* node_ptr, struct loop_block_list* block_ptr, struct block_element* block_element_ptr )
{
  struct block_list *block_list_ptr;
  struct branch_node *branch_ptr;

  for (block_list_ptr = block_element_ptr->unknown;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    branch_ptr->state = UNKNOWN;
  }

  for (block_list_ptr = block_element_ptr->branch;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    branch_ptr->state = BRANCH;
  }

  for (block_list_ptr = block_element_ptr->fallthru;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    branch_ptr->state = FALLTHRU;
  }

  for (block_list_ptr = block_element_ptr->unfellthru;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (branch_ptr->state != FALLTHRU)
      branch_ptr->state = UNKNOWN;
  }

  for (block_list_ptr = block_element_ptr->unbranched;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (branch_ptr->state != BRANCH)
      branch_ptr->state = UNKNOWN;
  }

  for (block_list_ptr = block_element_ptr->ft_cause_b;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (block_ptr && block_ptr->next &&
        block_ptr->next->block_number == 1 + block_ptr->block_number)
      branch_ptr->state = BRANCH;
  }

  for (block_list_ptr = block_element_ptr->ft_cause_ft;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (block_ptr && block_ptr->next &&
        block_ptr->next->block_number == 1 + block_ptr->block_number)
      branch_ptr->state = FALLTHRU;
  }

  for (block_list_ptr = block_element_ptr->b_cause_b;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (block_ptr && block_ptr->next &&
        block_ptr->next->block_number != 1 + block_ptr->block_number)
      branch_ptr->state = BRANCH;
  }

  for (block_list_ptr = block_element_ptr->b_cause_ft;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    branch_ptr = lookup_branch(node_ptr, block_list_ptr->block_number);
    if (block_ptr && block_ptr->next &&
        block_ptr->next->block_number != 1 + block_ptr->block_number)
      branch_ptr->state = FALLTHRU;
  }
}

int set_path_numbers( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;
  int n = 0;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
    path_ptr->number = ++n;

  node_ptr->num_paths = n;

  return n;			/* return number of paths found */
}

void print_path_num_list( struct path_num_node* number_list )
{
  struct path_num_node *p;

  for (p = number_list; p; p = p->next)
    printc ("%4d", p->number);
  printc ("\n");
}

/*
 * count_blocks - find out how many blocks are in the node 2/7/99
 */
void count_blocks( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;
  struct loop_block_list *block_ptr;
  int max_block_num = 0;
  extern bool verbose;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    for (block_ptr = path_ptr->path_list;
         block_ptr;
         block_ptr = block_ptr->next)
    {
      poss_update (&max_block_num, block_ptr->block_number, "");
    }
  }

  if (verbose)
    printc ("There are %d blocks in this node.\n", max_block_num);

  node_ptr->num_blocks = max_block_num;
}

/*
 * infeasible_paths - detect infeasible paths in this node 7/24/98
 * We are going to be temporarily modifying the branch states for each path,
 * so at the end we need to re-initialize the states back to UNKNOWN.
 *
 * When we are looking at a block in a path, we look up the
 * branch-affecting information to possibly update the branch states.
 *
 * For each path in the node
 *   For each block in this path
 *     set the branch states according to how this block influences branches
 *     if there's a branch in this block
 *       check for consistent state information (inconsistent --> infeasible)
 */

void infeasible_paths( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct block_element  *block_element_ptr;
  struct instruct *inf_inst_ptr;
  struct loop_path_node *path_ptr;
  struct loop_block_list *block_ptr;
  bool feasible, found_loop;
  struct branch_node *branch_ptr;
  int i;
  extern bool verbose;

  count_blocks(node_ptr);

  if (verbose)
    printc ("\nentering infeasible_paths\n");

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (verbose)
      printc ("path %d\n", path_ptr->number);

    feasible = true;
    found_loop = false;

    for (block_ptr = path_ptr->path_list;
         block_ptr;
         block_ptr = block_ptr->next)
    {
      if (verbose)
	printc ("  looking at block %d\n", block_ptr->block_number);

      block_element_ptr = Locate_Inf_Block(inf, block_ptr->block_number);

      /* Added by Sibin M on Sept 27, 2004.
       * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
       * If not, then skip over the rest of the loop.
       */
      if( !block_element_ptr )
	      continue ;

      /* End of additions by Sibin M on Sept 27, 2004.
       * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
       * If not, then skip over the rest of the loop.
       */

      /*
       * if this block has no instructions, it represents a child loop
       * 12/5/98
       */
      inf_inst_ptr = block_element_ptr->instruct_list;
      if (inf_inst_ptr &&
          ! Locate_Stats_Inst (node_ptr->instruct_list, inf_inst_ptr->number))
      {
	if (verbose)
	  printc ("  block %d is a child loop:  don't apply its constraints\n",
		  block_ptr->block_number);
        found_loop = true;
        continue;
      }

      /*
       * We do not want to eliminate a path if an inner loop nullified
       * a constraint.  2/7/99
       */
      if (found_loop)
      {
	if (verbose) {
	  printc ("  this is block after child loop\n");
	  printc ("  in_effect:\n");
	  showdvect (block_element_ptr->in_effect);
	  printc ("  in_unknown:\n");
	  showdvect (block_element_ptr->in_unknown);
	  printc ("  in_jump:\n");
	  showdvect (block_element_ptr->in_jump);
	}

        for (i = 1; i <= node_ptr->num_blocks; ++i)
        {
          if (din (block_element_ptr->in_unknown, i))
          {
	    if (verbose)
            printc ("  the in makes unknown branch in block %d\n", i);
          }
        }
      }


      /* block_element_ptr now points to this block's inf information
         because its number matches that of (the node's) block_ptr */

      /*   for each block in this list, look at the number,
           and go to the "node_ptr->branch" list and modify
           their states.                                    */

      set_branch_states(node_ptr, block_ptr, block_element_ptr);

      /* if this block contains a conditional branch, look up its
         own state to make sure it's consistent */

      branch_ptr = lookup_branch(node_ptr, block_ptr->block_number);

      if (branch_ptr)
      {
        /*
         * case where the inner loop made branch unknown 2/7/99
         */
        if (found_loop & din(block_element_ptr->in_unknown,
                             block_ptr->block_number))
          ;

        else if (branch_ptr->state == BRANCH &&
            block_ptr->next &&
            block_ptr->next->block_number == 1 + block_ptr->block_number)
        {
          feasible = false;
        }
        else if (branch_ptr->state == FALLTHRU &&
                 block_ptr->next &&
                 block_ptr->next->block_number != 1 + block_ptr->block_number)
        {
          feasible = false;
        }
      }
      found_loop = false;
    }

    /*
     * at this point we could call print_branch_info, passing the node_ptr
     * to print the status of the branches in this path
     */

    if (! feasible)
    {
      if (verbose)
	printc ("Path %d is not feasible: type set to NONE\n", path_ptr->number);

      /*
       * rather than take the path out of l.l., its type is set to
       * none so we can still use it in analysis for time being.  1 dec 98
       */
      path_ptr->path_type_mask = NONE;
    }

    /* re-initialize all branch states to UNKNOWN before
       looking at the next path */

    clear_branch_info(node_ptr);
  }

  if (verbose)
    printc ("returning from infeasible_paths\n");
}

/*
 * set_preheader_branch_states - in reality, the states of branches in
 * a loop is established in the preheader before entering the loop
 * 7/26/98
 */
void set_preheader_branch_states( struct inf_func_element* inf, struct tnode* node_ptr )
{
  int num_preheader_blocks, pred_block_num;
  struct block_element *block_element_ptr;
  struct block_list *block_list_ptr;
  struct loop_element *loop_element_ptr;

  extern bool verbose;

  /* look up header block number in inf data structure
     copied from Time_Path step 19                         */

  loop_element_ptr = Find_Loop(inf, node_ptr->loop_paths->path->header_block);
  block_element_ptr = Locate_Inf_Block(inf,
                                  node_ptr->loop_paths->path->header_block);

  num_preheader_blocks = 0;

  for (block_list_ptr = block_element_ptr->pred;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    if (! Is_In_Block_List (block_list_ptr->block_number,
                            loop_element_ptr->block_list))
    {
      pred_block_num = block_list_ptr->block_number;
      ++num_preheader_blocks;
      if (num_preheader_blocks > 1)
      {
	if (verbose)
	  printc ("HEY - there's more than 1 pred block.  Stop now.\n");

        block_element_ptr = NULL;
        break;
      }

      if (verbose)
	printc ("Pred block %d is not in loop\n",
		block_list_ptr->block_number);

      block_element_ptr = Locate_Inf_Block(inf, pred_block_num);
    }
  }

  if (block_element_ptr)
  {
    /* still need a block pointer before we can call set_branch_states ? */

    set_branch_states(node_ptr, (struct loop_block_list *) NULL,
                      block_element_ptr);
    if (verbose)
      print_branch_info(node_ptr);
  }

}


/*
 * This function is like Rob's for looking up a block list, but for
 * a different type.
 */
int is_in_loop_block_list( int number, struct loop_block_list* list )
{
  struct loop_block_list *list_ptr;

  for (list_ptr = list; list_ptr; list_ptr = list_ptr->next)
    if (number == list_ptr->block_number)
      return 1;
  return 0;
}

/*
 * is_in_loop - Determine if a particular block number is that of one
 * in the specified loop.
 */
int is_in_loop( int number, struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;


  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (is_in_loop_block_list (number, path_ptr->path_list))
      return 1;
  }
  return 0;
}


/*
 * path_influence - see how each path influences what paths can
 * execute on the next iteration 7/25/98
 *
 * initialize branch states for loop preheader
 * For each path P in the node
 *   set branch states for all blocks in the path P
 *   for each path Q in the node
 *     for each block in the path Q
 *       if there's a branch in this block
 *         check for consistent branch-state information
 *         Here, inconsistent means Q can't execute immediately after P
 *     if Q can't follow P, put Q in P's "can't list"
 */
void path_influence( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr, *path_ptr2;
  struct loop_block_list *block_ptr;
  struct block_element *block_element_ptr;
  struct branch_node *branch_ptr;
  int succ_block_num, i;
  bool feasible, found_loop;
  struct path_num_node *temp, *head, *tail;
  struct block_element *inf_bptr;
  struct instruct *inf_inst_ptr;
  extern bool verbose;

  if (verbose)
    printc ("\nentering path_influence\n");

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    set_preheader_branch_states(inf, node_ptr);		/* initialize states */

    temp = head = tail = NULL;
    if (verbose)
      printc ("path P = %d\n", path_ptr->number);

    for (block_ptr = path_ptr->path_list;
         block_ptr;
         block_ptr = block_ptr->next)
    {
      block_element_ptr = Locate_Inf_Block(inf, block_ptr->block_number);

      /* Added by Sibin M on Sept 27, 2004.
       * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
       * If not, then skip over the rest of the loop.
       */
      if( !block_element_ptr )
	      continue ;

      /* End of additions by Sibin M on Sept 27, 2004.
       * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
       * If not, then skip over the rest of the loop.
       */

      set_branch_states(node_ptr, block_ptr, block_element_ptr);
    }

    if (verbose)
      print_branch_info(node_ptr);

    for (path_ptr2 = node_ptr->loop_paths->path;
         path_ptr2;
         path_ptr2 = path_ptr2->next)
    {
      feasible = true;
      found_loop = false;

      for (block_ptr = path_ptr2->path_list;
           block_ptr;
           block_ptr = block_ptr->next)
      {
        branch_ptr = lookup_branch(node_ptr, block_ptr->block_number);


        /*
         * This code before the "if (branch_ptr)" is added so that
         * we can detect if an inner loop will nullify a constraint.
         * Unfortunately, it hurts lu100 by 119801  5/19/99
         */
        block_element_ptr = Locate_Inf_Block(inf, block_ptr->block_number);

	/* Added by Sibin M on Sept 27, 2004.
	 * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
	 * If not, then skip over the rest of the loop.
         */
        if( !block_element_ptr )
  	      continue ;

        /* End of additions by Sibin M on Sept 27, 2004.
         * Check to see if a valid block pointer has been returned by the Locate_Inf_Block.
         * If not, then skip over the rest of the loop.
         */

        /*
         * if this block has no instructions, it represents a child loop
         * Copied from infeasible_paths  5/19/99
         */
        inf_inst_ptr = block_element_ptr->instruct_list;
        if (inf_inst_ptr &&
            ! Locate_Stats_Inst (node_ptr->instruct_list, inf_inst_ptr->number))
        {
          /*
           * We do not want to eliminate a path if an inner loop nullified
           * a constraint.
           */
          found_loop = true;
	  if (verbose) {
	    printc ("  block %d is after child loop\n",
		    block_ptr->block_number);
	    printc ("  in_effect:\n");
	    showdvect (block_element_ptr->in_effect);
	    printc ("  in_unknown:\n");
	    showdvect (block_element_ptr->in_unknown);
	    printc ("  in_jump:\n");
	    showdvect (block_element_ptr->in_jump);
	  }
          for (i = 1; i <= node_ptr->num_blocks; ++i)
          {
            if (din (block_element_ptr->in_unknown, i))
            {
	      if (verbose)
		printc ("  the in makes unknown branch in block %d\n", i);
            }
          }
        }


        if (branch_ptr)
        {
          if (found_loop)
          {
            if (din(block_element_ptr->in_unknown, block_ptr->block_number))
            {
	      if (verbose)
		printc ("inner loop nullifies constraint\n");
              continue;
            }
          }
          if (branch_ptr->state == BRANCH &&
              block_ptr->next &&
              block_ptr->next->block_number == 1 + block_ptr->block_number)
          {
	    if (verbose)
	      printc ("bad transition:  branch state, but next is fall thru\n");
            feasible = false;
          }

          /* added test to be sure not header of child loop 3/1/01 */

          else if (branch_ptr->state == FALLTHRU &&
                   block_ptr->next &&
                   block_ptr->next->block_number != 1 + block_ptr->block_number
                     &&
                  ! found_loop)
          {
	    if (verbose) {
	      printc ("bad transition: fallthru state, but next not fall thru\n");

            /* new diagnostic output 3/1/01 */
	      printc ("  this block = %d, next block = %d\n",
		      block_ptr->block_number, block_ptr->next->block_number);
	    }

            feasible = false;
          }

          /*
           * If path2 is a continue path, and there is a branch in its
           * last block, and we know its state to not lead us to the header;
           * of if path2 is an exit path, and there is a branch in its
           * last block, and we know its state to not lead us to a block
           * outside the loop, then this path2 cannot feasibly follow path1.
           * 11/26/98
           */
          if (! block_ptr->next)
            if (path_ptr2->path_type_mask & BACKEDGE)
            {
              /*
               * To be feasible, the state of the branch
               * must lead us to header
               * branch_ptr->state is either FALLTHRU or BRANCH.
               * We look up the header block number,
               * and the number of this block.
               */

	      if (verbose)
		printc ("%d is last block in poss dest cont path\n",
			block_ptr->block_number);

              if (branch_ptr->state == FALLTHRU &&
                  node_ptr->loop_header_block !=
                              1 + block_ptr->block_number)
                feasible = false;
              if (branch_ptr->state == BRANCH &&
                  node_ptr->loop_header_block ==
                              1 + block_ptr->block_number)
                feasible = false;

            }
            else if (path_ptr2->path_type_mask == EXIT)
            {
              /*
               * To be feasible, the state of the branch
               * must lead us to a block outside of loop
               * In other words, if the next block we execute is in
               * the loop, then this transition is infeasible.
               */

	      if (verbose)
		printc ("%d is last block in poss dest exit path %d\n",
			block_ptr->block_number, path_ptr2->number);

              if (branch_ptr->state == FALLTHRU)
              {
		if (verbose)
		  printc ("fallthru state\n");

                /*
                 * If our block_number + 1 is in the loop, infeasible.
                 */
                if (is_in_loop (1 + block_ptr->block_number, node_ptr))
                  feasible = false;
              }
              if (branch_ptr->state == BRANCH)
              {
		if (verbose)
		  printc ("branch state\n");

                /*
                 * Find out which successor block is not block_number + 1,
                 * this must be the branch target.  If the branch target
                 * is in the loop, infeasible.
                 */
                inf_bptr = Locate_Inf_Block (inf, block_ptr->block_number);

                if (inf_bptr->succ->block_number != 1 + inf_bptr->block_number)
                  succ_block_num = inf_bptr->succ->block_number;
                else
                  succ_block_num = inf_bptr->succ->next->block_number;

		if (verbose)
		  printc ("successor block number is %d\n", succ_block_num);

                if (is_in_loop (succ_block_num, node_ptr))
                  feasible = false;
              }
                
            }
        }
      }

      /*
       * Nothing can follow an exit path.  Any transition from an exit
       * path is infeasible.  11/26/98
       */
      if (path_ptr->path_type_mask == EXIT)
      {
	if (verbose)
	  printc ("  nothing can come after an exit path\n");
        feasible = false;
      }

      if (feasible)
      {
	if (verbose)
	  printc ("  %d-->%d is feasible\n", path_ptr->number,
		  path_ptr2->number);
      }
      else
      {
        temp = (struct path_num_node *)
               malloc (sizeof (struct path_num_node));
        temp->number = path_ptr2->number;
        if (! head)
        {
          tail = temp;
          head = temp;
          temp->prev = NULL;
          temp->next = NULL;
        }
        else
        {
          temp->next = NULL;
          temp->prev = tail;
          tail->next = temp;
          tail = temp;
        }
      }
    }
    path_ptr->cant_next = head;

    if (verbose) {
      printc ("Here is the \"can't go next\" list for path %d:\n",
	      path_ptr->number);
      print_path_num_list(path_ptr->cant_next);
    }
  }

  if (verbose)
    printc ("returning from path_influence\n");
}

/*
 * can_imm_follow - See if one path can immediately follow another.
 */
int can_imm_follow( struct loop_path_node* path_ptr1, struct loop_path_node* path_ptr2 )
{
  struct path_num_node *p;

  for (p = path_ptr1->cant_next; p; p = p->next)
    if (p->number == path_ptr2->number)
      return 0;
  return 1;
}

/*
 * create_imm_follow - Create the 2-D array that stores which paths
 * can and cannot immediately follow the others.
 * Because paths are numbered from 1, we don't use row or column 0.
 * So "num_paths" is 1 extra.
 */
void create_imm_follow( struct tnode* node_ptr )
{
  static int **imm_follow;
  int *row;
  int i, num_paths = 1;
  struct loop_path_node *path_ptr1, *path_ptr2;

  num_paths = 1 + node_ptr->num_paths;

  imm_follow = (int **) calloc (num_paths, sizeof (int *));
  for (i = 0; i < num_paths; ++i)
  {
    row = (int *) calloc (num_paths, sizeof (int));
    imm_follow[i] = row;
  }

  for (path_ptr1 = node_ptr->loop_paths->path;
       path_ptr1;
       path_ptr1 = path_ptr1->next)
    for (path_ptr2 = node_ptr->loop_paths->path;
         path_ptr2;
         path_ptr2 = path_ptr2->next)
      imm_follow[path_ptr1->number][path_ptr2->number] =
                           can_imm_follow (path_ptr1, path_ptr2) ? 1 : 0;

  node_ptr->imm_follow = imm_follow;
}

/*
 * print_imm_follow - Print the matrix of paths that can or cannot follow,
 * using Ys and Ns.  11/27/98
 */
void print_imm_follow( struct tnode* node_ptr )
{
  int i, j;

  printc ("This is the path immediately follow array:\n");

  for (i = 1; i <= node_ptr->num_paths; ++i)
  {
    for (j = 1; j <= node_ptr->num_paths; ++j)
      printc ("%-3c", node_ptr->imm_follow[i][j] ? 'Y' : 'N');
    printc ("\n");
  }
}

/*
 * is_dead_end - Given the path number, see if it is that of an exit
 * path or infeasible path within the given node (loop).  11/28/98
 * A "none" path type means infeasible, so it is a dead end.
 */
int is_dead_end( struct tnode* node_ptr, int number )
{
  struct loop_path_node *path_ptr;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
    if (path_ptr->number == number)
      break;

  if (! path_ptr)
    error ("is_dead_end(): can't find path within node");

  return path_ptr->path_type_mask == EXIT || path_ptr->path_type_mask == NONE;
}

/*
 * is_none - Given a path number, tell if its the number of an infeasible
 * path.
 */
int is_none( struct tnode* node_ptr, int number )
{
  struct loop_path_node *path_ptr;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
    if (path_ptr->number == number)
      break;

  if (! path_ptr)
    error ("is_none(): can't find path within node");

  return path_ptr->path_type_mask == NONE;
}

/*
 * how_long_to_reach - How many iterations we need to traverse before
 * getting from one path to another.  If we can never make it, return
 * the value 0.  Note that we cannot return a value greater than the
 * number of paths:  the worst case would be to have to go through
 * every other path.  11/27/98
 * When we call this recursive procedure from the outside, the value
 * of mask should be 0 representing that we haven't gone to any paths yet.
 */
int how_long_to_reach( struct tnode* node_ptr, int num1, int num2, int mask )
{
  struct loop_path_node *path_ptr;
  int ret_val, dist, shortest_dist;
  extern bool verbose;

  if (verbose)
    printc ("entering how_long_to_reach with num1 = %d, num2 = %d, mask = %d\n",
	    num1, num2, mask);

  if (node_ptr->imm_follow[num1][num2])
  {
    if (verbose)
      printc ("  returning 1\n");
    return 1;
  }
  else
  {
    /*
     * We can't go directly, so we will go to wherever we can directly
     * hoping that we eventually get to num2.  If num1 is an exit path,
     * then we should immediately return 0.
     */
    if (is_dead_end (node_ptr, num1))
    {
      if (verbose)
	printc ("  it's a dead end:  returning 0\n");
      return 0;
    }
    /*
     * If the path we're going to is infeasible, then we can't get there.
     * 12/6/98
     */
    if (is_none (node_ptr, num2))
    {
      if (verbose)
	printc ("  dest path is infeasible:  returning 0\n");
      return 0;
    }
    /*
     * have a for-loop go thru all the paths that we can immediately reach.
     * and also mark where we've been.  don't go down a path that we've
     * encountered already.  add 1 for each path in the traversal
     */
    mask |= 1 << num1;
    shortest_dist = MAXINT;
    for (path_ptr = node_ptr->loop_paths->path;
         path_ptr;
         path_ptr = path_ptr->next)
    {
      dist = 0;

      /*
       * if we can get to this path immediately and haven't been there
       * already, go
       */
      if (node_ptr->imm_follow[num1][path_ptr->number] &&
          ! (mask & 1 << path_ptr->number))
      {
	if (verbose)
	  printc ("  about to make recursive call with %d, %d, %d\n",
		  path_ptr->number, num2, mask);

        /*
         * if our recursive call can't get there, then we can't get
         * there via this recursive call    Dec. 1, 1998
         */
        ret_val = how_long_to_reach (node_ptr, path_ptr->number, num2, mask);
        if (ret_val == 0)
        {
	  if (verbose)
	    printc ("  just returned -- can't go that way\n");
          dist = 0;
        }
        else
        {
	  if (verbose)
	    printc ("  just returned; adding one to total dist\n", ret_val + 1);
          dist = ret_val + 1;
        }
      }

      if (dist > 0 && dist < shortest_dist)
      {
	if (verbose)
	  printc ("  revising shortest distance to %d\n", dist);
        shortest_dist = dist;
      }

      /*
       * if we can't get to this path immediately, try the next path
       */
    }
    if (shortest_dist == MAXINT)
      shortest_dist = 0;			/* not reachable */
    return shortest_dist;
  }

}

/*
 * reachability - use the above how_long_to_reach repeatedly to build
 * information for all the paths.  started 11/28/98
 * store in dynamically allocated 2D array
 */
void reachability( struct tnode* node_ptr )
{
  int i, j, *row, num_paths;
  static int** reach;
  extern bool verbose;

  if (verbose)
    printc ("entering reachability\n");

  num_paths = 1 + node_ptr->num_paths;

  reach = (int **) calloc (num_paths, sizeof (int *));
  for (i = 0; i < num_paths; ++i)
  {
    row = (int *) calloc (num_paths, sizeof (int));
    reach[i] = row;
  }

  for (i = 1; i <= node_ptr->num_paths; ++i)
    for (j = 1; j <= node_ptr->num_paths; ++j)
    {
      if (verbose)
	printc ("about to find reach for i=%d, j=%d\n", i, j);
      reach[i][j] = how_long_to_reach (node_ptr, i, j, 0);
    }
  node_ptr->path_reach = reach;

  if (verbose)
    printc ("returning from reachability\n");
}

/*
 * print_path_reach - Diagnostic output of 2D array path_reach for a loop.
 */
void print_path_reach( struct tnode* node_ptr )
{
  int i, j;

  printc ("here is path reach info\n");

  for (i = 1; i <= node_ptr->num_paths; ++i)
  {
    for (j = 1; j <= node_ptr->num_paths; ++j)
      printc ("%-3d", node_ptr->path_reach[i][j]);
    printc ("\n");
  }
}

/*
 * find_unique_iters - Find range of iterations that are unique to each
 * path.  12/2/98
 */
void find_unique_iters( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr, *path_ptr2;
  int reduce_amount;
  extern bool verbose;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (path_ptr->path_type_mask == NONE)
      continue;


    path_ptr->unique_min = path_ptr->min;
    path_ptr->unique_max = path_ptr->max;

    if (verbose)
      printc ("path %d initializing unique range to [%d..%d]\n",
	      path_ptr->number, path_ptr->min, path_ptr->max);

    for (path_ptr2 = node_ptr->loop_paths->path;
         path_ptr2;
         path_ptr2 = path_ptr2->next)
    {
      if (path_ptr2->path_type_mask == NONE || path_ptr == path_ptr2)
        continue;
      if (verbose)
	printc ("  examining competing path %d range %d..%d\n",
		path_ptr2->number, path_ptr2->min, path_ptr2->max);

      /*
       * identical point ranges 1/16/99
       */
      if (path_ptr->min == path_ptr->max &&
          path_ptr2->min == path_ptr2->max &&
          path_ptr->min == path_ptr2->min)
      {
	if (verbose)
	  printc ("making path_ptr's take away 1 since both pts are the same\n");
        path_ptr->take_away = 1;
      }

      /*
       * case of a point range and the competing range may have this
       * same point deleted in itself:  don't change this path's range
       */
      if (path_ptr->min == path_ptr->max &&
          path_ptr2->min <= path_ptr->min &&
          path_ptr2->max >= path_ptr->max &&
          path_ptr2->take_away == 1)
      {
	if (verbose)
	  printc ("  not changing unique range:  path_ptr2's take away = %d\n",
		  path_ptr2->take_away);
        continue;
      }
      /*
       * case of path2 having a point range.
       */
      if (path_ptr2->min == path_ptr2->max &&
          path_ptr->min <= path_ptr2->min &&
          path_ptr->max >= path_ptr2->max &&
          path_ptr->take_away == 1)
      {
	if (verbose)
	  printc ("  not changing unique range:  path_ptr's take away = %d\n",
		  path_ptr->take_away);
        continue;
      }

      /*
       * There are six cases.
       *      ********
       *                 ********    no change
       *
       *      ********
       *          ********           trim path_ptr->unique_max down to
       *                             path_ptr2->min - 1
       *      ********
       *   ********                  trim path_ptr->unique_min up to
       *                             path_ptr2->max + 1
       *            ********
       *   *******                   no change
       *
       *          ******
       *   *******************       becomes empty set
       *
       *
       *   *******************
       *          ******             no change to range, but take_away becomes
       *                             path_ptr-> max - min + 1.
       *
       * Also need to take into consideration the max iters of the
       * competing path.  We can't reduce the set of unique iters of
       * path_ptr by more than path_ptr2->max_iter.
       */

      else if (path_ptr2->min > path_ptr->max)
        continue;

      else if (path_ptr2->min >= path_ptr->min &&
               path_ptr2->min <= path_ptr->max &&
               path_ptr2->max >= path_ptr->max)
      {

        if (path_ptr2->max_iter == 1 && path_ptr->take_away > 0 &&
            path_ptr->already_took_away)
          continue;						/* fix ex4 */

        reduce_amount = path_ptr->unique_max - (path_ptr2->min - 1);
        if (reduce_amount > path_ptr2->max_iter)
        {
	  if (verbose)
	    printc ("  can only reduce by competing path's max = %d\n",
		    path_ptr2->max_iter);
          path_ptr->take_away += path_ptr2->max_iter;
        }
        else
        {
          path_ptr->unique_max = path_ptr2->min - 1;
	  if (verbose)
	    printc ("  changing unique max to %d\n", path_ptr->unique_max);
        }
      }

      else if (path_ptr2->min < path_ptr->min &&
               path_ptr2->max >= path_ptr->min &&
               path_ptr2->max <= path_ptr->max)
      {
        if (path_ptr2->max_iter == 1 && path_ptr->take_away > 0)
          continue;						/* fix ex4 */

        reduce_amount = (path_ptr2->max + 1) - path_ptr->unique_min;
        if (reduce_amount > path_ptr2->max_iter)
        {
	  if (verbose)
	    printc ("  can only reduce by competing path's max = %d\n",
		    path_ptr2->max_iter);
          path_ptr->take_away += path_ptr2->max_iter;
        }
        else
        {
          path_ptr->unique_min = path_ptr2->max + 1;
	  if (verbose)
	    printc ("  changing unique min to %d\n", path_ptr->unique_min);
        }
      }

      else if (path_ptr2->max < path_ptr->min)
        continue;

      else if (path_ptr2->min <= path_ptr->min &&
               path_ptr2->max >= path_ptr->max)
      {
        if (path_ptr2->max_iter == 1 && path_ptr->take_away > 0)
          continue;						/* fix ex4 */

        reduce_amount = path_ptr->max - path_ptr->min + 1;
        if (reduce_amount > path_ptr2->max_iter)
        {
	  if (verbose)
	    printc ("  can only reduce by competing path's max = %d\n",
		    path_ptr2->max_iter);
          path_ptr->take_away += path_ptr2->max_iter;
        }
        else
        {
          path_ptr->unique_min = 0;
          path_ptr->unique_max = 0;
	  if (verbose)
	    printc ("  unique range becomes empty\n");
        }
      }

      else if (path_ptr2->min >= path_ptr->min &&
               path_ptr2->min <= path_ptr->max &&
               path_ptr2->max >= path_ptr->min &&
               path_ptr2->max <= path_ptr->max)
      {
        if (path_ptr2->min != path_ptr2->max)  /* don't penalize twice */
        {
          path_ptr->take_away += path_ptr2->max - path_ptr2->min + 1;
	  if (verbose)
	    printc ("  incrmenting take_away by %d\n",
		    path_ptr2->max - path_ptr2->min + 1);
        }
      }

      else
        error ("  find_unique_iters(): range comparison case not handled\n");
    }

    /*
     * check to see if there are no unique iterations left for this path
     */
    if (path_ptr->unique_max < path_ptr->unique_min)
      path_ptr->unique_min = path_ptr->unique_max = 0;

    if (verbose)
      printc ("path %d unique iterations are [%d..%d]\n",
	      path_ptr->number, path_ptr->unique_min, path_ptr->unique_max);
  }
}

/*
 * print_path_iters - For this node, print the range of iterations
 * and the min and max iterations of every path.
 */
void print_path_iters( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    printc ("path %d  [%d..%d]  min_iter = %d   max_iter = %d\n",
            path_ptr->number, path_ptr->min, path_ptr->max,
            path_ptr->min_iter, path_ptr->max_iter);
  }
}

/*
 * ranges_intersect - See whether [a..b] and [c..d] have any elements in
 * common.  1/19/99
 */
int ranges_intersect( int a, int b, int c, int d )
{
  if (a > b || c > d)
    error ("ranges_intersect():  range is invalid");

  if (c > b || a > d)
    return 0;
  else
    return 1;
}

/*
 * compute_ranges - Find a unique range for each path.  We will create
 * a linked list of ranges of iterations.  Corresponding to each range
 * will be a list of paths that encompass that range, or part of that
 * range.  Note that the ranges given in the nodes of the linked list
 * are mutually exclusive.  Started 1/18/99.
 * We assume there will be at most 31 paths so the list of paths for
 * each range is implemented as a bit vector.
 *
 * Our first node will be the range for the first path.
 * For each path after the first in the node_ptr,
 *   If its range of iterations is mutually exclusive to all ranges
 *     that have been created, then create a new range.
 *   Otherwise, we have a situation in which the new path under consideration
 *     has a range of iterations that intersects with range(s) that have
 *     already been created.
 *     If it intersects with just one range, expand the existing range to
 *       include the new paths' range:  [1..50] + [25..60] = [1..60].
 *     If it intersects with more than one (mutually exclusive) existing range,
 *       then we need to collapse into a single range in the l.l.
 *       (for now, we just give up in this case, which is rare)
 */
void compute_ranges( struct tnode* node_ptr )
{
  struct range_node *range_head = NULL, *range_tail = NULL, *range_temp;
  struct loop_path_node *path_ptr;
  int num_intersections, num_ranges_created = 0;
  extern bool verbose;

  if (verbose)
    printc ("\nentering compute_ranges\n");
  
  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (verbose)
      printc ("looking at path %d\n", path_ptr->number);

    if (path_ptr->path_type_mask == NONE)
      continue;

    /*
     * The following is another infeasible case that hasn't yet been
     * handled until late in find_compares.
     */
    if (path_ptr->min > path_ptr->max)
      continue;

    /*
     * For the first path, we initialize the linked list of ranges.
     */
    if (num_ranges_created == 0)
    {
      range_temp = (struct range_node *)
                   malloc (sizeof (struct range_node));
      num_ranges_created = 1;
      path_ptr->range_index = 1;
      range_temp->number = 1;
      range_temp->min = path_ptr->min;
      range_temp->max = path_ptr->max;
      range_temp->path_mask = 1 << path_ptr->number;

      range_tail = range_temp;
      range_head = range_temp;
      range_temp->prev = NULL;
      range_temp->next = NULL;

      if (verbose)
	printc ("just created first range node for [%d..%d]\n",
		range_temp->min, range_temp->max);
    }

    /*
     * After the first path, we have to compare the range of iterations
     * for path_ptr against each range in the linked list so far.
     */
    else
    {
      num_intersections = 0;

      for (range_temp = range_head;
           range_temp;
           range_temp = range_temp->next)
      {
        if (ranges_intersect (range_temp->min, range_temp->max,
                              path_ptr->min, path_ptr->max))
          ++num_intersections;
      }

      if (verbose)
	printc ("I found %d intersections\n", num_intersections);

      /*
       * If no intersection, just create a new range.
       */
      if (num_intersections == 0)
      {
        range_temp = (struct range_node *) malloc (sizeof (struct range_node));
        range_temp->number = ++num_ranges_created;
        path_ptr->range_index = range_temp->number;
        range_temp->min = path_ptr->min;
        range_temp->max = path_ptr->max;
        range_temp->path_mask = 1 << path_ptr->number;

        range_temp->next = NULL;
        range_temp->prev = range_tail;
        range_tail->next = range_temp;
        range_tail = range_temp;

	if (verbose)
	  printc ("just created a new range for [%d..%d]\n",
		  range_temp->min, range_temp->max);
      }

      /*
       * If one intersection, expand the existing range.  Need to go back
       * to the l.l. to find the proper range to modify.
       */
      else if (num_intersections == 1)
      {
        for (range_temp = range_head;
             range_temp;
             range_temp = range_temp->next)
          if (ranges_intersect (range_temp->min, range_temp->max,
                                path_ptr->min, path_ptr->max))
            break;

        path_ptr->range_index = range_temp->number;
        range_temp->path_mask |= 1 << path_ptr->number;

        if (path_ptr->min < range_temp->min)
          range_temp->min = path_ptr->min;
        if (path_ptr->max > range_temp->max)
          range_temp->max = path_ptr->max;
      }

      /*
       * If > 1 intersection, collapse to a single range.
       * For the time being, for simplicity we will just destroy the l.l.
       */
      else
      {
	if (verbose)
	  printc ("I found > 1 itersection, so I'm quitting the l.l.\n");

        node_ptr->range_list = NULL;
        return;
      }

    }
  }
  node_ptr->range_list = range_head;


  /*
   * once we have the node's complete set of ranges, we can assign a maximum
   * number of iterations to each range.
   */
  for (range_temp = node_ptr->range_list;
       range_temp;
       range_temp = range_temp->next)
    range_temp->max_iter = range_temp->max - range_temp->min + 1;

  if (verbose)
    printc ("created a node_ptr->range_list; returning from compute_ranges\n\n");
}

/*
 * print_range_list - Print the ranges and corresponding path # information
 * for a node. 1/19/99
 */
void print_range_list( struct tnode* node_ptr )
{
  struct range_node *p;

  printc ("Here is range list.\n");
  for (p = node_ptr->range_list; p; p = p->next)
  {
    printc ("#%d is [%d..%d]  max iter = %d  path mask = %d\n",
            p->number, p->min, p->max, p->max_iter, p->path_mask);
  }
}

/*
 * preheader_check - Look at the outs of the preheaders to see if we
 * can trim down a max iter in the case of alternating paths.  started 2/2/99
 */
void preheader_check( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct range_node *range_node_ptr;
  struct loop_path_node *path_ptr;
  struct loop_element *loop_element_ptr;
  struct block_element *block_element_ptr, *pre_block, *inf_bptr;
  struct block_list *block_list_ptr;
  struct loop_block_list *block_ptr;
  int sum_path_max = 0, num_preheaders = 0, header_block_num, num_preds,
      succ_block_num;
  dvector preheader_out_effect, preheader_out_unknown, preheader_out_jump,
          pred_out_effect, pred_out_unknown, pred_out_jump;
  extern bool verbose;

  preheader_out_effect = dinit();
  preheader_out_unknown = dinit();
  preheader_out_jump = dinit();

  pred_out_effect = dinit();
  pred_out_unknown = dinit();
  pred_out_jump = dinit();

  /*
   * First, find the range that contains the first iteration of the loop.
   */
  for (range_node_ptr = node_ptr->range_list;
       range_node_ptr;
       range_node_ptr = range_node_ptr->next)
  {
    if (range_node_ptr->min == 1)
      break;
  }

  if (! range_node_ptr)
    return;				/* cure segv 4/19/99 */

  /*
   * Find the sum of all the max_iters of every path in this range's mask.
   * If this sum is <= the range's max, then there is nothing to trim away.
   */

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (range_node_ptr->path_mask & 1 << path_ptr->number)
      sum_path_max += path_ptr->max_iter;
  }

  if (verbose)
    printc ("range's max = %d, sum of paths max = %d\n",
	    range_node_ptr->max_iter, sum_path_max);

  if (sum_path_max <= range_node_ptr->max_iter)
    return;

  /*
   * At this point we know that the sum of the paths' max iters is greater
   * than that of the range.  Now, go to the preheaders' outs to find out who
   * should go first in the loop.
   */
  header_block_num = node_ptr->loop_paths->path->header_block;
  loop_element_ptr = Find_Loop (inf, header_block_num);
  block_element_ptr = Locate_Inf_Block (inf, header_block_num);
  for (block_list_ptr = block_element_ptr->pred;
       block_list_ptr;
       block_list_ptr = block_list_ptr->next)
  {
    if (! Is_In_Block_List (block_list_ptr->block_number,
                            loop_element_ptr->block_list))
    {
      pre_block = Locate_Inf_Block (inf, block_list_ptr->block_number);
      ++num_preheaders;
      if (num_preheaders == 1)
      {
        dcpy (&preheader_out_effect, pre_block->out_effect);
        dcpy (&preheader_out_unknown, pre_block->out_unknown);
        dcpy (&preheader_out_jump, pre_block->out_jump);
      }
      else
      {
        dunion (&preheader_out_effect, pre_block->out_effect);
        dunion (&preheader_out_unknown, pre_block->out_unknown);
        dunion (&preheader_out_jump, pre_block->out_jump);
      }
    }
  }

  if (verbose) {
    printc ("preheader out effect:\n");
    showdvect (preheader_out_effect);
    printc ("preheader out unknown:\n");
    showdvect (preheader_out_unknown);
    printc ("preheader out jump:\n");
    showdvect (preheader_out_jump);
  }

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if ((range_node_ptr->path_mask & 1 << path_ptr->number) == 0)
      continue;					/* not in this range */

    if (verbose)
      printc ("\nlooking at path %d\n", path_ptr->number);

    /*
     * See if this path obeys the preheader outs.
     */
    for (block_ptr = path_ptr->path_list;
         block_ptr;
         block_ptr = block_ptr->next)
    {
      if (! din (preheader_out_effect, block_ptr->block_number))
        continue;		/* preheader has no comment about this block */
#if 0
      /*
       * If this is not the header, then find out the predecessor outs
       * as well and combine them with the preheader.  5/19/99
       * This addition fixes the additional 15 cycles overestimation in neville
       * over its "ignore value dependence" analysis, but it hurts lu100 by
       * 17416 and sumoddeven by 3, both by 0.1 percentage points.
       */
      if (block_ptr->block_number != header_block_num)
      {
	if (verbose)
	  printc ("Looking up pred outs for block %d\n",
		  block_ptr->block_number);
        block_element_ptr = Locate_Inf_Block (inf, block_ptr->block_number);
        num_preds = 0;
        for (block_list_ptr = block_element_ptr->pred;
             block_list_ptr;
             block_list_ptr = block_list_ptr->next)
        {
          pre_block = Locate_Inf_Block (inf, block_list_ptr->block_number);
	  if (verbose)
	    printc ("%d is a predecessor block\n", pre_block->block_number);
          ++num_preds;
          if (num_preds == 1)
          {
            dcpy (&pred_out_effect, pre_block->out_effect);
            dcpy (&pred_out_unknown, pre_block->out_unknown);
            dcpy (&pred_out_jump, pre_block->out_jump);
          }
          else
          {
            dunion (&pred_out_effect, pre_block->out_effect);
            dunion (&pred_out_unknown, pre_block->out_unknown);
            dunion (&pred_out_jump, pre_block->out_jump);
          }
        }
	if (verbose) {
	  printc ("pred out effect:\n");
	  showdvect (pred_out_effect);
	  printc ("pred out unknown:\n");
	  showdvect (pred_out_unknown);
	  printc ("pred out jump:\n");
	  showdvect (pred_out_jump);
	}
        dunion (&preheader_out_effect, pred_out_effect);
        dunion (&preheader_out_unknown, pred_out_unknown);
        dinter (&preheader_out_jump, pred_out_jump);
      }
#endif

      /*
       * Let's first find the outside successor block just in case we are
       * going out of the loop.  Make it 0 if there isn't one.  6/16/99
       */
      if (verbose)
	printc ("block %d\n", block_ptr->block_number);
      inf_bptr = Locate_Inf_Block (inf, block_ptr->block_number);
      if (! is_in_loop (inf_bptr->succ->block_number, node_ptr))
        succ_block_num = inf_bptr->succ->block_number;

      else if (inf_bptr->succ->next &&
               ! is_in_loop (inf_bptr->succ->next->block_number, node_ptr))
        succ_block_num = inf_bptr->succ->next->block_number;

      else
        succ_block_num = 0;
      if (verbose)
	if (succ_block_num != 0) 
	  printc ("outside successor block number is %d\n", succ_block_num);

      /*
       * If there is a disagreement between what this path does
       * and the outs, then it can't execute on the 1st iteration.
       * So, instead of decrementing the path's max_iter, set
       * its min to 2.  6/16/99
       * The next thing to do would be to increment this 2 possibly
       * further based on a minimum path_reach number
       */
      if (din (preheader_out_jump, block_ptr->block_number))
      {
	if (verbose) {
	  printc ("the preheader says the branch in this block must jump\n");
	}
        if (block_ptr->next)
        {
          if (block_ptr->next->block_number == 1 + block_ptr->block_number)
          {
	    if (verbose)
	      printc ("path %d, block %d was supposed to jump but fell thru\n",
		      path_ptr->number, block_ptr->block_number);
            path_ptr->min = 2;
          }
        }
        else
        {
          /*
           * In this case, the block is at the end of the path, so
           * the next block to execute is certainly the one outside the loop
           */
          if (succ_block_num == 1 + block_ptr->block_number)
          {
	    if (verbose)
	      printc ("path %d, block %d was supposed to jump but fell thru\n",
		      path_ptr->number, block_ptr->block_number);
            path_ptr->min = 2;
          }
        }
      }
      else if (! din (preheader_out_unknown, block_ptr->block_number))
      {
	if (verbose)
	  printc ("the preheader says the branch in this block must fall thru\n");
        if (block_ptr->next)
        {
          if (block_ptr->next->block_number != 1 + block_ptr->block_number)
          {
	    if (verbose)
	      printc ("path %d, block %d was supposed to fall thru but jumped\n",
		      path_ptr->number, block_ptr->block_number);
            path_ptr->min = 2;
          }
        }
        else
        {
          if (succ_block_num != 1 + block_ptr->block_number)
          {
	    if (verbose)
	      printc ("path %d, block %d was supposed to fall thru but jumped\n",
		      path_ptr->number, block_ptr->block_number);
            path_ptr->min = 2;
          }
        }
      }
    }
  }
}

/*
 * first_turn - For each path that cannot execute on the first iteration,
 * find out when it can first execute, based on how long it takes to be
 * reached from those paths that can execute on the first iteration.
 * Let P be set of paths that can execute
 * on the first iteration, and let Q be the set of those that cannot.
 * For each path q in Q, find the minimum path_reach to q from any p in P
 * This number plus 1 is the first iteration on which q can execute.
 * 6/16/99
 */
void first_turn( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr, *path_ptr2;
  int reach, min_distance;
  extern bool verbose;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (path_ptr->min == 2)
    {
      min_distance = MAXINT;

      for (path_ptr2 = node_ptr->loop_paths->path;
           path_ptr2;
           path_ptr2 = path_ptr2->next)
      {
        reach = node_ptr->path_reach[path_ptr2->number][path_ptr->number];

        if (path_ptr2->min == 1 &&
            reach != 0 &&
            reach < min_distance)
          min_distance = reach;
      }
      if (reach == MAXINT)
        printc ("Warning:  path %d can't be reached by initial path\n");
      else
      {
        path_ptr->min = 1 + min_distance;
        path_ptr->max += min_distance;
	if (verbose)
	  printc ("resetting path %d min to %d and max to %\n",
		  path_ptr->number, path_ptr->min, path_ptr->max);
      }
    }
  }
}

/*
 * find_paths_range - Given a path_ptr, return the corresponding range pointer
 */
struct range_node *find_paths_range( struct tnode* node_ptr, struct loop_path_node* path_ptr )
{
  static struct range_node *range_ptr;

  /* Added by Sibin M on Sept 30, 2004.
   * Added the following sanity checks to avoid seg faults.
   */
  
  if( ( !node_ptr->range_list ) && ( !path_ptr->range_index ) )
	  return 0 ;

  /* End of addition by Sibin M on Sept 30, 2004.
   * Added the following sanity checks to avoid seg faults.
   */
  
  for (range_ptr = node_ptr->range_list;
       range_ptr;
       range_ptr = range_ptr->next)
    if (range_ptr->number == path_ptr->range_index)
      break;

  if (! range_ptr) {
	  assert(0) ;
    // error ("find_paths_range():  cannot find range corresponding to path");
  }
  return range_ptr;
}

/*
 * check_exit - If any exit path can execute on first iter, I have to allow
 * all of them to.  Otherwise it is not safe.  Let's say we
 * have 3 paths, 2 of them can't execute on 1st iteration, but
 * the third is very long.  Then it would take less time to
 * go 2 iterations than 1 iteration!  We will have a conservative
 * underestimation to handle this safely.
 * If no exit path can execute on 1st iteration, then we need to
 * update min_iterations.  Called directly from tbc.  6/16/99
 */
void check_exit( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;
  bool any_on_first = false;
  int min = MAXINT;
  extern bool verbose;

  if (verbose)
    printc ("entering check_exit()\n");

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (path_ptr->path_type_mask != EXIT)
      continue;

    if (path_ptr->min < min)
      min = path_ptr->min;

    if (path_ptr->min == 1)
    {
      any_on_first = true;
      break;
    }
  }

  if (any_on_first)
  {
    if (verbose)
      printc ("I have to allow all exit paths to execute on first iter.\n");

    for (path_ptr = node_ptr->loop_paths->path;
         path_ptr;
         path_ptr = path_ptr->next)
    {
      if (path_ptr->path_type_mask != EXIT)
        continue;

      path_ptr->min -= path_ptr->min - 1;
      path_ptr->max -= path_ptr->min - 1;		/* bring max down too */
    }
  }
  else if (min != MAXINT)
  {
    node_ptr->loop_paths->min_iterations = min;
    if (verbose)
      printc ("need to increase the min iterations of the loop to %d\n",
	      min);
  }
}

/*
 * find_compares - Look for comparing an induction variable to a constant
 *                 so that we can see if a certain path is taken on
 *                 certain iterations, and count how many. (started 24 oct 98)
 */
void find_compares( struct inf_func_element* inf, struct tnode* node_ptr, enum time_case tcase )
{
  struct loop_path_node *path_ptr, *path_ptr2;
  struct loop_block_list *bptr;
  struct block_element *inf_bptr;
  int abs_max_iterations, k, m, n, succ_min, succ_max;
  struct instruct *inf_inst_ptr;

  extern bool verbose;

  if (verbose) {
    printc ("entering find_compares\n\n");
    print_path_types (node_ptr);
  }

  /*
   * To encompass best case analysis, it may be better to
   * use min_iters instead of abs_max_iters.  6/16/99
   * in order to detect that an exit path may not execute on
   * 1st iteration in best case.
   * But watch out below in assigning a backedge path its max number:
   * max has to start out at least 1.
   */
  if (tcase == WORST)
    abs_max_iterations = node_ptr->loop_paths->abs_max_iterations;
  else
    abs_max_iterations = node_ptr->loop_paths->min_iterations;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (verbose)
      printc ("path %d\n", path_ptr->number);

    /*
     * initialize range of iterations to be [1..abs max]
     */
    path_ptr->min = 1;
    path_ptr->max = abs_max_iterations;

    /*
     * To encompass best case analysis, it may be better to
     * use min_iters instead of abs_max_iters.  6/16/99
     * in order to detect that an exit path may not execute on
     * 1st iteration in best case
     */
    if (path_ptr->path_type_mask == EXIT)
    {
      path_ptr->min = abs_max_iterations;
      path_ptr->max = abs_max_iterations;
    }
    else if (path_ptr->path_type_mask == BACKEDGE)
    {
      path_ptr->min = 1;
      path_ptr->max = larger (abs_max_iterations - 1, 1);
    }

    for (bptr = path_ptr->path_list; bptr; bptr = bptr->next)
    {
      if (verbose)
	printc ("  block %d\n", bptr->block_number);

      inf_bptr = Locate_Inf_Block (inf, bptr->block_number);

      if (inf_bptr->left_min == 0 && inf_bptr->dir == 'X')
        continue;		/* this block has no left/right info */

      /*
       * If this block has no instructions, then it represents a child
       * loop, so don't apply any of its value constraints.  11/26/98
       */
      inf_inst_ptr = inf_bptr->instruct_list;
      if (! Locate_Stats_Inst (node_ptr->instruct_list, inf_inst_ptr->number))
      {
	if (verbose)
	  printc ("  this block is a child loop:  don't apply its constraints\n");
        continue;
      }

      /*
       * We have a fall thru situation if:
       *   the next block number is 1 more than the current block number, or
       *   this is the last block, and the loop header is 1 more 
       */
      if (bptr->next && bptr->next->block_number == bptr->block_number + 1 ||
          ! bptr->next &&
               path_ptr->path_list->block_number == bptr->block_number + 1)
      {
        if (inf_bptr->dir == 'R')
	  {
	    if (verbose)
	      printc ("  fall thru situation, encountered a branch taken only"
		      " at iter %d\n", inf_bptr->dir_iter);

          /*
           * If this false condition that we are avoiding on this path takes
           * place at either end of our range of iterations, then trim that
           * number from our range.  11/26/98
           */
          if (inf_bptr->dir_iter == path_ptr->min)
            ++path_ptr->min;
          else if (inf_bptr->dir_iter == path_ptr->max)
            --path_ptr->max;

          continue;			/* our situation is not a branch */
        }

        /*
         * first check for a "iters_once" arising from an equality
         * branch in this block.
         * We fall thru on exactly one iteration.  Intersect this
         * with the current path iteration information.
         */
        else if (inf_bptr->dir == 'L')
        {
          /*
           * Performing this intersection may yield an empty range of
           * iterations, which we will detect later in this function.
           */
          if (inf_bptr->dir_iter > 0)
          {
            if (inf_bptr->dir_iter > path_ptr->min)
              path_ptr->min = inf_bptr->dir_iter;
            if (inf_bptr->dir_iter < path_ptr->max)
              path_ptr->max = inf_bptr->dir_iter;
          }
          else
          {
            /*
             * Need to detect that there is a match between the inner
             * and outer loop control variables.  If there is, we say
             * that this path executes once, but don't change the range
             * of iterations because we don't know which iteration will take.
             * If the variables will not match up, the path will not execute.
             */
            path_ptr->min_iter = 1;
            path_ptr->max_iter = 1;
          }
        }

        /*
         * fall thru, so look at the left numbers for this block
         */
        else					/* intersect */
        {
          if (inf_bptr->left_min > path_ptr->min)
            path_ptr->min = inf_bptr->left_min;
          /*
           * special case:  on last iteration, we should not care
           * about what the branch is doing at the end of the path
           * because we are going to exit the loop shortly, either
           * by branching out or falling out.
           */
          if (! bptr->next && path_ptr->max == abs_max_iterations)
            ;
          else if (inf_bptr->left_max < path_ptr->max)
            path_ptr->max = inf_bptr->left_max;
        }
      }
      else
      {
        if (inf_bptr->dir == 'L')
        {
	  if (verbose)
	    printc ("branch situation, encountering branch that falls thru only"
		    " at iter %d\n", inf_bptr->dir_iter);

          /*
           * If this false condition that we are avoiding on this path takes
           * place at either end of our range of iterations, then trim that
           * number from our range.  11/26/98
           */
          if (inf_bptr->dir_iter == path_ptr->min)
          {
	    if (verbose)
	      printc ("incrementing path_ptr->min from %d to %d\n",
		      path_ptr->min, path_ptr->min + 1);
            ++path_ptr->min;
          }
          else if (inf_bptr->dir_iter == path_ptr->max)
          {
	    if (verbose)
	      printc ("decrementing path_ptr->max from %d to %d\n",
		      path_ptr->max, path_ptr->max - 1);
            --path_ptr->max;
          }

          continue;			/* our situation is not a fall thru */
        }
        /*
         * first check for a "iters_once" arising from an equality
         * branch in this block.
         * We branch on exactly one iteration.  Intersect this
         * with the current path iteration information.
         */
        else if (inf_bptr->dir == 'R')
        {
          /*
           * Performing this intersection may yield an empty range of
           * iterations, which we will detect later in this function.
           */
          if (inf_bptr->dir_iter > 0)
          {
            if (inf_bptr->dir_iter > path_ptr->min)
              path_ptr->min = inf_bptr->dir_iter;
            if (inf_bptr->dir_iter < path_ptr->max)
              path_ptr->max = inf_bptr->dir_iter;
          }
          else
          {
            path_ptr->min_iter = 1;
            path_ptr->max_iter = 1;
          }
        }
        /*
         * branches, so look at the right numbers for this block
         */
        else					/* intersect */
        {
          if (inf_bptr->right_min > path_ptr->min)
            path_ptr->min = inf_bptr->right_min;
          /*
           * special case:  on last iteration, we should not care
           * about what the branch is doing at the end of the path
           * because we are going to exit the loop shortly, either
           * by branching out or falling out.
           */
          if (! bptr->next && path_ptr->max == abs_max_iterations)
            ;
          else if (inf_bptr->right_max < path_ptr->max)
            path_ptr->max = inf_bptr->right_max;
        }
      }
      if (verbose)
	printc ("  range so far is [%d..%d]\n", path_ptr->min, path_ptr->max);
    }
  }

  /*
   * Now, count how many iterations we have reserved for specific paths.
   * Go thru blocks again to see if we can subtract 1 for each equality
   * branch that was feasible but did not go the unique way.
   */

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {

    /*
     * If this is an "exit" path and we can tell that it cannot execute
     * on the last iteration, then make it a "none" path so the loop
     * analysis will not consider it.  Similarly, if there is a "continue"
     * path that cannot execute until the last iteration.
     * 6/16/99:  I assume this won't apply for best case, because we
     * may be changing the min iterations.
     */

    if(tcase == WORST &&
      (path_ptr->path_type_mask == EXIT && path_ptr->max < abs_max_iterations ||
       path_ptr->path_type_mask==BACKEDGE && path_ptr->min>=abs_max_iterations))
    {
      path_ptr->path_type_mask = NONE;
      if (verbose)
	printc ("Path %d type is being changed to none.\n", path_ptr->number);
    }

    if (path_ptr->max > abs_max_iterations ||
        path_ptr->min < 1 ||
        path_ptr->min > path_ptr->max)
    {
      path_ptr->min_iter = 0;			/* sounds infeasible */
      path_ptr->max_iter = 0;
      path_ptr->path_type_mask = NONE;

      if (verbose)
	printc ("Path %d's iteration range [%d..%d] is outside loop's"
		" iteration range [1..%d].\n", path_ptr->number,
		path_ptr->min, path_ptr->max, abs_max_iterations);
    }

    path_ptr->take_away = 0;

    for (bptr = path_ptr->path_list; bptr; bptr = bptr->next)
    {
      inf_bptr = Locate_Inf_Block (inf, bptr->block_number);

      if (inf_bptr->left_min == 0 && inf_bptr->dir == 'X')
        continue;               /* this block has no left/right info */

      /*
       * Take away an iteration from the min # if we skip an iteration
       * within our range.  The comparison to -1 deals with comparing
       * the loop control variable to a value external:  we don't know
       * exactly which iteration.  12/2/98
       * Note:  if dir_iter for the just once directive is -1, then should
       * not be subtracting the take_away from the max iters.
       * The -1 indicates we don't know the iteration it will occur in and we
       * also don't know if it will occur at all.  But it would occur at
       * most once.  12/13/98
       */

      if (bptr->next && bptr->next->block_number == bptr->block_number + 1 ||
          bptr->next == NULL &&
               path_ptr->path_list->block_number == bptr->block_number + 1)
      {
        if (inf_bptr->dir == 'R' &&
            (inf_bptr->dir_iter <= 0 ||
             inf_bptr->dir_iter >= path_ptr->min &&
             inf_bptr->dir_iter <= path_ptr->max))
        {
          ++path_ptr->take_away;
          path_ptr->already_took_away = true;	/* don't do twice 2/17/99 */
          if (inf_bptr->dir_iter >= 0)
            ++path_ptr->max_take_away;
        }
      }
      else
      {
        if (inf_bptr->dir == 'L' &&
            (inf_bptr->dir_iter <= 0 ||
             inf_bptr->dir_iter >= path_ptr->min &&
             inf_bptr->dir_iter <= path_ptr->max))
        {
          ++path_ptr->take_away;
          if (inf_bptr->dir_iter >= 0)
            ++path_ptr->max_take_away;
        }
      }
    }

    /*
     * take away the number of iterations from equality situations
     * we avoid on this path -- this looks good  Nov. 12, 1998
     * But we only do this if we haven't calculated max_iter and min_iter
     * already as a result of "iters_once -1" earlier.  MAXINT was
     * the default value when the path was created.
     */
    if (path_ptr->max_iter == MAXINT)
    {
      if (path_ptr->path_type_mask == NONE)
        path_ptr->max_iter = 0;
      else
        /*
         * special case -- if max==min, keep max_iter 1  (1/15/99)
         */
        if (path_ptr->max == path_ptr->min)
          path_ptr->max_iter = 1;
        else
          path_ptr->max_iter = path_ptr->max - path_ptr->min + 1
                               - path_ptr->max_take_away;
    }
  }

  /*
   * Now let's make use of information regarding how long it takes for
   * a path to get back to itself, if ever.  12/3/98
   */
  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    /*
     * If a path can never reach itself, its max iter is at most 1.
     */
    if (node_ptr->path_reach[path_ptr->number][path_ptr->number] == 0)
    {
      if (path_ptr->max_iter > 1)
        path_ptr->max_iter = 1;
      if (path_ptr->min_iter > 1)
        path_ptr->min_iter = 1;

      /*
       * We determine the range of min and max iters for this path
       * to be 1 less than the union of where this path immediately leads.
       * For example, if there are 2 paths we can immediately reach whose
       * iter ranges are [50..90] and [75..100] then we have to constrain
       * the range of this path to within [49.99] which is [minsucc..maxsucc].
       */
      succ_min = 0;
      succ_max = 0;

      for (path_ptr2 = node_ptr->loop_paths->path;
           path_ptr2;
           path_ptr2 = path_ptr2->next)
      {
        if (node_ptr->path_reach[path_ptr->number][path_ptr2->number] == 1)
        {
          if (succ_min == 0)			/* initialize */
          {
            succ_min = path_ptr2->min;
            succ_max = path_ptr2->max;
          }
          else
          {
            if (path_ptr2->min < succ_min)
              succ_min = path_ptr2->min;
            if (path_ptr2->max > succ_max)
              succ_max = path_ptr2->max;
          }
        }
      }
      if (succ_min != 0)			/* if we found a range */
      {
        if (succ_min - 1 > path_ptr->min)
        {
          path_ptr->min = succ_min - 1;
	  if (verbose)
	    printc ("changing path %d's min to %d\n", path_ptr->number,
		    succ_min - 1);
        }
        if (succ_max - 1 < path_ptr->max)
        {
          path_ptr->max = succ_max - 1;
	  if (verbose)
	    printc ("changing path %d's max to %d\n", path_ptr->number,
		    succ_max - 1);
        }
      }
    }
    /*
     * If this path's own path reach is greater than 1, then we use
     * the formula m = ceil (n/k) to find the maximum iterations.
     * Here, m is the total number of iterations possible for the
     * path, rather than for the entire loop.
     */
    if (node_ptr->path_reach[path_ptr->number][path_ptr->number] > 1)
    {
      k = node_ptr->path_reach[path_ptr->number][path_ptr->number];
      n = path_ptr->max - path_ptr->min + 1;
      m = (int) ceil (((double) n) / ((double) k));
      if (verbose)
	printc ("path %d:  ceil (%d/%d) = %d max iters\n", path_ptr->number,
		n, k, m);

      if (path_ptr->max_iter > m)
        path_ptr->max_iter = m;
      if (path_ptr->min_iter > m)
        path_ptr->min_iter = m;
    }
  }

  find_unique_iters(node_ptr);

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (path_ptr->unique_min == 0)
      path_ptr->min_iter = 0;

    else if (path_ptr->path_type_mask != NONE)
    {
      path_ptr->min_iter = path_ptr->unique_max - path_ptr->unique_min + 1
                           - path_ptr->take_away;
      if (path_ptr->min_iter < 0)
        path_ptr->min_iter = 0;			/* can't be negative */
    }
  }

  if (node_ptr->num_paths <= 31)
    compute_ranges(node_ptr);

  /*
   * if we alternate paths for an odd number of iterations, or some similar
   * situation, we need to reduce one or more of the the path's max iters.
   * In other words, we want the max iters of all the paths of a particular
   * range to sum to that range.  For example, max iters of 50 and 51 for
   * 101 iterations and not 51 and 51.
   * 2/2/99
   */
  preheader_check(inf, node_ptr);

  /*
   * Also, recalculate path's min if it can't go on first iter.  6/16/99
   */
  first_turn(node_ptr);

  if (verbose) {
    print_range_list(node_ptr);
    print_path_iters(node_ptr);
    print_path_types (node_ptr);
    printc ("returning from find_compares\n");
  }
}

/*
 * common_fm - To avoid an underestimation in worst case when we have
 * multiple paths sharing a common first miss, detect if the following
 * conditions are met  (1) and either (2) or (3) or (4):
 * 1.  a first miss (program line) common to more than one path
 * 2.  the paths have a different # of max or min iterations
 * 3.    or the paths are in different iteration range classes
 * 4.    or these paths cannot execute in every iter of their poss iter range
 * started 2/8/99
 * program line is  inst num / words per line
 * Can't just return true/false.  Need to know which program line and
 * which paths we are talking about.
 * So I probably need to have int* parameters for the the above info.
 * But what if there is more than one instance of a common fm?  Maybe
 * we need to call this function during twc and not at the beginning.
 */
bool common_fm( struct inf_func_element* inf, struct tnode* node_ptr, 
			struct loop_path_node* path_ptr, struct instruct* inf_inst_ptr )
{
  struct loop_path_node *path_ptr2;
  struct block_element *inf_bptr2;
  struct loop_block_list *bptr2;
  struct instruct *inf_inst_ptr2;
  struct inst_cat *stats_inst_ptr2;
  extern int words_per_line;
  int prog_line1, prog_line2;
  bool count_on_side = false;
  extern bool verbose;

  if (verbose)
    printc ("entering common_fm()\n");

  /*
   * look at the other paths to see if there is a match
   */
  for (path_ptr2 = node_ptr->loop_paths->path;
       path_ptr2;
       path_ptr2 = path_ptr2->next)
  {
    if (path_ptr2 == path_ptr)
      continue;

    for (bptr2 = path_ptr2->path_list; bptr2; bptr2 = bptr2->next)
    {
      inf_bptr2 = Locate_Inf_Block (inf, bptr2 -> block_number);

      for (inf_inst_ptr2 = inf_bptr2 -> instruct_list;
           inf_inst_ptr2;
           inf_inst_ptr2 = inf_inst_ptr2 -> next)
      {
        stats_inst_ptr2 = Locate_Stats_Inst (node_ptr->instruct_list,
                                             inf_inst_ptr2->number);
        if (stats_inst_ptr2 == NULL)
          break;

        prog_line1 = inf_inst_ptr->number / words_per_line;
        prog_line2 = inf_inst_ptr2->number / words_per_line;

        /*
         * Condition (1) is that they are both 'f' and are in
         * the same program line.
         */
        if (stats_inst_ptr2->cat_list->worst_case == 'f' &&
            prog_line1 == prog_line2)
        {
	  if (verbose)
	    printc ("path %d line %d inst %d, path %d line %d inst %d\n",
		    path_ptr->number, prog_line1, inf_inst_ptr->number,
		    path_ptr2->number, prog_line2, inf_inst_ptr2->number);
	  
          /*
           * condition (2) is that the paths have diff max_iter.
           */
          if (path_ptr->max_iter != path_ptr2->max_iter)
          {
            count_on_side = true;
	    if (verbose)
	      printc ("  max iters differ\n");
          }

          /*
           * condition (3) is that paths are in diff range classes.
           */
          if (path_ptr->range_index != path_ptr2->range_index)
          {
            count_on_side = true;
	    if (verbose)
	      printc ("  paths are not in same range class\n");
          }


          /*
           * condition (4) is that these paths cannot execute
           * in every iter of their possible iteration range
           */
          if (path_ptr->max_iter < path_ptr->max - path_ptr->min + 1 &&
              path_ptr2->max_iter < path_ptr2->max - path_ptr2->min + 1)
          {
            count_on_side = true;
	    if (verbose)
	      printc ("  paths can't execute in every iter of their"
		      " possible iteration range\n");
          }
        }
      }
    }
  }

  if (verbose)
    printc ("returning %s from common_fm()\n", count_on_side ? "true" : "false");

  return count_on_side;
}

/*
 * For best case analysis, we detect if there are any paths that are infeasible
 * as exit paths because they can only make the exiting transition on an
 * iteration number we will never get to, based on the value of min_iterations.
 * 4/29/99
 *
 * For each path in the node
 *   If it is not an exit path, ignore it and go on to the next.
 *   Look at the last block in the path.
 *   Since this can be an exit path, there is a transition after the last block
 *     to a block outside the loop.
 *   Find out which successor is the one outside the loop.
 *     in other words, not in this loop or any of its children.
 *   If the successor is numbered  last_block + 1 then
 *     we exit the loop by falling out of it
 *     look at the block's "left" range
 *   Else
 *     we exit the loop by jumping
 *     look at the block's "right" range
 *   If min_iterations < range's min then
 *     this path can't serve as an exit path.
 */
void infeasible_exit( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;
  struct loop_block_list *block_ptr;
  struct block_element *inf_bptr;
  struct block_list *succ_ptr;
  struct tnode *child_node;
  bool found_in_child;
  int range_min;
  extern bool verbose;

  if (verbose) {
    printc ("entering infeasible_exit()\n");
    printc ("min_iterations = %d\n", node_ptr->loop_paths->min_iterations);
  }

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (verbose)
      printc ("Looking at path %d\n", path_ptr->number);

    if ((path_ptr->path_type_mask & EXIT) == 0)
      continue;

    for (block_ptr = path_ptr->path_list;
         block_ptr;
         block_ptr = block_ptr->next)
    {
      if (block_ptr->next)
        continue;
      
      if (verbose)
	printc ("  looking at the last block: %d\n", block_ptr->block_number);

      inf_bptr = Locate_Inf_Block (inf, block_ptr->block_number);

      for (succ_ptr = inf_bptr->succ;
           succ_ptr;
           succ_ptr = succ_ptr->next)
      {
	if (verbose)
	  printc ("    successor block is %d\n", succ_ptr->block_number);

        /*
         * It is possible that the successor is in a child loop.
         * in such a case, that is not really an exit from the loop.
         */
        if (is_in_loop (succ_ptr->block_number, node_ptr))
          continue;
        found_in_child = false;
        for (child_node = node_ptr->child;
             child_node;
             child_node = child_node->next)
          if (is_in_loop (succ_ptr->block_number, child_node))
          {
            found_in_child = true;
            break;
          }
        if (found_in_child)
          continue;

        if (succ_ptr->block_number == 1 + block_ptr->block_number)
        {
	  if (verbose) {
	    printc ("      We exit the loop by falling out of it.\n");
	    printc ("      left min = %d\n", inf_bptr->left_min);
	  }
          range_min = inf_bptr->left_min;
        }
        else
        {
	  if (verbose) {
	    printc ("      We exit the loop by jumping.\n");
	    printc ("      right min = %d\n", inf_bptr->right_min);
	  }
          range_min = inf_bptr->right_min;
        }

        if (range_min > 0 && node_ptr->loop_paths->min_iterations < range_min)
        {
	  if (verbose) {
	    printc ("      Path %d is infeasible as exit.\n", path_ptr->number);
	  }
          path_ptr->path_type_mask &= ~ EXIT;

          if (! inf_bptr->infeasible_succ)
          {
            inf_bptr->infeasible_succ=Create_Block_List(succ_ptr->block_number);
	    if (verbose) 
	      printc ("      just created infeasible succ list for block %d"
		      " containing %d\n", inf_bptr->block_number,
		      succ_ptr->block_number);
          }
          else
          {
            Append_Block_Blocklist (succ_ptr->block_number,
                                    &inf_bptr->infeasible_succ);
	    if (verbose)
	      printc ("      just added %d to block %d's infeasible succ list\n",
		      succ_ptr->block_number, inf_bptr->block_number);
          }
        }
      }
    }
  }

  if (verbose)
    printc ("returning from infeasible_exit()\n");
}

/*
 * For each path in the node, we need to find out if there is an infeasible
 * transition out of a child loop into the rest of this path.  4/30/99
 * This function was needed to fix the 9-cycle underestimation in symb.
 * For each block B (except the first) in a path:
 *   feasible = false
 *   for each predecessor P of B
 *     if B is not marked as infeasible successor (done in infeasible_exit)
 *       feasible = true
 *     This is because in order for the path to be infeasible, EVERY
 *       predecessor P has to consider B infeasible.
 *   if feasible == false
 *     mark this path as infeasible (set its path_type_mask to NONE).
 */
void infeasible_transition( struct inf_func_element* inf, struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;
  struct loop_block_list *block_ptr;
  struct block_element *inf_bptr, *inf_bptr2;
  bool feasible;
  struct block_list *pred_ptr;
  extern bool verbose;

  if (verbose)
    printc ("entering infeasible_transition\n");

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    if (verbose)
      printc ("Path %d\n", path_ptr->number);

    /*
     * This analysis is only relevant if there is more than one block.
     */
    if (path_ptr->path_list && path_ptr->path_list->next)
    {
      for (block_ptr = path_ptr->path_list->next;
           block_ptr;
           block_ptr = block_ptr->next)
      {
	if (verbose)
	  printc ("  block %d\n", block_ptr->block_number);

        feasible = false;

        inf_bptr = Locate_Inf_Block (inf, block_ptr->block_number);

        for (pred_ptr = inf_bptr->pred;
             pred_ptr;
             pred_ptr = pred_ptr->next)
        {
          inf_bptr2 = Locate_Inf_Block (inf, pred_ptr->block_number);

          if (! Is_In_Block_List (inf_bptr->block_number,
                                  inf_bptr2->infeasible_succ))
          {
	    if (verbose)
	      printc ("    %d is not in %d's infeasible succ list\n",
		      inf_bptr->block_number, pred_ptr->block_number);

            feasible = true;
          }
        }

        if (! feasible)
        {
	  if (verbose)
	    printc ("path looks infeasible after looking at all the "
		    " predecessors of block %d\n", inf_bptr->block_number);

          path_ptr->path_type_mask = NONE;
        }
      }
    }
  }
  if (verbose)
    printc ("returning from infeasible_transition\n");
}

/*
 * Infeasible continue:  If a loop has N iterations, and a continue
 * path cannot execute on any of the first N-1 iterations, then
 * it cannot serve as a continue path.  5/21/99
 * In other words, if the path's min_iter >= N.
 * However, we don't want to change the path_type_mask, so I've
 * made this boolean.
 * If the min_iter >= N, then we check the path type mask.  If
 * the path is only a continue path, it is infeasible.
 */
bool infeasible_continue( struct tnode* node_ptr, struct loop_path_node* path_ptr, enum time_case tcase )
{
  int iterations;
  extern bool verbose;

  if (tcase == WORST)
    iterations = node_ptr->loop_paths->abs_max_iterations;
  else
    iterations = node_ptr->loop_paths->min_iterations;

  if (path_ptr->min >= iterations &&
      path_ptr->path_type_mask == BACKEDGE)
  {
    if (verbose)
      printc ("Path %d can't continue.\n", path_ptr->number);
    return true;
  }
  return false;
}

/*
 * print_path_types - Print the path type mask for each path.
 */
void print_path_types( struct tnode* node_ptr )
{
  struct loop_path_node *path_ptr;

  for (path_ptr = node_ptr->loop_paths->path;
       path_ptr;
       path_ptr = path_ptr->next)
  {
    printc ("path %d is ", path_ptr->number);
    if (path_ptr->path_type_mask == NONE)
      printc ("none");
    if (path_ptr->path_type_mask & BACKEDGE)
      printc ("continue ");
    if (path_ptr->path_type_mask & EXIT)
      printc ("exit");
    printc ("\n");
  }
}
