/*****************************************************************************
*
*   FILE NAME       :   INPUT.C
*   PROGRAMMER(S)   :   Robert Arnold
*   DATE CREATED    :   June 30, 1993
*
******************************************************************************
*
*   MODIFICATION HISTORY:
*
*   Date       Change
*   --------   ----------------------------------------------------------------
*   07/03      Added "include time.h"
*   02/10/94   Added code to calculate the instruction number. <rda 02/10/94>
*   02/16/94   Added code to calculate instruction offset from beginning of
*              function. <rda 02/16/94>
******************************************************************************
*/

#include "instdef.h"
#include "timeds.h"
#include "inf.h"       /* Define constants/tokens used by VPO in the INF file */
#ifndef INST_H
#include "inst.h"
#endif

/* Added by Sibin M on Oct 25, 2004. 
 * Process of cleaning up the TA for C++
 */

#include "input.h"
#include "timedecl.h"
#include "lib.h"
#include "instdecl.h"
#include "isa.h"
#include "treedecl.h"
#include "mtadeldecl.h"

#include <ctype.h>
#include <cstdlib>
#include <unistd.h>

/* Added by Sibin M on Oct 25, 2004. 
 * Process of cleaning up the TA for C++
 */

// #include <malloc.h>
#include <cstdio>
#include <cstring>

#ifdef DISPLAY
#define printc printf
#else
#define printc  
#endif

// #define __SIB_DEBUG__

#ifdef __SIB_DEBUG__
extern bool sib_verbose ;
#endif // __SIB_DEBUG__

extern int memory_count;		/* bytes of dynamic allocation */

/******************************************************************************
*
*   FUNCTION NAME : Process_INF_File
*
*   ASSUMPTIONS:
*
*     1. All functions will be preceeded by the value FUNC_IDENTIFIER
*
*     2. The conclusion of each function will be indicated by the 
*        value LIST_TERMINATOR
*
*     3. There will be no extra integers following the last function declaration
*******************************************************************************/
void Process_INF_File( struct inf_func_element** inf, string inputfile, struct inst_block* inst_set )
{
  FILE *inf_fp;

  int number = 0;
  int file_status = 0;     /* Used to determine if EOF has been read in */
  int func_number = 0;     /* Number assigned to function when read in */

  string  inf_filename;     /* filename used to open up the inf file */
  string  func_name;        /* string to read in function name from inf file */

  int     inst_number;   /* Inst offset from beginning of func <rda 02/10/94>*/
  string  token;         /* Char/string token scanned from  inf file */

  bool cont = true;     /* Used to break out of outermost loop when */
                                              /* EOF condition is true */

  struct inf_func_element *head_func_ptr;	/* initialize shortly */
  struct inf_func_element *curr_func_ptr = NULL;
  struct inf_func_element *func_ptr = NULL;
  extern bool  verbose;

  head_func_ptr = *inf;		/* new method for mult files (7/13) */

  for (curr_func_ptr = head_func_ptr;
       curr_func_ptr && curr_func_ptr->next;
       curr_func_ptr = curr_func_ptr->next)
    ;

  strcpy (inf_filename, inputfile);
  strcat (inf_filename, File_EXT);

  inf_fp = fopen (inf_filename, "r");      /* Open inf file for read access */
  if (inf_fp == NULL)
  {
    printf("Unable to find %s.inf file\n",inf_filename);
    error ("Unable to find .inf file");
  }

  while (number != EOF && cont == true)
  {
                                /* Skip all numbers till see FUNC_IDENTIFIER*/
                                /* which indicates the start of a function */

    while ( fscanf(inf_fp,"%d\n", &number)  && 
            number != FUNC_IDENTIFIER )
      ;

    fscanf(inf_fp, "%s\n",func_name);       /* Scan In The Function Name  */

    func_number++;

    if (verbose) {
      printc ("Reading function %s from .inf file\n", func_name);
    }

    func_ptr = Create_Inf_Func_Element (func_name, func_number, inf_filename);

    file_status = fscanf(inf_fp, "%s", token);

    while ( file_status  != EOF  &&
            strcmp(Func_IDENTIFIER, token) != 0 )
    {
                        /* NOTE: The INF file is being scanned from this*/
                        /*       function AND the function to read in the*/
                        /*       list of instructions for each block.    */
                        /*       Since the list of instructions does not */
                        /*       contain a seperate "end of list" symbol,*/
                        /*       the control flow is more complex.       */


                        /* At this point the parsing is inside a function*/

      inst_number = 0;     /* Initial inst # for each func <rda 02/10/94> */

      while ( (file_status != EOF)  &&
               (strcmp(Func_IDENTIFIER,token) != 0) )
	{
		/* This while loop will check for end of current function */

                  /* Should be reading in a loop struct def, */
                  /* a block struct def, or a block instruction*/
                  /* STRUCT_DEF indicates reading in a loop or */
                  /* block definition.*/

        if (strcmp(STRUCT_DEF, token) == 0)
        {
                                        /* Process a loop or block definition */
          fscanf(inf_fp, "%s", token);

          if (strcmp("time", token) == 0)	/* reading time structure def */
          {
            Read_In_Time_Structure (inf_fp, func_ptr);
            fscanf(inf_fp, "%s", token);
          }

          if (strcmp(LOOP, token) == 0)
          {                       /* Reading in a loop structure definition */

            Read_In_Loop_Structure (inf_fp, func_ptr);
            fscanf(inf_fp, "%s", token);
                                  /* Read In The next token from the INF File */
          }
          else if (strcmp(BLOCK, token) == 0)                     
          {                      /* Reading in a block structure definition */

            Read_In_Block_Structure (inf_fp, func_ptr, &inst_number,
                                     token,&file_status,inst_set);
                                         /* Read in the structure information */
          }
        }
      }
    }

                /* At this point should have completely read in a function.  */


    if (file_status == EOF)
      cont = false;
      		/* If EOF has been read in, then break out of this function */

           /* At this point insert func_ptr into the linked list of functions */

    if (head_func_ptr)
            /* Assuming common case that more than 1 function will be read in*/ 
    {
      curr_func_ptr->next = func_ptr;
      curr_func_ptr = curr_func_ptr->next;
    }
    else
    {
      head_func_ptr = func_ptr;
      curr_func_ptr = func_ptr;
    }
  }

  *inf = head_func_ptr;
}


// FILE *inf_fp;                                          /* Ptr to the INF file */
// struct inf_func_element *inf;  /*Ptr to struct holding func info being read in*/
void Read_In_Time_Structure( FILE * inf_fp, struct inf_func_element* inf )
{
  extern int mhz;
  int number, lineno, lower_bound, upper_bound, found;
  struct timing_label_element  *head_label, *temp_label;
  struct line_list_element *line_list_temp;
  char c;

  /*
   * read in the constraint number and line number
   */
  fscanf (inf_fp, "%d %d ", &number, &lineno);

  /*
   * read in lower bound thru the space that precedes the upper bound
   */
  fscanf (inf_fp, "%d%c", &lower_bound, &c);
  switch (c)
  {
    case 's' : lower_bound = lower_bound * mhz * 1000000;
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'm' : lower_bound = lower_bound * mhz * 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'u' : lower_bound = lower_bound * mhz;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'n' : lower_bound = lower_bound * mhz / 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;
  }

  /*
   * read in upper bound
   */
  fscanf (inf_fp, "%d%c", &upper_bound, &c);
  switch (c)
  {
    case 's' : upper_bound = upper_bound * mhz * 1000000;
               break;

    case 'm' : upper_bound = upper_bound * mhz * 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;

    case 'u' : upper_bound = upper_bound * mhz;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;

    case 'n' : upper_bound = upper_bound * mhz / 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;
  }


  found = 0;
  for (head_label = inf->timing_label;
       head_label;
       head_label = head_label->next)
  {
    if (head_label->number == number)
    {
      found = 1;
      break;
    }
  }
  if (! found)
  {
    memory_count += sizeof (struct timing_label_element);
    temp_label = (struct timing_label_element *) 
                 malloc (sizeof (struct timing_label_element));
    temp_label->number = number;
    temp_label->earliest_line = lineno;

    temp_label->lower_bound = lower_bound;
    temp_label->upper_bound = upper_bound;

    temp_label->line_list = NULL;
    memory_count += sizeof (struct line_list_element);
    line_list_temp = (struct line_list_element *)
                     malloc (sizeof (struct line_list_element));
    line_list_temp->lineno = lineno;
    line_list_temp->next = temp_label->line_list;
    temp_label->line_list = line_list_temp;


    temp_label->next = inf->timing_label;
    inf->timing_label = temp_label;
  }
  else					/* already in list of labels */
  {
    if (head_label->lower_bound == LIST_TERMINATOR &&
        lower_bound != LIST_TERMINATOR)
    {
      head_label->lower_bound = lower_bound;	/* initialize lower bound */
    }

    if (head_label->lower_bound != LIST_TERMINATOR &&
        lower_bound != LIST_TERMINATOR &&
        lineno <= head_label->earliest_line)
    {
      head_label->earliest_line = lineno;
      head_label->lower_bound = lower_bound;	/* reset lower bound */
    }

    if (head_label->upper_bound == LIST_TERMINATOR &&
        upper_bound != LIST_TERMINATOR)
    {
      head_label->upper_bound = upper_bound;	/* initialize upper bound */
    }

    if (head_label->upper_bound != LIST_TERMINATOR &&
        upper_bound != LIST_TERMINATOR &&
        lineno <= head_label->earliest_line)
    {
      head_label->earliest_line = lineno;
      head_label->upper_bound = upper_bound;	/* reset upper bound */
    }

    /* now let's append new lineno to label's list of line numbers */

    memory_count += sizeof (struct line_list_element);
    line_list_temp = (struct line_list_element *)
                     malloc (sizeof (struct line_list_element));
    line_list_temp->lineno = lineno;
    line_list_temp->next = head_label->line_list;
    head_label->line_list = line_list_temp;
  }
}


/*******************************************************************************
*
*   FUNCTION NAME : Read_In_Loop_Structure
*
*   ASSUMPTIONS:
*
*      1. All lists of blocks in the inf file will be terminated with the
*         LIST_TERMINATOR value
*
*      2. Any loop iteration that can not be determined by the compiler 
*         will be set to the value of the LIST_TERMINATOR.  This is assumed 
*         so that the code can prompt the user for some specified value.
*
*      3. Assumption will be that the inf pointer will be pointing to memory
*         that has already been malloc'ed.  This pointer can either be the head
*         of the list (for the function main), or the "next" pointer inside
*         the inf function for all other subsequent functions.
*
*      4. The loop header block will be the first block read in from the list
*         of blocks contained in the loop.
*******************************************************************************
*   MODIFICATION HISTORY:
*
*   Date       Change
*   --------   ----------------------------------------------------------------
*   02/21/94   Added code to set the header field to the loop header block.
*******************************************************************************/
// FILE *inf_fp;                                          /* Ptr to the INF file */
// struct inf_func_element *inf;  /*Ptr to struct holding func info being read in*/
void Read_In_Loop_Structure( FILE * inf_fp, struct inf_func_element* inf )
{
  char   c, *s;
  char   identifier[80], command[200], number_str[11];
  extern int mhz;
  extern bool traditional;
  int    number;                   /* Loop number assigned by vpo compiler */
  int    nest_level;               /* Loop nesting level */
  int    min_iterations;           /* Minimum loop iterations */
  int    max_iterations;           /* Maximum loop iterations */
  int    lower_bound;              /* user's expected time bounds */
  int    upper_bound;
  int    i, min, max, mode, inner_begin_offset, inner_limit_value,
         inner_increment, outer_number,
         relop, iter_mode, dummy,
         abs_min_iterations, abs_max_iterations,
         avg_min_iterations, avg_max_iterations,
         induction_init, induction_limit, induction_incr,
         total_iters, induction_init_offset = 0,
         bc_total_iters, wc_total_iters,
         inner_initial_value, inner_limit_offset,
         min_initial, max_initial, min_limit, max_limit;

  bool inner_constant_limit, need_to_average = false,
               early_exit = false, found_early_exit = false;

  struct loop_element *loop_ptr, *p, *parent;
                                     /* Ptr to Loop struct to be added
                                                        to inf struct */
  struct loop_element *curr_ptr;    /* Ptr to last loop struct in linked list*/

	/* taking out the old summation solver, 
  extern char *eval();
	and adding the new summation solver
	*/
  extern bool  verbose;


  /*
   * Inf loop info includes, inter alia:
   *   number                 Loop number assigned by vpo
   *   nest_level             Nesting level within function
   *   min_iterations         Min iter loop will execute
   *   max_iterations         Max iter loop will execute
   */

  fscanf (inf_fp, "%d %d ", &number, &nest_level);
  if (verbose)
    printc ("\nExamining loop %d at nesting level %d\n", number, nest_level);

  fscanf (inf_fp, "%d ", &min_iterations);

  if (min_iterations == INDUCTION_VAR_KNOWN)
  {
    /*
     * If the compiler knows the induction variable of the loop,
     * we will get five pieces of information:
     *   register RTL, initial value RTL, limit RTL, constant increment,
     *   and then the letter 's' or 'e' for Single or Early exit
     *   Early exit means that in best case, only one iteration
     *   this will impact the expression sent to ctadel for inner loop.
     * Either of the initial value or the limit may be a reg +/- constant.
     * If we don't recognize the format, we won't handle this information.
     *   In other words, if limit is sum of 2 reg's.  10/13/98
     */

    /*
     * throw away induction register
     */
    fscanf (inf_fp, "%c%c%d%c ", &c, &c, &dummy, &c);
    fscanf (inf_fp, "%c", &c);
    if (c != 'r')
    {
      ungetc (c, inf_fp);
      fscanf (inf_fp, "%d ", &induction_init);
      if (verbose)
	printc ("induction_init = %d\n", induction_init);
    }
    else
    {
      /*
       * throw away induction init register
       * the register number   r[number]   may be followed by +/- constant
       */
      fscanf (inf_fp, "%c%d%c%c", &c, &dummy, &c, &c);

      if (c != ' ')
      {
        fscanf (inf_fp, "%d ", &induction_init_offset);
        if (c == '-')
          induction_init_offset *= -1;
      }
      if (verbose)
	printc ("induction_init_offset = %d\n", induction_init_offset);
    }

    if (fscanf (inf_fp, "%d ", &induction_limit) == 1)
    {
      if (verbose)
	printc ( "induction_limit = %d\n", induction_limit);
    }
    else
    {
      /*
       * throw away induction limit register
       */
      fscanf (inf_fp, "%c%c%d%c ", &c, &c, &dummy, &c);
    }

    fscanf (inf_fp, "%d %c %d ", &induction_incr, &c, &min_iterations);

    if (c == 'e')
    {
      if (verbose)
	printc ("setting early_exit to true\n");
      early_exit = true;
    }
  }

  if (min_iterations == ASK_USER)
  {
    min_iterations = 0;
    max_iterations = 0;
    mode = 1;						/* positive sign */

    do
    {
      for (i = 0; i < 80; ++i)
      {
        fscanf (inf_fp, "%c", &c);
        if (c == ' ' || c == '+' || c == '-')
          break;
        identifier[i] = c;
      }
      if (! isdigit (identifier[0]))
      {
        printf ("What is the minimum value of %s in loop %d? \n",
                identifier, number);
        scanf ("%d", &min);

        printf ("What is the maximum value of %s in loop %d? \n",
                identifier, number);
        scanf ("%d", &max);

      }
      else
      {
        /*
         * in the case of a constant, the max and min are this constant
         */
        max = min = atoi (identifier);
      }

      min_iterations += min * mode;
      max_iterations += max * mode;

      for (i = 0; i < 80; ++i)
        identifier[i] = '\0';				/* erase identifier */

      if (c == '-')
        mode = -1;					/* negative sign */
    } while (c != ' ');

    do
    {
      fscanf (inf_fp, "%c", &c);			/* skip max iter */
    } while (c != ' ');

    abs_min_iterations = min_iterations;
    abs_max_iterations = max_iterations;

  }
  else if (min_iterations == COMPUTE_ITERATIONS ||
           min_iterations == MULTIPLE_EXIT)
  {
    /*
     * After seeing a -3 in the loop line, we expect to see:
     *
     * 1.  iter_mode:  an averaging mode number -- 0 means use averaging method,
     *     1 means don't use averaging in b.c. since loop could exit in 1 iter
     *
     *     Next, The initial value and limit are constant and RTL or vice versa
     *     so the next two numbers are  inner_begin_offset, inner_limit_value
     *                              or  inner_initial_value, inner_limit_offset.
     *
     * 2.  an RTL like r[10]+1 indicating the initial value for this loop
     *     in terms of the control variable of the outer loop (like r[10])
     * 3.  limit for the loop (this also could be RTL or constant)
     * 4.  relop:  a number indicating which type of comparison is being made
     * 5.  inner_increment:  increment for the loop
     * 6.  outer_number:  loop number of the outer loop
     *                    upon which the inner depends
     * 7.  outer_initial_value or outer_depend_number:
     *     initial value for control value of outer loop
     *     c followed by a constant, or
     *     r followed by a register and a loop number assoc with that register
     * 8.  outer_limit_value or outer_depend_number
     *     limit of control value of outer loop
     *     c followed by a constant, or
     *     r followed by a register and a loop number assoc with that registero
     * 9.  outer_increment:  increment of control value of outer loop
     */

     /*
      * We make the following assumptions:
      * 1.  Constant stride
      * 2.  initial or the limit must be constant, or they must
      *     both differ by a constant, as in V+c1 and V+c2.
      * 3.  form of nonconstant must be   variable +/- constant
      * 4.  lower bound is always <= upper bound
      *     Otherwise, don't call Ctadel, but say num iters = 1.
      */

    need_to_average = true;	/* save the averaging for later 4/10/99 */

    /*
     * iter_mode is 1 if b.c. could exit in 1 iter.  0 means use avg method
     */

    fscanf (inf_fp, "%d ", &iter_mode);
    if (verbose)
      printc ("iter mode = %d\n", iter_mode);

    /*
     * The second component could be a constant or begin with a register.
     * first read the rtl expression, for example "r[10]+1"
     *   there may be a constant to add or subtract from the register
     *   to indicate the initial value of the inner (this) loop
     *   control variable.  r[10] is the loop ctrl variable for outer loop
     */
    fscanf (inf_fp, "%c", &c);
    if (c == 'r')
    {
      if (verbose)
	printc ("the initial value for this inner loop contains a register\n");

      do
      {
        fscanf (inf_fp, "%c", &c);
      } while (c != ']');

      /*
       * need to prepare for syntax like r[16] without any + or - after it
       */

      fscanf (inf_fp, "%c", &c);		/* the char after the ']' */

      if (c == ' ')
      {
        inner_begin_offset = 0;		/* only 7 numbers to read, not 8 */
      }

      else
      {
        if (c == '-')
          mode = -1;
        else if (c == '+')
          mode = 1;		/* is inner_begin_offset pos or neg? */
        else
          error ("Read_In_Loop_Structure(): unknown char after register\n");


        fscanf (inf_fp, "%d ", &inner_begin_offset);

        inner_begin_offset *= mode;
      }
      if (verbose)
	printc ("inner begin offset = %d\n", inner_begin_offset);
    }
    else			/* keep reading digits of this constant */
    {
      ungetc (c, inf_fp);
      fscanf (inf_fp, "%d ", &inner_initial_value);
      if (verbose)
	printc ("inner initial value = %d\n", inner_initial_value);
    }

    /*
     * The third component could be a constant or begin with a register.
     */
    fscanf (inf_fp, "%c", &c);
    if (c == 'r')
    {
      inner_constant_limit = false;
      if (verbose)
	printc ("the limit for this inner loop contains a register\n");

      do
      {
        fscanf (inf_fp, "%c", &c);
      } while (c != ']');

      /*
       * need to prepare for syntax like r[16] without any + or - after it
       */

      fscanf (inf_fp, "%c", &c);		/* the char after the ']' */

      if (c == ' ')
      {
        inner_limit_offset = 0;		/* only 7 numbers to read, not 8 */
      }

      else
      {
        if (c == '-')
          mode = -1;
        else if (c == '+')
          mode = 1;		/* is inner_begin_offset pos or neg? */
        else
          error ("Read_In_Loop_Structure(): unknown char after register\n");


        fscanf (inf_fp, "%d ", &inner_limit_offset);

        inner_limit_offset *= mode;
      }
      if (verbose)
	printc ("inner limit offset = %d\n", inner_limit_offset);
    }
    else			/* keep reading digits of this constant */
    {
      inner_constant_limit = true;
      ungetc (c, inf_fp);
      fscanf (inf_fp, "%d ", &inner_limit_value);
      if (verbose)
	printc ("inner limit value = %d\n", inner_limit_value);
    }

    /*
     * Note, need to relax the restriction about either (2) or (3) being
     * constant.  4/13/99
     */



    fscanf (inf_fp, "%d ", &relop);

    if (fscanf (inf_fp, "%d ", &inner_increment) != 1)
    {

      printf ("Loop %d needs a constant increment.  What should it be? ",
              number);
      scanf ("%d", &inner_increment);
    }

    fscanf (inf_fp, "%d ", &outer_number);
    if (verbose)
      printc ("relop = %d,  inner increment = %d,  outer number = %d\n",
	      relop, inner_increment, outer_number);

    /*
     * Adjust the limit value or offset based on the relational operator.
     * 4/4/99
     */
    if (relop == REL_LT || relop == REL_NE)
    {
      if (inner_constant_limit)
      {
        --inner_limit_value;
	if (verbose)
	  printc ("inner limit value adjusted to %d\n", inner_limit_value);
      }
      else
      {
        --inner_limit_offset;
	if (verbose)
	  printc ("inner limit offset adjusted to %d\n", inner_limit_offset);
      }
    }

    /*
     * Next we read about the control value of the outer loop.  3/31/99
     */
    fscanf (inf_fp, "%c ", &c);
    if (c == 'c')
    {
      fscanf (inf_fp, "%d ", &dummy);	   /* throw away outer initial value */
    }
    else
    {
      do
      {
        fscanf (inf_fp, "%c", &c);
      } while (c != ' ');

      fscanf (inf_fp, "%d ", &dummy);	   /* throw away outer depend number */
    }

    fscanf (inf_fp, "%c ", &c);
    if (c == 'c')
    {
      fscanf (inf_fp, "%d ", &dummy);	     /* throw away outer limit value */
    }
    else
      /*
       * warning:  the first thing to scan now is an rtl expression,
       * not simply a register.  For example, we could see r[19]+2.
       * Skip over the rtl expression.  Same with the above.  4/2/99
       */
    {
      do
      {
        fscanf (inf_fp, "%c", &c);
      } while (c != ' ');
 
      fscanf (inf_fp, "%d ", &dummy);	   /* throw away outer depend number */
    }

    fscanf (inf_fp, "%d ", &dummy);		/* throw away outer increment */



  }
  else
  {
    need_to_average = false;
    fscanf (inf_fp, "%d ", &max_iterations);

    abs_min_iterations = min_iterations;
    abs_max_iterations = max_iterations;
  }

  /*
   * read in lower bound thru the space that precedes the upper bound
   */
  fscanf (inf_fp, "%d%c", &lower_bound, &c);
  switch (c)
  {
    case 's' : lower_bound = lower_bound * mhz * 1000000;
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'm' : lower_bound = lower_bound * mhz * 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'u' : lower_bound = lower_bound * mhz;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;

    case 'n' : lower_bound = lower_bound * mhz / 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               fscanf (inf_fp, "%c", &c);		/* throw away ' ' */
               break;
  }

  /*
   * read in upper bound
   */
  fscanf (inf_fp, "%d%c", &upper_bound, &c);
  switch (c)
  {
    case 's' : upper_bound = upper_bound * mhz * 1000000;
               break;

    case 'm' : upper_bound = upper_bound * mhz * 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;

    case 'u' : upper_bound = upper_bound * mhz;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;

    case 'n' : upper_bound = upper_bound * mhz / 1000;
               fscanf (inf_fp, "%c", &c);		/* throw away 's' */
               break;
  }


                      /* If either min_iterations or max_iterations     */
                      /* is equal to LIST_TERMINATOR, then the          */
                      /* compiler was unable to determine the minium    */
                      /* or maximum number of iterations, so prompt the */
                      /* user for this information.                     */

  if (min_iterations == LIST_TERMINATOR)
  {
     printf ("WARNING: FUNCTION %s Contains Loop At Nest Level", inf->func);
     printf (" %d That Has An Undetermined Minimum Number of Iterations\n",
             nest_level);
     printf ("         Enter Minimum Number Of Iterations-> ");
     scanf  ("%d", &min_iterations);			/* ch took out the \n */
  }

  if (max_iterations == LIST_TERMINATOR)
  {
     printf ("WARNING: FUNCTION %s Contains Loop At Nest Level", inf->func);
     printf (" %d That Has An Undetermined Maximum Number of Iterations\n",
             nest_level);
     printf ("         Enter Maximum Number Of Iterations-> ");
     scanf  ("%d", &max_iterations);			/* ch took out the \n */
  }
/*
  abs_min_iterations = min_iterations;
*/
  abs_max_iterations = max_iterations;

                   /* Allocate memory for loop struct and init ptrs */

                   /* Create a node for the loop element.  Passing */
                   /* value of 0 into loop header block because it */
                   /* is unknown at this point in time.            */

  loop_ptr = Create_Loop_Element(number, nest_level, min_iterations, 
             max_iterations, abs_min_iterations, abs_max_iterations,
             induction_init, induction_limit, lower_bound, upper_bound, 0);

  loop_ptr->induction_incr = induction_incr;
  loop_ptr->early_exit = early_exit;

  if (verbose)
    printc ("setting loop %d early_exit to %s\n", loop_ptr->number,
          early_exit ? "true" : "false");

    /*
     * But also need to compute the ABSOLUTE min and max iterations.
     * 4/10/99
     * abs max iterations = max_limit - min_initial + 1
     * abs min iterations = larger (1, min_limit - max_initial + 1)
     */

    min_initial = find_min_initial (inf, loop_ptr);
    max_initial = find_max_initial (inf, loop_ptr);
    min_limit = find_min_limit (inf, loop_ptr);
    max_limit = find_max_limit (inf, loop_ptr);

    abs_min_iterations = larger (1, min_limit - max_initial + 1);
    abs_max_iterations = max_limit - min_initial + 1;

    loop_ptr->abs_min_iterations = abs_min_iterations;
  /*
   * store the begin_offset, limit and increment in case a further nested
   * loop needs to look it up  3/24/99  someday maybe this should be
   * incorporated in the function being called above
   */
  loop_ptr->constant_limit = inner_constant_limit;
  loop_ptr->begin_value = inner_initial_value;
  loop_ptr->begin_offset = inner_begin_offset;
  loop_ptr->limit_value = inner_limit_value;
  loop_ptr->limit_offset = inner_limit_offset;
  loop_ptr->increment = inner_increment;
  loop_ptr->outer_number = outer_number;


  Read_In_Block_List (inf_fp, &loop_ptr->block_list);
                            /* Read in list of blocks contained in the loop */
                                                                     
  loop_ptr->header = loop_ptr->block_list->block_number;            
          /* Set header value to header block number for loop <rda 02/21/94> */

  Read_In_Block_List (inf_fp, &loop_ptr->exit_list);
               /* Read in list of exit blocks out of the loop */


  loop_ptr->need_to_average = need_to_average;

  if (loop_ptr->need_to_average)
  {
    /*
     * But also need to compute the ABSOLUTE min and max iterations.
     * 4/10/99
     * abs max iterations = max_limit - min_initial + 1
     * abs min iterations = larger (1, min_limit - max_initial + 1)
     */

    min_initial = find_min_initial (inf, loop_ptr);
    max_initial = find_max_initial (inf, loop_ptr);
    min_limit = find_min_limit (inf, loop_ptr);
    max_limit = find_max_limit (inf, loop_ptr);

    abs_min_iterations = larger (1, min_limit - max_initial + 1);
    abs_max_iterations = max_limit - min_initial + 1;

    loop_ptr->abs_min_iterations = abs_min_iterations;

    if (verbose)
      printc ("now we will calculate total # iterations for loop %d\n",
	      loop_ptr->number);

    strcpy (command, "sum(1, ");

    for (p = loop_ptr;   ; p = parent)
    {
      parent = find_parent_loop (inf, p);

      if (p != loop_ptr)		/* print "by" for 2nd and later loops */
        strcat (command, " by ");

      strcat (command, "loop");
      itoa (p->number, number_str);
      strcat (command, number_str);
      strcat (command, " = ");
      if (verbose) {
	printc ("p = loop %d\n", p->number);
      }

      if (! p->need_to_average)
      {
	if (verbose)
	  printc ("no need to average: constant number of iterations\n");

        /*
         * If this loop can exit early, then we use the initial value
         * on both sides of the ..  4/22/99
         */
        if (p->early_exit)
        {
          found_early_exit = true;
        }
	if (verbose)
	  printc ("limit is initial+incr(num_iter-1) = %d + %d(%d-1) \n",
                p->induction_init, p->induction_incr, p->max_iterations);

        itoa (p->induction_init, number_str);
        strcat (command, number_str);
        strcat (command, "..");
        itoa (p->induction_init + p->induction_incr * (p->max_iterations - 1),
              number_str);
        strcat (command, number_str);
      }
      else if (p->constant_limit)
      {
	if (verbose)
	  printc ("inner limit is constant\n");

        strcat (command, "loop");
        itoa (p->outer_number, number_str);
        strcat (command, number_str);
        if (p->begin_offset > 0)
          strcat (command, "+");
        if (p->begin_offset != 0)
        {
          itoa (p->begin_offset, number_str);
          strcat (command, number_str);
        }
        strcat (command, " .. ");
        itoa (p->limit_value, number_str);
        strcat (command, number_str);
      }
      else
      {
	if (verbose)
	  printc ("inner limit is variable\n");

        itoa (p->begin_value, number_str);
        strcat (command, number_str);
        strcat (command, " .. loop");
        itoa (p->outer_number, number_str);
        strcat (command, number_str);
        if (p->limit_offset > 0)
          strcat (command, "+");
        if (p->limit_offset != 0)
        {
          itoa (p->limit_offset, number_str);
          strcat (command, number_str);
        }
      }

      if (! p->need_to_average)
        break;
    }
    strcat (command, ")");
    if (verbose)
      printc ("the command is %s\n", command);
			/* this is out.
    s = eval (command);
    if (verbose)
      printc ("Ctadel says %s\n", s);
    total_iters = atoi (s);
		*/
    total_iters = mtadel(command);
		
    printf( "HERE !!!!\n" ) ;
		 
#ifdef __SIB_DEBUG__
  sib_verbose = true ;
  if( sib_verbose )
  {	  
	  printc( "\n---------------------------------------\n" ) ;
	  printc( "Inside: Read_In_Loop_Structure()\n" ) ;
	  printc( "Input to MTADEL : %s\n", command ) ;
	  printc( "Ouput from MTADEL : %d \n", total_iters ) ;
	  printc( "---------------------------------------\n" ) ;
  }
#endif // __SIB_DEBUG
		
    printf("total_iters (per mtadel()): %d\n", total_iters);

    loop_ptr->total_iters = total_iters;

    parent = find_parent_loop (inf, loop_ptr);

    avg_min_iterations = total_iters / parent->total_iters;
    avg_max_iterations = int_ceil (total_iters, parent->total_iters);
    if (verbose) {
      printc ("surrounding loop's total is %d\n", parent->total_iters);
      printc ("computed avg iterations:  min = %d, max = %d\n",
	      avg_min_iterations, avg_max_iterations);
    }
    min_iterations = avg_min_iterations;
    max_iterations = avg_max_iterations;

    /*
     * In the traditional analysis, the number of iterations was identical
     * to what we now refer to as the "absolute" max/min number of iterations.
     */
    if (traditional)
    {
      min_iterations = abs_min_iterations;
      max_iterations = abs_max_iterations;
    }

    if (loop_ptr->need_to_average)
      abs_min_iterations = larger(1, min_limit - max_initial + 1);

    /*
     * if outer loop can possibly iterate just once
     */

    if (found_early_exit)
    {
      if (verbose) 
	printc ("now calculate a special total iteration count for bc\n");
      strcpy (command, "sum(1, ");
      for (p = loop_ptr;   ; p = parent)
      {
        parent = find_parent_loop (inf, p);
        if (p != loop_ptr)              /* print "by" for 2nd and later loops */
        strcat (command, " by ");

        strcat (command, "loop");
        itoa (p->number, number_str);
        strcat (command, number_str);
        strcat (command, " = ");

	if (verbose) {
	  printc ("p = loop %d\n", p->number);
	}
        if (! p->need_to_average)
        {
	  if (verbose)
	    printc ("no need to average: constant number of iterations\n");

          if (p->early_exit)
          {
	    if (verbose) {
	      printc ("this loop has an early exit\n");
	    }

            if (p->min_iterations != 1)
            {
              printf ("  this loop is insane, min iterations is not 1\n");
              exit(1);
            }
            itoa (p->induction_init, number_str);
            strcat (command, number_str);
            strcat (command, "..");
            strcat (command, number_str);
          }
        }
        else if (p->constant_limit)
        {
	  if (verbose)
	    printc ("inner limit is constant\n");

          strcat (command, "loop");
          itoa (p->outer_number, number_str);
          strcat (command, number_str);
          if (p->begin_offset > 0)
            strcat (command, "+");
          if (p->begin_offset != 0)
          {
            itoa (p->begin_offset, number_str);
            strcat (command, number_str);
          }
          strcat (command, " .. ");
          itoa (p->limit_value, number_str);
          strcat (command, number_str);
        }
        else
        {
	  if (verbose)
	    printc ("inner limit is variable\n");

          itoa (p->begin_value, number_str);
          strcat (command, number_str);
          strcat (command, " .. loop");
          itoa (p->outer_number, number_str);
          strcat (command, number_str);
          if (p->limit_offset > 0)
            strcat (command, "+");
          if (p->limit_offset != 0)
          {
            itoa (p->limit_offset, number_str);
            strcat (command, number_str);
          }
        }

        if (! p->need_to_average)
          break;
      }
      strcat (command, ")");
      if (verbose)
	printc ("the special bc command is %s\n", command);
      /*s = eval (command);
      if (verbose)
	printc ("Ctadel says %s\n", s);

      total_iters = atoi (s);
			*/
			total_iters = mtadel(command);
			printf("total_iters (per mtadel()): %d\n", total_iters);
      abs_min_iterations = avg_min_iterations = min_iterations = total_iters;
    }

    if (iter_mode == 1)		/* in best case, could exit early */
    {
      min_iterations = 1;
      abs_min_iterations = 1;
    }
    if (verbose)
      printc ("Abs min iter = %d, Abs max iter = %d\n",
            abs_min_iterations, abs_max_iterations);
    if (verbose)
      printc ("Avg min iter = %d, Avg max iter = %d\n",
            avg_min_iterations, avg_max_iterations);

    loop_ptr->min_iterations = min_iterations;
    loop_ptr->max_iterations = max_iterations;
    loop_ptr->abs_min_iterations = abs_min_iterations;
    loop_ptr->abs_max_iterations = abs_max_iterations;
  }

  /*
   * Reconsile the total number of iterations for loops that do not
   * need to be averaged.  Simply based on the next outer loop.  4/13/99
   */
  else
  {
    parent = find_parent_loop (inf, loop_ptr);
    if (parent)
      loop_ptr->total_iters = parent->total_iters * max_iterations;
    else
      loop_ptr->total_iters = 1;
  }


  if (inf->loop)    /* Insert new loop elemnt struct into linked list*/
  {
    curr_ptr = inf->loop;            /* Move curr_ptr to the end of the l.l.*/
    while (curr_ptr->next)
      curr_ptr = curr_ptr->next;

    curr_ptr->next = loop_ptr;    /* Insert loop_ptr at the end of the l.l. */
  }
  else                             /* Insert loop_ptr as head of the l.l. */
    inf->loop = loop_ptr;

  fscanf (inf_fp, "%c", &c);		/* throw away newline at end of line */
  if (c != '\n')
    ungetc (c, inf_fp);
}


int find_min_initial( struct inf_func_element* inf, struct loop_element* loop )
{
  extern bool  verbose;
  int min_initial = 0;
  struct loop_element *p;

  for (p = loop; p; p = find_parent_loop(inf, p))
  {
    if (! p->need_to_average)				/* both are constant */
    {
      if (verbose)
	printc ("add the induction init value %d from loop %d\n",
              p->induction_init, p->number);
      min_initial += p->induction_init;
      break;
    }
    else if (! p->constant_limit)			/* initial is const */
    {
      if (verbose)
	printc ("add the begin value %d from loop %d\n",
              p->begin_value, p->number);

      min_initial += p->begin_value;
      break;
    }
    else
    {
      if (verbose)
	printc ("add the begin offset %d from loop %d\n",
              p->begin_offset, p->number);

      min_initial += p->begin_offset;
    }
  }
  if (verbose)
    printc ("returning %d from find_min_initial()\n", min_initial);

  return min_initial;
}

int find_max_initial( struct inf_func_element* inf, struct loop_element* loop )
{
  int max_initial = 0;
  struct loop_element *p;
  extern bool verbose;

  /*
   * need to have a preliminary step that looks at the initial value
   * of the innermost loop, while for each outer loop we look at the limits
   */
  if (! loop->need_to_average)
    max_initial = loop->induction_init;
  else if (! loop->constant_limit)
    max_initial = loop->begin_value;
  else
  {
    max_initial = loop->begin_offset;

    for (p = find_parent_loop(inf,loop); p; p = find_parent_loop(inf, p))
    {
      if (! p->need_to_average)				/* both are constant */
      {
	if (verbose) {
	  printc ("add the induction limit value %d from loop %d\n",
		  p->induction_limit, p->number);
	}

        max_initial += p->induction_limit;
        break;
      }
      else if (! p->constant_limit)			/* initial is const */
      {
	if (verbose) {
	  printc ("add the limit offset %d from loop %d\n",
		  p->limit_offset, p->number);
	}

        max_initial += p->limit_offset;
        break;
      }
      else
      {
	if (verbose) {
	  printc ("add the limit value %d from loop %d\n",
		  p->limit_value, p->number);
	}

        max_initial += p->limit_value;
      }
    }
  }
  if (verbose)
    printc ("returning %d from find_max_initial()\n", max_initial);

  return max_initial;
}

int find_min_limit( struct inf_func_element* inf, struct loop_element* loop )
{
  int min_limit = 0;
  struct loop_element *p;
  extern bool verbose;

  /*
   * need to have a preliminary step that looks at the limit value
   * of the innermost loop, while for each outer loop we look at the initial
   */
  if (! loop->need_to_average)
    min_limit = loop->induction_limit;
  else if (loop->constant_limit)
    min_limit = loop->limit_value;
  else
  {
    min_limit = loop->limit_offset;

    for (p = find_parent_loop(inf,loop); p; p = find_parent_loop(inf, p))
    {
      if (! p->need_to_average)				/* both are constant */
      {
	if (verbose)
	  printc ("add the induction init value %d from loop %d\n",
		  p->induction_init, p->number);

        min_limit += p->induction_init;
        break;
      }
      else if (! p->constant_limit)			/* initial is const */
      {
	if (verbose)
	  printc ("add the begin value %d from loop %d\n",
		  p->begin_value, p->number);

        min_limit += p->begin_value;
        break;
      }
      else
      {
	if (verbose)
	  printc ("add the begin offset %d from loop %d\n",
		  p->begin_offset, p->number);

        min_limit += p->begin_offset;
      }
    }
  }
  if (verbose)
    printc ("returning %d from find_min_limit()\n", min_limit);

  return min_limit;
}

int find_max_limit( struct inf_func_element* inf, struct loop_element* loop )
{
  int max_limit = 0;
  struct loop_element *p;
  extern bool verbose;

  for (p = loop; p; p = find_parent_loop(inf, p))
  {
    if (! p->need_to_average)				/* both are constant */
    {
      if (verbose)
	printc ("add the induction limit value %d from loop %d\n",
		p->induction_limit, p->number);

      max_limit += p->induction_limit;
      break;
    }
    else if (p->constant_limit)				/* limit is const */
    {
      if (verbose)
	printc ("add the limit value %d from loop %d\n",
		p->limit_value, p->number);

      max_limit += p->limit_value;
      break;
    }
    else
    {
      if (verbose)
	printc ("add the limit offset %d from loop %d\n",
		p->limit_offset, p->number);

      max_limit += p->limit_offset;
    }
  }
  if (verbose)
    printc ("returning %d from find_max_limit()\n", max_limit);

  return max_limit;
}


struct loop_element *find_parent_loop( struct inf_func_element* inf, struct loop_element* loop )
{
  static struct loop_element *parent;
  struct loop_element *p;
  extern bool verbose;

  for (p = inf->loop; p; p = p->next)
  {
    if (p->nest_level == loop->nest_level - 1 &&
        Is_In_Block_List (loop->header, p->block_list))
    {
      parent = p;
      if (verbose)
	printc ("the parent of loop %d is loop %d\n", loop->number, p->number);
      break;
    }
  }

  if (! parent)
  {
    if (verbose)
      printc ("I could not find parent loop.\n");
    parent = NULL;
  }

  return parent;
}


/*
 * int_ceil - an integer-oriented ceiling function.  I can't seem to
 * get the real one to work for the purpose I'm using above.
 * Note:  it may be more efficient to say  (num-1) / denom + 1.
 */
int int_ceil( int num, int denom )
{
  if (num % denom == 0)
    return num / denom;
  else
    return num / denom + 1;
}

/******************************************************************************
*
*   FUNCTION NAME : Read_In_Block_Structure
*
*   ASSUMPTIONS:
*
*      1. All lists of blocks in the inf file will be terminated with the
*         LIST_TERMINATOR value
*
*      2. Assumption will be that the inf pointer will be pointing to memory
*         that has already
*         been malloc'ed.  This pointer can either be the head of the list
*         (for the function main), or the "next" pointer inside the inf
*         function for all other subsequent functions.
*
*      3. The block structure will have 3 lists of blocks in this order:
*            1. preds  - list of block numbers that preceed the current block
*            2. succs  - list of block numbers that succeed the current block
*            3. doms   - list of block numbers that dominate the current block
*
*         with all block lists starting with their name as the string token
*         ('preds', 'succs', 'doms') and all lists being terminated by
*         the LIST_TERMINATOR.
******************************************************************************/
// FILE              *inf_fp;                          /* Ptr to the INF file */
// struct inf_func_element *inf;  /*Ptr to struct holding func info being read in*/
// int               *inst_number;
// string            token;
// int               *file_status;
// struct inst_block *inst_set;      /* Ptr to list of all inst for this machine */

void Read_In_Block_Structure( FILE* inf_fp, struct inf_func_element* inf, int* inst_number, 
			      string token, int* file_status, struct inst_block* inst_set)
{
  int   block_number;              /* Block number assigned by vpo compiler */
  int   start_line;                /* Source code start line for this block */
  int   stop_line;                 /* Source code stop line for this block */

  struct block_element *block_ptr;   /* Ptr to Block struct to be added
                                        to inf struct */
  struct block_element *curr_ptr;    /* Ptr to last loop struct in l.l. */

  extern bool verbose;

  fscanf (inf_fp, "%d lines %d-%d ",
          &block_number,      /* Scan in the block number assigned by VPO */
          &start_line,        /* Scan src code line number block starts */
          &stop_line);        /* Scan src code line number block ends */

  block_ptr = Create_Block_Element(block_number);


                      /* Determine if this block's number is outside */
                      /* the function's current range of block numbers.*/
                      /* IF so, update the appropriate variable.*/

  if (block_ptr->block_number < inf->min_block_number)
    inf->min_block_number = block_ptr->block_number;
  else if (block_ptr->block_number > inf->max_block_number)
    inf->max_block_number = block_ptr->block_number;

                      /* Next token scanned in should be "preds" */
                      /* If this is not the token scanned in, then */
                      /* print an error message and exit the program */

  fscanf (inf_fp, "%s", token);
  if (strcmp(PREDS, token) != 0)
  {
    printf ("ERROR: MISSING PREDICATE LIST FOR FUNCTION %s\n", inf->func);
    exit(1);
  }
  Read_In_Block_List (inf_fp, &block_ptr->pred);

                     /* Next token scanned in should be "succs" */
                     /* If this is not the token scanned in, then */
                     /* print an error message and exit the program */

  fscanf (inf_fp, "%s", token);
  if (strcmp(SUCCS, token) != 0)
  {
    printf ("ERROR: MISSING SUCCESSOR LIST FOR FUNCTION %s\n", inf->func);
    exit(1);
  }
  Read_In_Block_List (inf_fp, &block_ptr->succ);

                     /* Next token scanned in should be "doms" */
                     /* If this is not the token scanned in, then */
                     /* print an error message and exit the program */

  fscanf (inf_fp, "%s", token);


   /* I added code to read in the block information that may affect
    * conditional branches in the same function.  6/29/98
    */

  if (strcmp("makes_unknown", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->unknown);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("makes_fallthru", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->fallthru);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("makes_branch", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->branch);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("makes_unknown_if_not_fellthru", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->unfellthru);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("makes_unknown_if_not_branched", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->unbranched);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("fallthru_causes_branch", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->ft_cause_b);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("fallthru_causes_fallthru", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->ft_cause_ft);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("branch_causes_branch", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->b_cause_b);
    fscanf (inf_fp, "%s", token);
  }
  if (strcmp("branch_causes_fallthru", token) == 0)
  {
    Read_In_Block_List (inf_fp, &block_ptr->b_cause_ft);
    fscanf (inf_fp, "%s", token);
  }

  if (strcmp("iters_range", token) == 0)
  {
    /*
     * The ranges were all initalized to [0..0] when block was created.
     * A range of [MAXINT..-1] means not feasible.
     */
    fscanf (inf_fp, " [%d..%d] [%d..%d]\n", &block_ptr->left_min,
            &block_ptr->left_max, &block_ptr->right_min, &block_ptr->right_max);

    if (verbose)
      printc ("block %d:  left [%d..%d]     right [%d..%d]\n",
	      block_number, block_ptr->left_min, block_ptr->left_max,
	      block_ptr->right_min, block_ptr->right_max);

    fscanf (inf_fp, "%s", token);
  }
  else if (strcmp("iters_once", token) == 0)
  {
    /*
     * Something will happen just once for an equality-branch.
     * Direction L means fall thru.  R means taken branch.
     * The dir_iter says which iteration this will take place.
     * A dir_iter value of -1 means that the branch is comparing
     * against a loop-invariant nonconstant value (as in  i != j).
     */
    fscanf (inf_fp, " %c %d\n", &block_ptr->dir, &block_ptr->dir_iter);

    if (verbose)
      printc ("block %d:  %s on iteration %d\n", block_number,
	      block_ptr->dir == 'L' ? "fall thru" : "taken",
	      block_ptr->dir_iter);

    fscanf (inf_fp, "%s", token);
  }

  if (strcmp(DOMS, token) != 0)
  {
    if (verbose)
      printf ("MISSING DOMINATOR LIST: BLOCK %d, FUNCTION %s\n", block_number,
            inf->func);
    exit(1);
  }
  Read_In_Block_List (inf_fp, &block_ptr->dom);

  Read_In_Block_Instruct (inf_fp, &block_ptr->instruct_list, inst_number,
                          token, file_status,inst_set);

  if (inf->block)
  {
    curr_ptr = inf->block;          /* Move curr_ptr to the end of the l.l. */
    while (curr_ptr->next)
      curr_ptr = curr_ptr->next;

    curr_ptr->next = block_ptr;     /* Insert block_ptr at end of the l.l. */
  }
  else                               /* Insert block_ptr as head of the l.l. */
   inf->block = block_ptr;
}


/******************************************************************************
*
*   FUNCTION NAME : Read_In_Block_List
*
*   ASSUMPTIONS:
*
*      1. All lists of blocks in the inf file will be terminated with
*         the LIST_TERMINATOR value
*
*      2. For the sake of efficiency, the assumption is that the common case
*         a block list will contain more than 1 block.
*         This affects the IF statement structure.
*
*      3. The attempt has been made to code this procedure generically
*         so that many functions have the ability to use it.
******************************************************************************/
void Read_In_Block_List( FILE* inf_fp, struct block_list ** block_ptr)
{
  int token;                 /* used read in each block number from inf file */
  struct block_list *head_ptr;   /* ptr to the head of the list of blocks */
  struct block_list *curr_ptr;   /* ptr to the last block on the list     */
  struct block_list *b_ptr;      /* ptr each new block placed on list   */

  head_ptr = NULL;
  curr_ptr = NULL;

  fscanf (inf_fp, "%d", &token);      /* Read in a block # from the inf file */
  while (token != LIST_TERMINATOR)
  {                                   /* Block # Has Been Read In */
                                /* Allocate memory for the block list element */
    b_ptr = Create_Block_List(token);


    /* Insert Block Onto the block_ptr list */
    /* Assuming common case to be more than 1 block to be placed in a list*/

    if (head_ptr)
    {
      curr_ptr->next = b_ptr;       /* Attach b_ptr node to the block list */
      curr_ptr = curr_ptr->next;    /* Move curr_ptr to point to
                                                  last node in list */
    }
    else                       /* Block Will be the first block on the list */
    {
      head_ptr = b_ptr;             /* Set block_ptr to point to this node */
      curr_ptr = head_ptr;          /* Set curr_ptr to point to this node */
    }

    fscanf (inf_fp, "%d", &token);   /* Scan in the next token */
  }


  *block_ptr = head_ptr;   
}


/******************************************************************************
*
*   FUNCTION NAME : Read_In_Block_Instruct
*
*   ASSUMPTIONS:
*
*      1. The list of instructions for each block will be terminated
*         with the List TERMINATOR.
*
*   HISTORY:
*      02/14/94 - Fixed a scanning bug by setting token to be end of line
*                 when tok is at end of line.  <rda 02/14/94>
*
******************************************************************************/
// FILE *inf_fp;
// struct instruct **inst_list;    /* Ptr to struct holding list of block inst */
// int               *inst_number;
// string            token;
// int               *file_status;
// struct inst_block *inst_set;    /* List of all possible inst for this machine */
void Read_In_Block_Instruct( FILE* inf_fp, struct instruct** inst_list, 
			     int* inst_number, string token,
                             int* file_status, struct inst_block* inst_set)
{
  int    inst_cond_code;
  int    inst_data_type;
  int    inst_type;

  int    op1_addr;
  int    op1_data_type;
  string op1_reg_list;

  int    op2_addr;
  int    op2_data_type;
  string op2_reg_list;

  int    op3_addr;
  int    op3_data_type;
  string op3_reg_list;

  char tok;
  int    i;



  string inst_name;
  string called_func;

  struct instruct *head_ptr = NULL;
  struct instruct *curr_ptr = NULL;
  struct instruct *inst_ptr = NULL;

  static int prog_inst_number = 0;
  static int PC = 0;

  strcpy (op1_reg_list, "");
  strcpy (op2_reg_list, "");
  strcpy (op3_reg_list, "");
  strcpy (called_func, "");

  while ( (*file_status = fscanf(inf_fp, "%s", token)) != EOF  &&
           strcmp(Func_IDENTIFIER, token) != 0                &&
	  strcmp(STRUCT_DEF, token) != 0                       )
    {
      strcpy (op1_reg_list, "");
    strcpy (op2_reg_list, "");
    strcpy (op3_reg_list, "");
    strcpy (called_func, "");

    fscanf(inf_fp, " %d %d %d %d %d ", &inst_type,
                                       &inst_data_type,
                                       &inst_cond_code,
                                       &op1_addr,
                                       &op1_data_type);


    fscanf(inf_fp, "%c", &tok);
    while (tok != REG_LIST_START)
      fscanf(inf_fp, "%c", &tok);

    i = 0;
    do
    {
      fscanf (inf_fp, "%c", &tok);

      if (tok != REG_LIST_TERMINATOR)
      {
        op1_reg_list[i] = tok;
        i++;
      }
    }
    while  (tok != REG_LIST_TERMINATOR && i < MAX_STRING_LENGTH-1);

    op1_reg_list[i] = '\0';

    fscanf (inf_fp, "%d %d ", &op2_addr, &op2_data_type);

    fscanf(inf_fp, "%c", &tok);
    while (tok != REG_LIST_START)
      fscanf(inf_fp, "%c", &tok);
    i = 0;
    do
    {
      fscanf (inf_fp, "%c", &tok);

      if (tok != REG_LIST_TERMINATOR)
      {
        op2_reg_list[i] = tok;
        i++;
      }
    }
    while  (tok != REG_LIST_TERMINATOR && i < MAX_STRING_LENGTH-1);

    op2_reg_list[i] = '\0';

    fscanf (inf_fp, "%d %d ", &op3_addr, &op3_data_type);

    fscanf(inf_fp, "%c", &tok);
    while (tok != REG_LIST_START)
      fscanf(inf_fp, "%c", &tok);

    i = 0;
    do {
      fscanf (inf_fp, "%c", &tok);

      if (tok != REG_LIST_TERMINATOR)
      {
        op3_reg_list[i] = tok;
        i++;
      } 
    }
    while  (tok != REG_LIST_TERMINATOR && i < MAX_STRING_LENGTH-1);
    op3_reg_list[i] = '\0';

    if (is_call(inst_type))
      fscanf (inf_fp, "%s", called_func);


    Determine_Inst_Name(inst_type, inst_set, inst_name);

    inst_ptr = Create_Instruct((*inst_number)++, inst_name, inst_type,
                               inst_data_type, inst_cond_code,
                               op1_data_type, op1_reg_list,
                               op2_data_type, op2_reg_list,
                               op3_data_type, op3_reg_list,
                               called_func);


/*    Assign this inst a number representing its offset from beg of program. */

    inst_ptr->prog_offset = prog_inst_number;
    ++prog_inst_number;

    inst_ptr->PC = PC;
    PC +=4;
    
    if (head_ptr)
    {
      curr_ptr->next = inst_ptr;
      curr_ptr = curr_ptr->next;
    }
    else
    {
      head_ptr = inst_ptr; 
      curr_ptr = inst_ptr;
    }
  }
  *inst_list = head_ptr;
}
