/*****************************************************************************
*
* 
*   FILE NAME       :   INST.C
*   PROGRAMMER(S)   :   Robert Arnold
*   DATE CREATED    :   July 3, 1993
*
*   LAST EDITED : Mon Apr 18 17:12:50 1994 by arnold (Robert Arnold) on grep
*
*
*   DESCRIPTION:
*      This file will contain all functions necessary to reading in the
*      instructions specified for the specific hardware platform
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "time.h"

/* Added by Sibin M on Oct 25, 2004.
 * Process of cleaning up the TA for C++
 */
 
#include "instdecl.h"
#include "isa.h"
#include "timedecl.h"
#include "lib.h"
 
/* Added by Sibin M on Oct 25, 2004.
 * Process of cleaning up the TA for C++
 */

#ifndef INST_H
#include "inst.h"
#endif

#ifdef DISPLAY
#define printc printf
#else
#define printc  
#endif

// #define __SIB_DEBUG__

#ifdef __SIB_DEBUG__
extern bool sib_verbose ;
#endif // __SIB_DEBUG__

int  inst_end_cycles[CARD_INST_SET][NUM_DATASIZES][NUM_STAGES][NUM_CASES];


void print_reg_info( int src_needed[], int  src_stage[], int release_time[] )
{
  int i;
  string reg;

  for (i = 0; i < NUMREGS; ++i)
  {
    if (src_needed[i] != NEVER || release_time[i] != NEVER)
    {
      int2reg (i, reg);
      printc ("%s : ", reg);
      if (src_needed[i] != NEVER)
        printc ("needed by %d  before the %s  ; ", src_needed[i],
                stage_name (src_stage[i]));
      else
        printc ("                               ");
      if (release_time[i] != NEVER)
        printc ("ready at %d", release_time[i]);
      printc ("\n");
    }
  }
}

int largest_entry( int array[], int base, int howmany )
{
  int i, max = array[base];

  for (i = base + 1; i < base + howmany; ++i)
  {
    poss_update (&max, array[i], "");
  }
  return max;
}

int smallest_entry( int array[], int base, int howmany )
{
  int i, min = array[base];

  for (i = base + 1; i < base + howmany; ++i)
  {
    if (array[i] < min && array[i] != 0)
      min = array[i];
  }
  return min;
}

int next_stage( int this_stage, int i_end_cycles[] )
{
  while (1)
  {
    ++this_stage;
    if (this_stage == NUM_STAGES)
      return NO_STAGE;
    if (i_end_cycles[this_stage] == 0)
      continue;
    else return this_stage;
  }
}

int prev_stage( int this_stage, int i_end_cycles[] )
{
  while(1)
  {
    --this_stage;
    if (this_stage == NO_STAGE)
      return NO_STAGE;
    if (i_end_cycles[this_stage] == 0)
      continue;
    else return this_stage;
  }
}

int get_end_use( int stage, int num_path_inst, 
		 int this_inst_end_cycles[], 
		 row *end_use )
{
  int prev = prev_stage (stage, this_inst_end_cycles);
  int next = next_stage (stage, this_inst_end_cycles);
  int ready, j;
#ifdef __SIB_DEBUG__
      if( sib_verbose )
      {
	      printc( " --- \n" ) ;
	      printc( "get_end_use() : stage = %d \t prev = %d \t next = %d \n", stage, prev, next ) ;
	      printc( "num_path_inst = %d \n", num_path_inst ) ;
	      printc( "end_use[%d][prev] = %d \n", num_path_inst, end_use[num_path_inst][prev] ) ;			       
	      printc( " --- \n" ) ;
      }
#endif // __SIB_DEBUG__	    

  /* we have to make sure there IS a next stage */
  if (next != NO_STAGE  &&           
      this_inst_end_cycles[stage] > 0 && this_inst_end_cycles[next] > 0)
  {
    ready = 0;
    for (j = num_path_inst - 1; j >= 0; j--)
    if (end_use[j][next] > ready)
    {
      ready = end_use[j][next];
      break;
    }
    return  larger(ready, end_use[num_path_inst][prev] + /* 1 */
                      this_inst_end_cycles[prev] - this_inst_end_cycles[stage]);
  }
  /* otherwise there may be no next stage for this instruction */

  else if (this_inst_end_cycles[stage] > 0)
  {
    return  end_use[num_path_inst][prev] + /* 1 */
                this_inst_end_cycles[prev] - this_inst_end_cycles[stage];
  }

  /* otherwise the inst has neither stage */

  else return 0;
}



void Determine_Inst_Name( int inst_type, 
			  struct inst_block* inst_list, 
			  string name )
{
  int    i;
  struct inst_block *inst_ptr;

  for (inst_ptr = inst_list;
       inst_ptr && inst_ptr->num != inst_type;
       inst_ptr = inst_ptr->next) ;

  if (inst_ptr)
  {
    for (i = 0; i < 20; i++)
    {
      name[i] = inst_ptr->inst[i];
    }
  }

}


/*********
*
*   FUNCTION NAME : Determine_Inst_Set
*
**********/
struct inst_block * Determine_Inst_Set()
{
  FILE   *inst_fp;
  int  inst_num;
  extern bool only_fetch;
  char   inst_name[20];
  static char   instruct_file[] = "../ConfigFiles/instset.curr" ;
  struct inst_block  *inst_set_list = NULL;
  struct inst_block  *inst = NULL;
  struct inst_block  *curr_inst = NULL;
  FILE *stage_fp[NUM_CASES];
  string name;
  int i, this_inst, j, tcase;
  int  single_stage[NUM_STAGES], double_stage[NUM_STAGES];

  inst_fp = fopen (instruct_file, "r");    /* Open file containing instruction
                                                        list for read access */
  if (! inst_fp)
    error ("I can't find the file instset.curr");

  stage_fp[(int) WORST] = fopen ("../ConfigFiles/worst_stages_curr", "r");
  if (! stage_fp[(int) WORST])
    error ("I can't find the file worst_stages_curr");

  stage_fp[(int) BEST] = fopen ("../ConfigFiles/best_stages_curr", "r");
  if (! stage_fp[(int) BEST])
    error ("I can't find the file best_stages_curr");


  while (fscanf(inst_fp,"%d %s", &inst_num, inst_name) == 2)
    {
    if (inst_num <0) break;
    inst = Create_Inst_Block(inst_num, inst_name);

    if (inst_set_list)
    {
      curr_inst->next = inst;
      curr_inst = curr_inst->next; 
    }
    else
    {
      inst_set_list = inst;
      curr_inst = inst;
    }  
  }

  for (tcase = WORST; tcase <= BEST; tcase++)
  {
    fscanf (stage_fp[tcase], "other");
    for (i = 0; i < NUM_STAGES; ++i) {
      fscanf (stage_fp[tcase], "%d", &single_stage[i]);
    }
    for (i = 0; i < NUM_STAGES; ++i ){
      fscanf (stage_fp[tcase], "%d", &double_stage[i]);
    }
    for (i = 0; i < CARD_INST_SET; i++)
      for (j = 0; j < NUM_STAGES; ++j)
      {
        inst_end_cycles[i][SINGLE_SIZE][j][tcase] = single_stage[j];
        inst_end_cycles[i][DOUBLE_SIZE][j][tcase] = double_stage[j];
      }

/* now read the stage requirements for the flops */

    while (1)
    {
      fscanf (stage_fp[tcase], "%d:%s\n", &this_inst, name);

      if (this_inst < 0)
        break;
      for (i = 0; i < NUM_STAGES; ++i)
      {
        fscanf (stage_fp[tcase], "%d ", &single_stage[i]);
        inst_end_cycles[this_inst][SINGLE_SIZE][i][tcase] = single_stage[i];
      }
      for (i = 0; i < NUM_STAGES; ++i)
      {
        fscanf (stage_fp[tcase], "%d ", &double_stage[i]);
        inst_end_cycles[this_inst][DOUBLE_SIZE][i][tcase] = double_stage[i];
      }

    }   
    fclose (stage_fp[tcase]);

    if (only_fetch)
    {
      for (i = 0; i < CARD_INST_SET; ++i)
      {
        inst_end_cycles[i][SINGLE_SIZE][IF][tcase] = 1;
        inst_end_cycles[i][DOUBLE_SIZE][IF][tcase] = 1;

        for (j = 1; j < NUM_STAGES; ++j)	/* other stages are empty */
        {
          inst_end_cycles[i][SINGLE_SIZE][j][tcase] = 0;
          inst_end_cycles[i][DOUBLE_SIZE][j][tcase] = 0;
        }
      }
    }
  }
  return (inst_set_list);
}



