/*
*****************************************************************************
* 
*   FILE NAME       :   LIB.C
*   PROGRAMMER(S)   :   Robert Arnold
*   DATE CREATED    :   July 19, 1993
*
*   DESCRIPTION :
*      This file contains library functions on the structures defined in
*      the file time.h
*
*   DEBUGGING ROUTINES:
*      Print_Cache_Block(struct cache_block *)
*      Print_Cache_List(struct cache_block *)
*      Print_Category_List(struct category_list *)   -- Prints list of inst cat
*      Print_Inst_Cat_List(struct inst_cat *)        -- Prints tnode l. of inst
******************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include "stats.h"
#include "time.h"
#include "tree.h"

/* Added by Sibin M on Oct 25, 2004.
 * Process of cleaning up the TA for C++
 */
 
#include "lib.h"
#include <cstring>
 
/* Added by Sibin M on Oct 25, 2004.
 * Process of cleaning up the TA for C++
 */

#ifndef INST_H
#include "inst.h"
#endif

#ifdef DISPLAY
#define printc printf
#else
#define printc  
#endif

extern int memory_count;

/*********************************************************
*
*        FUNCTIONS FOR BLOCK_LIST
*
**********************************************************/
/******************************************************************
*
*   FUNCTION NAME : Create_Block_List
*
*******************************************************************/
struct block_list * Create_Block_List( int block_number )
{
  static struct block_list *block_ptr;

  memory_count += sizeof (struct block_list);
  block_ptr = (struct block_list *)
              malloc (sizeof(struct block_list));
  block_ptr->block_number = block_number;
  block_ptr->next = NULL;

  return block_ptr;
}


/******************************************************************
*
*   FUNCTION NAME : Append_Block_Blocklist
*
*   DESCRIPTION:
*      This function will take the block number and append it to the
*      end of the specified path.
*
*******************************************************************/
void Append_Block_Blocklist( int block_number, struct block_list **path )
{
  struct block_list *block_ptr;
  struct block_list *list_ptr;

  block_ptr = Create_Block_List (block_number);

  if (*path)
  {
    for (list_ptr = *path; list_ptr->next; list_ptr = list_ptr->next)
      ;
    list_ptr->next = block_ptr; 
  }
  else
    *path = block_ptr;
}


/******************************************************************
*
*   FUNCTION NAME : Delete_Blocklist 
*
*   DESCRIPTION:
*      This routine will search a list of blocks for the specified block number
*      and delete the block out of the list if found.
*
*******************************************************************/
void Delete_Blocklist( int block_number, struct block_list **list )
{
  struct block_list   *block_ptr;
  struct block_list   *list_ptr = *list;
  struct block_list   *prev_ptr;

  for (block_ptr=*list; block_ptr; block_ptr=block_ptr->next)
  {
    if (block_number == block_ptr->block_number)
    {
      if (block_ptr == list_ptr)              
      {
        *list = list_ptr->next;
        block_ptr->next = NULL;
      }
      else                                   
      {
        for (prev_ptr=*list;
             prev_ptr->next != block_ptr;
             prev_ptr=prev_ptr->next)
          ;
        prev_ptr->next = block_ptr->next;
        block_ptr->next = NULL;
      }
      free (block_ptr);
      memory_count -= sizeof (struct block_list);
    }
  }
}


/******************************************************************
*
*   FUNCTION NAME : Is_In_Block_List
*
*   DESCRIPTION:
*      This function will determine if the block_number specified
*      is contained in the list of blocks passed in.
*
*******************************************************************/
bool Is_In_Block_List( int block_number, struct block_list *list )
{
  struct block_list *list_ptr;

  for (list_ptr = list; list_ptr; list_ptr = list_ptr->next)
    if (block_number == list_ptr->block_number)
    {
      return true;
    }
  return false;
}


/*********************************************************
*
*        FUNCTIONS FOR CACHE BLOCK
*
**********************************************************/
void Add_Block_To_Cache_List( int line_number, int inst_number, 
			      int instance, char best_case, 
			      char worst_case, struct category_list *cat_list,
			      struct cache_block **list )
{
  struct cache_block   *block_ptr = NULL;

  struct category_list *clist = NULL;

            /* If cat_list->next == NULL then inst */
            /* will always be a hit so no need to */
            /* add to cache list. */

  if ( ((cat_list->best_case == Hit_INST_CATEGORY) ||
        (cat_list->worst_case == Hit_INST_CATEGORY)  )  &&
        (cat_list->next == NULL)                           )
    return;


  Copy_Category_List(&clist,cat_list);
  block_ptr = Create_Cache_Block(line_number, inst_number, instance, 
                                 best_case, worst_case,
                                 clist);
  block_ptr->next = *list;
  *list = block_ptr;
}


/******************************************************************
*
*   FUNCTION NAME : Create_Cache_Block
*
*******************************************************************/
struct cache_block * Create_Cache_Block( int line_number, int inst_number,
					 int instance, char best_case, char worst_case, 
					 struct category_list *cat_list )
{
  static struct cache_block *block_ptr;

  memory_count += sizeof (struct cache_block);
  block_ptr = (struct cache_block *) malloc (sizeof(struct cache_block));

  block_ptr->line_number = line_number;
  block_ptr->inst_number = inst_number;
  block_ptr->instance = instance;
  block_ptr->trans_occurred = false;
  block_ptr->cat_list = cat_list;
  block_ptr->best_case_cat = best_case;
  block_ptr->worst_case_cat = worst_case;
  block_ptr->next = NULL;

  return (block_ptr);
}

int cache_blocks_match( struct cache_block* list1, struct cache_block* list2 )
{
  struct cache_block *p1, *p2;
  int p1_sum = 0, p1_size = 0, p1_sum2 = 0, p2_sum = 0, p2_size = 0, p2_sum2 =0;
  extern bool verbose;

  for (p1 = list1; p1; p1 = p1->next)
  {
    p1_sum += p1->inst_number;
    p1_sum2 += p1->inst_number * p1->inst_number;
    p1_size++;
  }
  for (p2 = list2; p2; p2 = p2->next)
  {
    p2_sum += p2->inst_number;
    p2_sum2 += p2->inst_number * p2->inst_number;
    p2_size++;
  }

  if (verbose)
    printc ("list 1's sum = %d    list 2's sum = %d\n", p1_sum, p2_sum);

  if (p1_sum == p2_sum && p1_sum2 == p2_sum2 && p1_size == p2_size)
  {
    if (verbose)
      printc ("cache blocks match\n");
    return 1;
  }
  else
  {
    if (verbose)
      printc ("cache blocks do not match\n");
    return 0;
  }
}


/******************************************************************
*
*   FUNCTION NAME : Copy_Cache_List
*
*******************************************************************/
void Copy_Cache_List( struct cache_block** target_list, 
		      struct cache_block* source_list )
{
  struct cache_block *block_ptr = NULL;
  struct cache_block *sblock_ptr = NULL;
  struct cache_block *tblock_ptr = NULL;

  struct category_list *cat_list = NULL;

  *target_list = NULL;

  for (sblock_ptr = source_list;
       sblock_ptr; 
       sblock_ptr = sblock_ptr->next)
  {
    Copy_Category_List (&cat_list, sblock_ptr->cat_list);
    block_ptr = Create_Cache_Block(sblock_ptr->line_number, 
                                   sblock_ptr->inst_number, 
                                   sblock_ptr->instance,
                                   sblock_ptr->best_case_cat,
                                   sblock_ptr->worst_case_cat, 
                                   cat_list);

    block_ptr->trans_occurred = sblock_ptr->trans_occurred;
    if (*target_list)
      tblock_ptr->next = block_ptr;
    else
      *target_list = block_ptr;

    tblock_ptr = block_ptr;      
  }  
}


/************************************************************
*
*   FUNCTION NAME : Print_Cache_Block
*
*   DESCRIPTION:
*      USED FOR DEBUGGING INSIDE DBX 
*
*************************************************************/

void Print_Cache_Block( struct cache_block* cblock )
{
  printc ("inst_number              = %d\n", cblock->inst_number);
  printc ("line number              = %d\n", cblock->line_number);
  printc ("instance                 = %d\n", cblock->instance);

  if (cblock->trans_occurred == true)
    printc ("transition occurred      = true\n");
  else
    printc ("transition occurred      = false\n");
/*
  printc ("worst case prev category = %c\n", cblock->prev_wc_cat);
  printc ("best case prev category  = %c\n", cblock->prev_bc_cat);
*/
  printc ("worst case last category = %c\n", cblock->worst_case_cat);
  printc ("best case last category  = %c\n", cblock->best_case_cat);
  Print_Category_List(cblock->cat_list);
  printc ("\n\n");
}


/*********************************************************
*
*   FUNCTION NAME : Print_Cache_List
*
*   DESCRIPTION:
*      USED FOR DEBUGGING INSIDE DBX 
*
**********************************************************/

void Print_Cache_List( struct cache_block* cblock_list )
{
  struct cache_block *cb_list;

  for (cb_list = cblock_list; cb_list; cb_list = cb_list->next)
    Print_Cache_Block(cb_list);

  printc ("\n\n\n\n");
}


void Union_Cache_List( struct cache_block** target_list, 
		       struct cache_block* source_list )
{
  struct cache_block *block_ptr = NULL;
  struct cache_block *sblock_ptr = NULL;
  struct cache_block *tblock_ptr = NULL;

  struct cache_block *tmp_ptr = NULL;

  struct category_list *cat_list = NULL;

  extern bool verbose;

  for (tblock_ptr = *target_list; 
       tblock_ptr && tblock_ptr->next; 
       tblock_ptr = tblock_ptr->next)
    ;

  for (sblock_ptr = source_list; sblock_ptr ; sblock_ptr = sblock_ptr->next)
  {
    for (tmp_ptr = *target_list; tmp_ptr ; tmp_ptr = tmp_ptr->next)
    {
      if ( ( tmp_ptr->line_number == sblock_ptr->line_number) &&
           ( tmp_ptr->instance    == sblock_ptr->instance)       )
      {
        if ((Matching_Cache_Blocks(tmp_ptr, sblock_ptr) == false) )
        { 
	  if (verbose) {
	    printc ("\n\n");
	    printc ("WARNING: During Cache Union 2 blocks with same cache",
		    " line %d did not match\n", sblock_ptr->line_number);
	    Print_Cache_Block(sblock_ptr);
	    printc ("\n"); 
	    Print_Cache_Block(tmp_ptr);
	    printc ("\n\n");
	  }

        } 
        break;
      } 
    }  
    if (tmp_ptr == NULL)
    { 
      Copy_Category_List (&cat_list, sblock_ptr->cat_list);
      block_ptr = Create_Cache_Block(sblock_ptr->line_number,
                                     sblock_ptr->inst_number,
                                     sblock_ptr->instance,
                                     sblock_ptr->best_case_cat,
                                     sblock_ptr->worst_case_cat, 
                                     cat_list);

      block_ptr->trans_occurred = sblock_ptr->trans_occurred;
      if (*target_list)
        tblock_ptr->next = block_ptr;
      else
        *target_list = block_ptr;

      tblock_ptr = block_ptr;
    }
  }  
}


void Update_Cache_Block( struct cache_block* cache_block )
{
  struct category_list *block_ptr;

  if (cache_block  && cache_block->cat_list)
  {
    cache_block->prev_bc_cat = cache_block->best_case_cat;
    cache_block->prev_wc_cat = cache_block->worst_case_cat;
    cache_block->best_case_cat = cache_block->cat_list->best_case;
    cache_block->worst_case_cat = cache_block->cat_list->worst_case;

    block_ptr = cache_block->cat_list;
    cache_block->cat_list = block_ptr->next;
    block_ptr->next = NULL;

    free(block_ptr);
    memory_count -= sizeof (struct category_list);
  }
}


void Update_Cache_List( struct cache_block* cache_list )
{
  struct cache_block *block_ptr;

  for (block_ptr = cache_list; block_ptr; block_ptr = block_ptr->next)
    Update_Cache_Block(block_ptr);
}


/* Let's write a GENERAL function that has two parameters that tells what 
   transition to look for. (8/1) */

bool transition( char from, char to, 
			 struct cache_block* cache_block,
			 enum time_case tcase )
{
  bool  transition = false;

  switch (tcase)
  {
    case WORST : if ((cache_block) &&
                     (cache_block->cat_list) &&
                     (cache_block->worst_case_cat == from) &&
                     (cache_block->cat_list->worst_case == to))
                   transition = true;
                 break;

    case BEST  : if ((cache_block) &&
                     (cache_block->cat_list) &&
                     (cache_block->best_case_cat == from) &&
                     (cache_block->cat_list->best_case == to))
                   transition = true;
                 break;

    default    : fprintf (stderr, "transition(): illegal timing case %d\n",
                                  tcase);
                 exit(1);
  }

  return transition;
}


/* A function to count how many (fh)list elements are still considered f or m */

int Determine_Num_Still_SomeMiss( struct cache_block* pcache_list, 
				  struct cache_block* ccache_list, 
				  enum time_case tcase )
{
  int num_still_somemiss = 0, found;
  struct cache_block  *ccache_ptr, *pcache_ptr;

  if (tcase == WORST)
  {
    for (ccache_ptr = ccache_list; ccache_ptr; ccache_ptr = ccache_ptr->next)
    {
      if (ccache_ptr->worst_case_cat == Firstmiss_INST_CATEGORY ||
          ccache_ptr->worst_case_cat == Miss_INST_CATEGORY)
      {

        found = 0;
        for (pcache_ptr = pcache_list; 
             pcache_ptr; 
             pcache_ptr = pcache_ptr->next)
        {
          if (pcache_ptr->line_number == ccache_ptr->line_number &&
              pcache_ptr->instance    == ccache_ptr->instance)
          {
            found = 1;
            break;
          }
        }
        
        if (found == 0)

          ++num_still_somemiss;
      }
    }
  }

  return num_still_somemiss;
}


/* This function will look at a child's (fh) cache list, and for each element:
   if (it is currently categorized as a first hit  AND  prev cat is also fh)
      if (not in parent's cache list (no match with both line # & instance))
         Put it the parent's cache list
      else 
         increment counter, because it's already in parent's list.   (4/28) 
*/
 
int Determine_Num_New_Miss2Hit( struct cache_block* pcache_list, 
				struct cache_block* ccache_list,
				enum time_case tcase )
{
  int num_new_miss2hit = 0, found;
  struct cache_block  *ccache_ptr, *pcache_ptr;

  if (tcase == WORST)
  {
    for (ccache_ptr = ccache_list; ccache_ptr; ccache_ptr = ccache_ptr->next)
    {
      if (ccache_ptr->worst_case_cat == Firsthit_INST_CATEGORY &&
          ccache_ptr->prev_wc_cat != Firsthit_INST_CATEGORY)
      {
        found = 0;
        for (pcache_ptr = pcache_list; 
             pcache_ptr; 
             pcache_ptr = pcache_ptr->next)
        {
          if (pcache_ptr->line_number == ccache_ptr->line_number &&
              pcache_ptr->instance    == ccache_ptr->instance)
          {
            found = 1;
            break;
          }
        }
        
        if (found != 0)
          ++num_new_miss2hit;
      }
    }
  }
  else 		/* for BEST case */
  {
    for (ccache_ptr = ccache_list; ccache_ptr; ccache_ptr = ccache_ptr->next)
    {
      if (ccache_ptr->best_case_cat == Firsthit_INST_CATEGORY &&
          ccache_ptr->prev_bc_cat == Firsthit_INST_CATEGORY)
      {
        found = 0;
        for (pcache_ptr = pcache_list; 
             pcache_ptr; 
             pcache_ptr = pcache_ptr->next)
        {
          if (pcache_ptr->line_number == ccache_ptr->line_number &&
              pcache_ptr->instance    == ccache_ptr->instance)
          {
            found = 1;
            break;
          }
        }
        
        if (found != 0)
          ++num_new_miss2hit;
      }
    }
  }
  return num_new_miss2hit;  
}


/* I need a similar function for counting f->f transitions. (4/18) */

int Determine_Num_FirstMiss_FM_Trans( struct cache_block* pcache_list, 
				      struct cache_block* ccache_list,
				      enum time_case tcase,
				      int* fut_trans )
{
  int number_transitions = 0;

  struct cache_block *pcache_ptr;
  struct cache_block *ccache_ptr;

  *fut_trans = 0;

  for (ccache_ptr = ccache_list; ccache_ptr ; ccache_ptr = ccache_ptr->next)
  {
    for (pcache_ptr = pcache_list;
         pcache_ptr; 
         pcache_ptr = pcache_ptr->next)
      if (ccache_ptr->line_number == pcache_ptr->line_number &&
          ccache_ptr->instance    == pcache_ptr->instance    &&
          ccache_ptr->trans_occurred == false)
        break;
				/* call the right function ! (4/18) */

    if (tcase == WORST)
    {
      if (! pcache_ptr &&
          ccache_ptr->worst_case_cat == 'f')
      {
        number_transitions++;
        ccache_ptr->trans_occurred = true;

        if (transition('f', 'f', ccache_ptr, tcase) == true)
          ++*fut_trans;
      }
    }
    else /* BEST case */
    {
      if (! pcache_ptr &&
          ccache_ptr->best_case_cat == 'f')
      {
        number_transitions++;
        ccache_ptr->trans_occurred = true;

        if (transition('f', 'f', ccache_ptr, tcase) == true)
          ++*fut_trans;
      }
    }
  }
  return number_transitions;  
}


/* A general function to detect a transition in the list (8/2) */

bool ever_transition( char from, char to, 
			      struct category_list* clist, 
			      enum time_case tcase )
{
  struct category_list *from_ptr;

  switch (tcase)
  {
    case WORST : for (from_ptr = clist;
                      from_ptr && from_ptr->next;
                      from_ptr = from_ptr->next)
                 {
                   if ((from_ptr->worst_case == from) &&
                       (from_ptr->next->worst_case == to))
                   {
                     return true;
                   }
                 }
                 break;

    case BEST  : for (from_ptr = clist;
                      from_ptr && from_ptr->next;
                      from_ptr = from_ptr->next)
                 {
                   if ((from_ptr->best_case == from) &&
                       (from_ptr->next->best_case == to))
                   {
                     return true;
                   }
                 }
                 break;

    default    : fprintf (stderr, "ever_transition():illegal timing case %d\n",
                                  tcase);
                 exit(1);
  }
  return false;		/* we didn't find the transition in the list */
}


/* Determine if a f OR m instruction EVER becomes an i.  (4/28) */

bool SomeMiss_FirstHit_Trans( struct category_list *clist, 
				      enum time_case tcase )
{
  struct category_list *sm_ptr;

  if (tcase == WORST)
  {
    for (sm_ptr = clist;
         sm_ptr && sm_ptr->next;
         sm_ptr = sm_ptr->next)
    {
      if (  ((sm_ptr->worst_case == Firstmiss_INST_CATEGORY) ||
             (sm_ptr->worst_case == Miss_INST_CATEGORY))       &&
            (sm_ptr->next->worst_case == Firsthit_INST_CATEGORY) )
        return true;

    }
  }
  else		/* for BEST case */
  {
    for (sm_ptr = clist;
         sm_ptr && sm_ptr->next;
         sm_ptr = sm_ptr->next)
    {
      if (  ((sm_ptr->best_case == Firstmiss_INST_CATEGORY) ||
             (sm_ptr->best_case == Miss_INST_CATEGORY))       &&
            (sm_ptr->next->best_case == Firsthit_INST_CATEGORY) )
        return true;

    }
  }
  return false;
}


bool Matching_Cache_Blocks( struct cache_block *block1, 
				    struct cache_block *block2 )
{
  struct category_list *list1_ptr;
  struct category_list *list2_ptr;

  for (list1_ptr = block1->cat_list, list2_ptr = block2->cat_list;
       list1_ptr && list2_ptr;
       list1_ptr = list1_ptr->next, list2_ptr = list2_ptr->next)
  {
    if (list1_ptr->worst_case != list2_ptr->worst_case ||
        list1_ptr->best_case !=  list2_ptr->best_case)
    {
      return false;
    } 
  } 

  if (list1_ptr == NULL  && list2_ptr != NULL  ||
      list1_ptr != NULL  && list2_ptr == NULL    )
  {
    return false;
  } 
  return true;
}


/************************************************************
*
*   FUNCTION NAME: Inst_In_Cache
*
*************************************************************/
bool Inst_In_Cache( struct cache_block* fm_list, int line_number )
{
  struct cache_block *block_ptr;

  for (block_ptr = fm_list; block_ptr; block_ptr = block_ptr->next)
  {
    if (block_ptr->line_number == line_number)
      return true;
  } 
  return false;
}


/*********************************************************
*
*        FUNCTIONS FOR CATEGORY_LIST
*
**********************************************************/
struct category_list * Create_Category_List( char best_case, char worst_case )
{
  struct category_list *block_ptr = NULL;

  memory_count += sizeof (struct category_list);
  block_ptr = (struct category_list *)
              malloc (sizeof(struct category_list));
  block_ptr->best_case = best_case;
  block_ptr->worst_case = worst_case;
  block_ptr->next = NULL;

  return block_ptr;
}


void Copy_Category_List( struct category_list** target, 
			 struct category_list* source )
{
  struct category_list *block_ptr;
  struct category_list *tcurr_ptr;
  struct category_list *scurr_ptr;

  *target = NULL;

  for (scurr_ptr = source; scurr_ptr; scurr_ptr = scurr_ptr->next)
  {
    block_ptr = Create_Category_List (scurr_ptr->best_case, scurr_ptr->worst_case);

    if (*target)
      tcurr_ptr->next = block_ptr;
    else
      *target = block_ptr;

    tcurr_ptr = block_ptr;
  }  
}


/*********************************************************
*
*   FUNCTION NAME : Print_Category_List
*
*   DESCRIPTION:
*      USED FOR DEBUGGING INSIDE DBX 
*
**********************************************************/

void Print_Category_List( struct category_list* clist )
{
  struct category_list *clist_ptr;

  for (clist_ptr = clist; clist_ptr; clist_ptr = clist_ptr->next)
    printc (" %c/%c ", clist_ptr->worst_case, clist_ptr->best_case);
  
  printc ("\n");
}


void Print_Cat_List( struct category_list* cat_list )
{
  struct category_list *cat_ptr;

  for (cat_ptr = cat_list; cat_ptr; cat_ptr = cat_ptr->next)
  {
    printc ("     Best Case  = %c\n", cat_ptr->best_case);
    printc ("     Worst Case = %c\n", cat_ptr->worst_case);
  }  
} 


/*********************************************************
*
*        FUNCTIONS FOR INST_CAT
*
*   inst_cat is the representation of the instructions
*   produced from the static cache simulator.
*
**********************************************************/
/*********************************************************
*
*   FUNCTION NAME : Inst_Same_Cache_Line
*
*   DESCRIPTION:
*      Given 2 instructions, this function will determine if
*      they are from the same cache line.
*
*
*   ASSUMPTIONS:
*
**********************************************************/
bool Inst_Same_Cache_Line( struct inst_cat* inst1_ptr, 
				   struct inst_cat* inst2_ptr )
{
  if (inst1_ptr->line_number == inst2_ptr->line_number)
    return true;
  else
    return false;
}


/*********************************************************
*
*        FUNCTIONS FOR BLOCK_ELEMENT
*
**********************************************************/
/*********************************************************
*
*   FUNCTION NAME : Create_Block_Element
*
**********************************************************/
struct block_element * Create_Block_Element( int block_number ) /* assigned by vpo */
{
  struct block_element * block_ptr;

  memory_count += sizeof(struct block_element );
  block_ptr = (struct block_element *)
              malloc (sizeof(struct block_element ));

  block_ptr->block_number = block_number;
  block_ptr->instruct_list = NULL;
  block_ptr->pred = NULL;
  block_ptr->succ = NULL;
  block_ptr->infeasible_succ = NULL;
  block_ptr->unknown = NULL;
  block_ptr->branch = NULL;
  block_ptr->fallthru = NULL;
  block_ptr->unfellthru = NULL;
  block_ptr->unbranched = NULL;
  block_ptr->ft_cause_b = NULL;
  block_ptr->ft_cause_ft = NULL;
  block_ptr->b_cause_b = NULL;
  block_ptr->b_cause_ft = NULL;
  block_ptr->left_min = 0;
  block_ptr->left_max = 0;
  block_ptr->right_min = 0;
  block_ptr->right_max = 0;
  block_ptr->dir = 'X';
  block_ptr->dir_iter = 0;
  block_ptr->dom = NULL;
  block_ptr->next = NULL;

  return block_ptr;
}


/*********************************************************
*
*   FUNCTION NAME : Create_Inf_Func_Element
*
**********************************************************/
//			(name,              /* Function name */
//			number,             /* Number assigned to function*/
//			filename)           /* inf filename func read from */
struct inf_func_element * Create_Inf_Func_Element( string name, 
						   int number, 
						   string filename )
{
  struct inf_func_element *inf_ptr;

  memory_count += sizeof(struct inf_func_element);
  inf_ptr = (struct inf_func_element *)
            malloc (sizeof(struct inf_func_element));

  strcpy (inf_ptr->func, name);
  inf_ptr->number = number;
  strcpy (inf_ptr->inf_filename, filename);

  /*
   * Next 2 values used in comparisons to determine the range of the
   * block numbers contained within each function.  This range of
   * numbers should be used in determining paths across function calls.
   */
  inf_ptr->min_block_number = 0;
  inf_ptr->max_block_number = MAXINT;

  inf_ptr->timing_label = NULL;	/* l.l. of @-timing labels (10/1) */
  inf_ptr->block = NULL;
  inf_ptr->loop = NULL;
  inf_ptr->path = NULL;
  inf_ptr->next = NULL;

  return inf_ptr;
}

/*********************************************************
*
*   FUNCTION NAME : Create_Inst_Block
*
**********************************************************/
struct inst_block * Create_Inst_Block( int inst_num, char inst[MAX_INST_LENGTH] )
{
  int i;

  struct inst_block *inst_ptr;

  memory_count += sizeof (struct inst_block);
  inst_ptr = (struct inst_block *) malloc (sizeof(struct inst_block));
  inst_ptr->num = inst_num;
  for(i = 0; i < MAX_INST_LENGTH; inst_ptr->inst[i] = inst[i], i++);
  inst_ptr->next = NULL;

  return inst_ptr;
}


int Find_Max_Inst( struct instruct* inst_list )
{
  struct instruct *inst_ptr;
 
  if (inst_list == NULL)
    return -1;

  for (inst_ptr = inst_list; inst_ptr->next; inst_ptr = inst_ptr->next);

  return inst_ptr->number;
}


/*********************************************************
*
*   FUNCTION NAME : Create_Instruct
*
**********************************************************/
struct instruct * Create_Instruct( int number, string inst, int inst_type,
				   int data_type, int cond_code, int op1_data_type,
				   string op1_reg_list, int op2_data_type, 
				   string op2_reg_list, int op3_data_type,
				   string op3_reg_list, string called_func )
{
  int i;

  struct instruct *inst_ptr;

  memory_count += sizeof (struct instruct);
  inst_ptr = (struct instruct *) malloc (sizeof(struct instruct));

  inst_ptr->number = number;
  for(i = 0; i < MAX_INST_LENGTH; inst_ptr->inst[i] = inst[i], i++);
  inst_ptr->inst_type = inst_type;
  inst_ptr->data_type = data_type;
  inst_ptr->cond_code = cond_code;

  inst_ptr->op1.data_type = op1_data_type;
  for (i=0; i < MAX_STRING_LENGTH; i++)
    inst_ptr->op1.reg_list[i] = op1_reg_list[i];

  inst_ptr->op2.data_type = op2_data_type;
  for (i=0; i < MAX_STRING_LENGTH; i++)
    inst_ptr->op2.reg_list[i] = op2_reg_list[i];

  inst_ptr->op3.data_type = op3_data_type;
  for (i=0; i < MAX_STRING_LENGTH; i++)
    inst_ptr->op3.reg_list[i] = op3_reg_list[i];

  strcpy (inst_ptr->called_func, called_func);

  inst_ptr->next = NULL;
   
  return inst_ptr;
}


/*********************************************************
*
*   FUNCTION NAME : Find_Last_Inst_Num
*
*   DESCRIPTION:
*      Given a list of instructions, this function will go through
*      the list and determine the last instruction based upon the
*      assigned instruction number.
*
*   ASSUMPTIONS:
*      1. The instruction number is assigned from the cache simulator
*         and read in when the .ist file is processed.
*
*      2. It is not assumed that the instructions are in ascending order.
*     
**********************************************************/
int Find_Last_Inst_Num( struct instruct* inst_list )
{
  struct instruct *inst_ptr;
  int    last_inst = 0;

  for (inst_ptr = inst_list; inst_ptr; inst_ptr=inst_ptr->next)
    if (inst_ptr->number > last_inst)
      last_inst = inst_ptr->number;

  return last_inst;
}


/*********************************************************
*
*   FUNCTION NAME : Create_Loop_Element
*
**********************************************************/
//		(number,             /* Loop number assigned by vpo compiler */
//		nest_level,         /* Nesting level of loop inside function */
//		min_iterations,     /* Minimum iterations of loop */
//		max_iterations,     /* Maximum iterations of loop */
//               abs_min_iterations,
//              abs_max_iterations,
//             induction_init,
//		induction_limit,
//               lower_bound,
//              upper_bound,
//		header)             /* Block number that is the loop header */
struct loop_element * Create_Loop_Element( int number, int nest_level,
					   int min_iterations, int max_iterations,
					   int abs_min_iterations, int abs_max_iterations,
					   int induction_init, int induction_limit,
					   int lower_bound, int upper_bound, int header )
{
  struct loop_element * loop_ptr;

  memory_count += sizeof (struct loop_element);
  loop_ptr = (struct loop_element *)
             malloc (sizeof(struct loop_element));

  loop_ptr->number = number;
  loop_ptr->nest_level = nest_level;
  loop_ptr->min_iterations = min_iterations;
  loop_ptr->max_iterations = max_iterations;

  loop_ptr->abs_min_iterations = abs_min_iterations;
  loop_ptr->abs_max_iterations = abs_max_iterations;
  loop_ptr->induction_init = induction_init;
  loop_ptr->induction_limit = induction_limit;

  loop_ptr->lower_bound = lower_bound;
  loop_ptr->upper_bound = upper_bound;
  loop_ptr->header = header;

  loop_ptr->block_list = NULL;
  loop_ptr->exit_list = NULL;
  loop_ptr->next = NULL;

  return loop_ptr;
}


/*********************************************************
*
*   FUNCTION NAME : Find_Loop
*
*   DESCRIPTION:
*      this function will take as input a pointer to an inf function
*      structure and a loop number and return a pointer to the loop
*      element structure that has the corresponding loop number.
*
*      function will return null if it can't find the loop.
*
*   ASSUMPTIONS:
*      1. Since the vpo compiler assigns loop numbers, it is assumed
*         that the loop numbers inside a function are unique.
*
**********************************************************/
struct loop_element * Find_Loop ( struct inf_func_element* inf_ptr,
				  int header_block )
{
  struct loop_element *loop_ptr;

  for (loop_ptr = inf_ptr->loop; loop_ptr; loop_ptr = loop_ptr->next)
    if (loop_ptr->header == header_block)
      return loop_ptr;

  return NULL;
}


/*********************************************************
*
*        FUNCTIONS FOR LOOP_NODE
*
**********************************************************/
/*********************************************************
*
*   FUNCTION NAME : Create_Loop_Node
*
*   DESCRIPTION:
*      This function will allocate memory for the struct loop_node and 
*      return a pointer to this memory location
*   ASSUMPTIONS:
*
*   NOTES:
*      1. min_iteration and max_iteration is duplicated in this struct
*         in order to place all important info in 1 struct when timing
*         a loop node.
*
**********************************************************/
struct loop_node * Create_Loop_Node( int number, int block_number,
				     int nest_level, int min_iterations, 
				     int max_iterations, int abs_min_iterations, 
				     int abs_max_iterations )
{
  struct loop_node *node_ptr;

  memory_count += sizeof (struct loop_node);
  node_ptr = (struct loop_node *) malloc (sizeof(struct loop_node));
  node_ptr->number       = number;
  node_ptr->header_block = block_number;
  node_ptr->nest_level = nest_level;
  node_ptr->min_iterations = min_iterations;
  node_ptr->max_iterations = max_iterations;
  node_ptr->abs_min_iterations = abs_min_iterations;
  node_ptr->abs_max_iterations = abs_max_iterations;
  node_ptr->path = NULL;
  node_ptr->next = NULL;

  return node_ptr;
}    


/*********************************************************
*
*   FUNCTION NAME : Create_Loop_Block_List
*
**********************************************************/
struct loop_block_list * Create_Loop_Block_List( int block_number, 
						 bool is_loop_node,
						 struct loop_node* node_ptr ) 
{
  struct loop_block_list *block_ptr;

  memory_count += sizeof (struct loop_block_list);
  block_ptr = (struct loop_block_list *)
              malloc (sizeof(struct loop_block_list));
  block_ptr->block_number = block_number;
  block_ptr->is_loop_node = is_loop_node;
  block_ptr->loop_node_ptr = node_ptr;
  block_ptr->next = NULL;

  return block_ptr;
}


/*********************************************************
*
*   FUNCTION NAME : Is_Same_Loop_Path
*
*   DESCRIPTION:
*
*      This function will take in two paths and determine if they are the
*      same path.  This is used to identify the corresponding EXIT path
*      to a given BACKEDGE.
*
**********************************************************/

bool Is_Same_Loop_Path( struct loop_block_list* path1, 
				struct loop_block_list* path2 )
{
  struct loop_block_list  *path1_ptr;
  struct loop_block_list  *path2_ptr;

  for(path1_ptr = path1, path2_ptr = path2; 
      path1_ptr && path2_ptr; 
      path1_ptr = path1_ptr->next, path2_ptr = path2_ptr->next)
    if (path1_ptr->block_number != path2_ptr->block_number)
      return false;

  if (path1_ptr && path2_ptr == NULL  ||
      path1_ptr == NULL && path2_ptr     )
    return false;
  else
    return true;
}


/*********************************************************
*
*   FUNCTION NAME : Append_Block_Pathlist
*
**********************************************************/
void Append_Block_Pathlist( int block_number, 
			    struct loop_node* loop_node_list, 
			    struct loop_block_list* path )
{
  bool is_loop_node = false;

  struct loop_block_list  *list_ptr=NULL;
  struct loop_block_list  *block_ptr;
  struct loop_node        *loop_node_ptr;

  struct loop_block_list * Create_Loop_Block_List ();
  struct loop_node *       Loop_Node_Ptr_Addr();

  loop_node_ptr = Loop_Node_Ptr_Addr(block_number, loop_node_list);
  if (loop_node_ptr)
    is_loop_node = true;

  block_ptr = Create_Loop_Block_List(block_number, is_loop_node, loop_node_ptr);

  if (path)
  {
    for (list_ptr = path; list_ptr->next; list_ptr = list_ptr->next)
      ;
  
    list_ptr->next = block_ptr;
  }
  else
    path = block_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Copy_Loop_Block_List
*
*********************************************************/
struct loop_block_list * Copy_Loop_Block_List( struct loop_block_list* list )
{
  struct loop_block_list *head_ptr = NULL;
  struct loop_block_list *curr_ptr = NULL;
  struct loop_block_list *block_ptr = NULL;

  struct loop_block_list *list_ptr;

  for (list_ptr = list; list_ptr; list_ptr = list_ptr->next)
  {
    block_ptr = Create_Loop_Block_List(list_ptr->block_number,
                                       list_ptr->is_loop_node,
                                       list_ptr->loop_node_ptr);
    if (head_ptr)
    {
      curr_ptr->next = block_ptr;
      curr_ptr = curr_ptr->next;
    } 
    else
    {
      head_ptr = block_ptr;
      curr_ptr = block_ptr;
    }
  }
  return head_ptr;
}


/*********************************************************
*
*        FUNCTIONS FOR LOOP_PATH_NODE
*
**********************************************************/
/********************************************************
*
*   FUNCTION NAME : Create_Loop_Path
*
*   DESCRIPTION:
*      This function will allocate memory for the struct loop_path_node
*      and return a pointer to that memory location.
*
*********************************************************/
struct loop_path_node * Create_Loop_Path( int block_number )
{
  int i;
  struct loop_path_node *node_ptr;  /* Create the node for the specific path */

  memory_count += sizeof(struct loop_path_node);
  node_ptr = (struct loop_path_node *)
             malloc (sizeof(struct loop_path_node));
  node_ptr->min      = 0;
  node_ptr->max      = 0;
  node_ptr->min_iter       = MAXINT;
  node_ptr->max_iter       = MAXINT;
  node_ptr->take_away      = 0;
  node_ptr->already_took_away = false;
  node_ptr->max_take_away  = 0;
  node_ptr->cant_next      = NULL;
  node_ptr->header_block   = block_number;
  node_ptr->path_list      = NULL;
  node_ptr->exit_list      = NULL;
  node_ptr->best_fm_list   = NULL;
  node_ptr->worst_fm_list  = NULL;
  node_ptr->path_type_mask = NONE;
  node_ptr->shared_fm_time = 0;
#if 0
  node_ptr->est.extra_time = 0;	
  node_ptr->est.extra_fh_time = 0;
  node_ptr->est.extra_fm_time = 0;
  node_ptr->est.bc_extra_time = 0;

  node_ptr->est.beg_occupant[NUM_STAGES] = NULL;
  node_ptr->est.end_occupant[NUM_STAGES] = NULL;
  for (i = 0; i < NUM_STAGES; ++i)
  {
    node_ptr->est.beg_cycles[i] = 0;
    node_ptr->est.end_cycles[i] = 0;
  }
  node_ptr->est.path_cycles = 0;
  node_ptr->est.inner_ff_trans = 0;
  node_ptr->est.fm_encountered = 0;
  node_ptr->est.fh_encountered = 0;
  for (i = 0; i < NUMREGS; ++i)
  {
    node_ptr->est.release_time[i] = 0;
    node_ptr->est.src_needed[i] = 0;
    node_ptr->est.src_stage[i] = NO_STAGE;
  }
  node_ptr->est.beg_cache_line = -1;
  node_ptr->est.first_word_num = -1;
  node_ptr->est.end_cache_line = -1;
  node_ptr->est.beg_avail_time = NULL;
  node_ptr->est.end_avail_time = NULL;
  node_ptr->est.fm_end_avail_time = NULL;
  node_ptr->est.first_line_delay = 0;
  node_ptr->est.last_line_delay = 0;
  node_ptr->est.delay_after_fm = 0;
  node_ptr->est.first_prog_offset = 0;
  node_ptr->est.beg_word_num = NULL;
  node_ptr->est.last_fm = 0;
  node_ptr->est.fm_dead_cycles = 0;
  node_ptr->est.last_miss_cat = '\0';		/* either a 'm' or a 'f' */
  node_ptr->est.finish_load = 0;	/* # cycles after IF done 'til loaded */
  node_ptr->est.delay_after_fm_head = NULL;
#endif
  node_ptr->next = NULL;
                                        /* Place the header block in the path */

  /*BOOKMARK - also add the new frequency model variables */

  node_ptr->icache_misses = 0;
  node_ptr->dcache_misses = 0;
  node_ptr->path_cycles = 0;
  node_ptr->overlap = 0;
  node_ptr->extra_latency = 0;
  node_ptr->latency_considered = 0;
  node_ptr->total_latency = 0;
  node_ptr->diff = 0;

  return node_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Create_Path_Element
*
*********************************************************/
struct path_element * Create_Path_Element( int block_number,
					   struct loop_node* func_loop_ptr )
{
  struct path_element  *path_ptr;

  struct loop_node * Loop_Node_Ptr_Addr();

  memory_count += sizeof(struct path_element);
  path_ptr = (struct path_element *)
             malloc (sizeof(struct path_element));
  path_ptr->block_number = block_number;
  path_ptr->succ = NULL;
  path_ptr->next = NULL;

  path_ptr->loop_node_ptr = Loop_Node_Ptr_Addr (block_number, func_loop_ptr);
  if (path_ptr->loop_node_ptr == NULL)
    path_ptr->is_loop_node = false;
  else
    path_ptr->is_loop_node = true;

  return path_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Find_Prev_Path_Block
*
*   DESCRIPTION:
*
*      Given a pointer to a block in a path_list, this function
*      will return a pointer to the previous block in the path.
*
*   WARNING:
*      It is assumed that the curr_block_ptr points to a block in the
*      path list.
*********************************************************/
struct loop_block_list * Find_Prev_Path_Block( int curr_block, 
					       struct loop_block_list* path_list )
{
  struct loop_block_list *block_ptr;

  if (curr_block == path_list->block_number)
    return NULL;
  else
    for (block_ptr = path_list;
         block_ptr && block_ptr->next;
         block_ptr = block_ptr->next)
    {
      if (block_ptr->next->block_number == curr_block)
        return block_ptr;
    }
  return NULL;				/* make lint happy */  
}


/*******************************************************
**
**               FUNCTIONS FOR NODE_BLOCK
**
*******************************************************/
/********************************************************
*
*   FUNCTION NAME : Append_Block_Path
*
*********************************************************/
void Append_Block_Path( int block_number, 
			struct loop_node* loop_node_list, 
			struct path_element* path )
{
  struct path_element *block_ptr;
  struct path_element *list_ptr;

  struct path_element * Create_Path_Element();

  block_ptr = Create_Path_Element (block_number, loop_node_list);
  for (list_ptr = path; list_ptr->succ; list_ptr = list_ptr->succ)
    ;
  list_ptr->succ = block_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Loop_Node_Ptr_Addr
*
*   DESCRIPTION:
*      This function will determine if a block number represents a loop node,
*      and if so, will return the address of that loop node.  If the block
*      does not represent a loop node, NULL will be returned.
*
*   ASSUMPTIONS:
*      1. the list of loop_nodes will be terminated by a pointer to NULL
*
*********************************************************/
struct loop_node * Loop_Node_Ptr_Addr( int block_number,
				       struct loop_node* loop_node_list )
{
  struct loop_node *node_ptr;

  for (node_ptr = loop_node_list; node_ptr; node_ptr=node_ptr->next)
    if (block_number == node_ptr->header_block)
      break;

  return node_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Multiple_Blocks_In_Blocklist
*
*   DESCRIPTION:
*      This function will determine if multiple successor blocks exist.
*
*********************************************************/
bool Multiple_Blocks_In_Blocklist( struct block_list* list )
{
  if (list && list->next)
    return true;
  else
    return false;
}


/********************************************************
*
*   FUNCTION NAME : Find_Pred_Blocklist
*
*   DESCRIPTION:
*      This function will take in a block number and
*      determine the list of successor blocks
*
*********************************************************/
struct block_list * Find_Pred_Blocklist( struct block_element*func_inf_ptr, 
					 int block_number )
{
  struct block_element *block_inf_ptr;
  struct block_list    *list_ptr = NULL;
   
  struct block_list    *head_ptr = NULL;
  struct block_list    *block_ptr = NULL;
  struct block_list    *curr_ptr = NULL;

  block_inf_ptr = Locate_Inf_Block( ( inf_func_element* )func_inf_ptr, block_number);

  if (block_inf_ptr)
  {
    for (list_ptr = block_inf_ptr->pred;list_ptr;list_ptr=list_ptr->next)
    {
      block_ptr = Create_Block_List(list_ptr->block_number);

      if (head_ptr)
      {
        curr_ptr->next = block_ptr;
        curr_ptr = curr_ptr->next;
      }           
      else
      {
        head_ptr = block_ptr;
        curr_ptr = block_ptr;
      }  
    }      
  }
  return head_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Copy_Loop_Path_Node_List
*
*********************************************************/
struct loop_path_node *  Copy_Loop_Path_Node_List( struct loop_path_node* list )
{
  struct loop_path_node *block_ptr = NULL;
  struct loop_block_list * Copy_Loop_Block_List();

  block_ptr = Create_Loop_Path (list->header_block);
  block_ptr->path_list = Copy_Loop_Block_List(list->path_list);

  block_ptr->exit_list = NULL;
  block_ptr->next = NULL;

  block_ptr->path_type_mask = list->path_type_mask;
  return block_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Copy_Path_List
*
*********************************************************/
struct path_element * Copy_Path_List( struct path_element* path, struct loop_node* func_loop )
{
  struct path_element *head_ptr = NULL;
  struct path_element *curr_ptr = NULL;
  struct path_element *block_ptr = NULL;
  struct path_element *list_ptr;

  struct path_element * Create_Path_Element();

  for (list_ptr = path; list_ptr; list_ptr = list_ptr->succ)
  {
    block_ptr = Create_Path_Element(list_ptr->block_number, func_loop);

    if (head_ptr)
    {
      curr_ptr->succ = block_ptr;
      curr_ptr = curr_ptr->succ;
    }
    else
    {
      head_ptr = block_ptr;
      curr_ptr = block_ptr;
    }
  }
  return head_ptr;
}


/********************************************************
*
*   FUNCTION NAME : Locate_Inf_Block
*
*   DESCRIPTION:
*      this function will take in as input a block number and
*      a pointer to a function inf structure and return a pointer
*      to the specified block element structure in the inf function.
*
*********************************************************/
struct block_element * Locate_Inf_Block( struct inf_func_element* inf, 
					 int block_number )
{
  struct block_element *block_ptr;

  for (block_ptr = inf->block; block_ptr; block_ptr = block_ptr->next)
  {
    if (block_ptr->block_number == block_number)
      return block_ptr;
  }  
  return NULL;
}


/*********************************************************
*
*        MISCELLANEOUS FUNCTIONS 
*
**********************************************************/
int Min( int int1, int int2 )
{
  if (int1 <= int2)
    return int1;
  else
    return int2;
}



void dump_INF(struct inf_func_element *myinf) 
{
  struct inf_func_element *temp;
  struct loop_element       *tloop;        
  struct block_element      *tblock;     
  struct path_element       *tpath;      
  struct loop_node          *tloop_nodes; 
  struct block_list *blist;
  struct instruct *inst;

  for(temp = myinf; temp; temp= temp->next) {
    printf("Name of func : %s\t number:%d\n",temp->func, temp->number);
    printf("read from %s \n",temp->inf_filename);
    printf("min blocks = %d  max blocks =%d\n",temp->min_block_number, temp->max_block_number);
    if (temp->timing_label == NULL) printf("Yes!! timing label is null\n");

    for (tloop= myinf->loop; tloop; tloop= tloop->next) {


      printf("number %d nest level %d miniter %d maxiter %d\n",tloop->number, tloop-> nest_level,
	     tloop-> min_iterations,tloop-> max_iterations);
      printf("block list ");
      for (blist = tloop->block_list; blist ; blist= blist->next) {
	printf(" %d",blist->block_number);
      }
      printf("\n");

         printf("exit list ");
      for (blist = tloop->exit_list; blist ; blist= blist->next) {
	printf(" %d",blist->block_number);
      }
      printf("\n");

    }
    
    for (tblock= myinf->block; tblock; tblock= tblock->next) {

      printf("instr list ");
      for (inst = tblock->instruct_list; inst ; inst= inst->next) {
	printf("num %d poff %d binst %d \n",inst->number, inst->prog_offset,
	       inst->block_inst);
	printf("name %s  type %d\n",inst->inst, inst->inst_type);
	printf("op1 %s op2 %s op3 %s func %s\n",inst->op1.reg_list,inst->op2.reg_list,inst->op2.reg_list,
	       inst->called_func);
      }
      printf("\n");
      
      printf("pred list ");
      for (blist = tblock->pred; blist ; blist= blist->next) {
	printf(" %d",blist->block_number);
      }
      printf("\n");
      
      printf("succ list ");
      for (blist = tblock->succ; blist ; blist= blist->next) {
	printf(" %d",blist->block_number);
      }
      printf("\n");
      printf("dom list ");
      for (blist = tblock->dom; blist ; blist= blist->next) {
	printf(" %d",blist->block_number);
      }
      printf("\n");
      
    }
    /*
    for (tpath= myinf->path; tpath; tpath= tpath->next) {
      Print_Loop_Node(tpath->loop_node_ptr);
      printf("is bl no %d a loop? %d \n",tpath->block_number,tpath-> is_loop_node);
    }
    for (tloop_nodes= myinf->loop_nodes; tloop_nodes; tloop_nodes= tloop_nodes->next) {
      Print_Loop_Node(tloop_nodes);
      }*/
  }
  

}



int larger( int a, int b )
{
  return (a > b) ? a : b;
}

