#include <stdio.h>
#include <ctype.h>
#include <strings.h>

#define MAXSTR 1024
#define TRUE 1
#define FALSE 0

#define FUNC 1000
#define OTHER 1001
#define LOOP 1002
#define BLOCK 1003

#define DOUBLE_WORD_OP 1
#define SINGLE_WORD_OP 0


/*now defining the character that terminates the labels */
#define LABEL_TERM ':'

/*******************ASSUMPTIONS**********************
 *  1) All labels are on a line by themselves
 *  2) Comments are preferably on a line by themselves
 *     or they don't contain any special characters that 
 *     can be found in labels or instructions 
 ****************************************************/

/*
 * line /w 80 chars (str),
 * pointer to (prev) line
 */
struct line {
  struct line *next;
  char str[MAXSTR];
};


/*                      
 * a stuct that maintains link list of all the instructions inside
 *
 */

struct instr {
  int BBid;         /*BB id*/
  int opcode;        
  char str[10];
  char rs[10];      /*int register info*/
  char rt[10];
  char rd[10];  
  char fs[10];      /*floating register info*/
  char ft[10];  
  char fd[10];  
  int imm;
  char offset[10];
  char target[MAXSTR];
  int size;         /*if info required for load,store,etc.*/
  int cc;           /*condition code info*/
  struct instr *next;
};


struct temp_inst {
  char str[10];
  int opcode;
  struct temp_inst *next;
};


/*
 * baslic block w/ textual successor (next), CFG
 * (preds) and (succs), Dominator- (doms) and
 * (loop) block list (only set iff Block == Kopf)
 * and 1st line (rtls) und block # nr
 */

/*CHANGE bblock to handle path info */
struct bblock {
  struct bblock *next;        /*next BB*/
  struct blist *preds;        /*list of preds*/
  struct blist *succs;        /*list of succs*/
  struct blist *doms;         /*list of dominators*/
  struct blist *loop;         /*list of BB forming loop..this BB is header*/
  struct line *rtls;          /*the lines in BB*/
  struct instr *instrs;       /*linked list of instructions in BB*/
  int nr;                     /*BB ID no.*/
  int num_inst;               /*num of instructions in BB */
  int offset;                 /*offset of BB */
  char label[MAXSTR];         /*the label of BB.. if any*/
  char target[MAXSTR];        /*target of BB .. if any*/
  int loopno;                 /*loop ID no.. if BB in loop*/
  int loopfreq;               /*loop nesting .. if BB in loop*/
  int *loop_blocklist;        /*list of BB ID's forming loop..also available in *loop*/
  int *exitpaths;             /*the BBs that the loop can exit to */
  int loop_bno;               /*num of blocks forming loop.. if BB in loop*/
  int numexits;               /* num of exit blocks */
};

/*
 * list of bblocks w/ one block (ptr)
 * and next list element (next)
 */
struct blist {
  struct blist *next;
  struct bblock *ptr;
};


/*List of functions appearing in isa.c*/

/*use following function to read the prologue info in an assembly file*/
void read_initialize(FILE *in, FILE *out);

/*reading basic blocks handled here */
struct line *read_block(FILE *in, char *mybuffy, char * mylabel, 
			char *mytarget, int * myinst, int *NEWFUNC);
/*reading the isa file to match up instructions to opcodes */
struct instr * read_instrs(struct temp_inst *head, struct line * rtls,
			   int id, int *num_inst);

/*making the cfg*/
void make_cfg(struct bblock *top);

/*writing the path file info*/
void path_info(FILE *outpath, struct bblock *top);

/*writing the inf file info*/
void inf_info(FILE *outinf, FILE *loopin, struct bblock *top);


/*********************************************************************/


/*List of functions appearing in CFG.c*/
/*all these funcitons are already defined and need not be changed 
  ... unless bugs are found ;) */


void error(char *s);

void compress_whitespace(char *s);

void write_rtls(FILE *out, struct line *rtl);

void write_block(FILE *out, struct bblock *cblk) ;

struct blist *add_blist(struct blist *list, struct bblock *nblk);

struct blist *add_sorted_blist(struct blist *root, struct bblock *nblk);

int is_in_blist(struct blist *list, struct bblock *cblk);

int is_in_sorted_blist(struct blist *list, struct bblock *cblk);

struct blist *reverse_blist(struct blist *plist, struct blist *list);

void print_blist(FILE *out, struct blist *list);

void free_blist(struct blist *list);

void find_target(struct bblock *cblk, struct bblock *tblk, char *label);

struct blist *modified_doms(struct bblock *cblk, int *change,
			    struct bblock *top);

void find_doms(struct bblock *top);

int dominates(struct bblock *dblk, struct bblock *cblk);

void insert(struct blist **loop, struct bblock *cblk, struct blist **stack);

void ReachUnder(struct bblock *cblk, struct bblock *dblk);

int subset(int *inner, int size1, int *outer, int size2) ;

void sort_list(int *list, int size);

void find_loops(struct bblock *top);

void find_nesting(struct bblock *top);
