/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: bit_enc.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Bit-encoding routines
 */

#include "scs.h"

#ifdef BIT_ENCODE
/*
 * print out the prediction for memory/cache fetches of the call graph
 */
void output_dag_prediction()
{
  struct funclist *fptr;
  struct pathlist *pptr;
  struct pathinfo *pinfo;
  struct fmline_t *fmptr;
  static struct fetchinfo_s fetchinfo = {0, 0};
  static struct fetchinfo_s pathinfo = {0, 0};
  static struct accessinfo_s afetchinfo = {0, 0, 0};
  static struct accessinfo_s apathinfo = {0, 0, 0};
  int refs;
#ifdef MERGE_STATES
  struct pathlist *pptr2;
  struct instinfo *iptr;
  vector v;
  int addr, addr2, i;
#ifdef DEBUG
  int maxi, from, to;
#endif
#endif

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
#ifdef MERGE_STATES
      if (pptr->path == pptr->path->equivs->path) { /* first in equiv. list */
	/*
	 * conflicts and misses
	 */
	v = vinit();
	pathinfo.mm = 0;
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next) {
	  pinfo = pptr2->path;
	  for (fmptr = pinfo->mmptr; fmptr; fmptr = fmptr->next)
	    if (!vin(v, fmptr->line)) {
	      vinsert(&v, fmptr->line);
	      pathinfo.mm++;
	    }
	}
	vfree(v);

	/*
	 * hits and first misses: find smallest address, mark all insts
	 * referenced by any path in the equivalence class, count them
	 */
	v = vinit();
	pathinfo.cache = 0;
	addr = pptr->path->equivs->path->insts->addr;
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next) {
	  pinfo = pptr2->path;
	  for (iptr = pinfo->insts; iptr; iptr = iptr->next)
	    if (iptr->addr < addr)
	      addr = iptr->addr;
	}
	addr >>= 2;
#ifdef DEBUG
	maxi = addr;
#endif
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next) {
	  pinfo = pptr2->path;
	  for (iptr = pinfo->insts; iptr; iptr = iptr->next)
	    for (i = 0; i < iptr->num; i++) {
	      addr2 = (iptr->addr >> 2) + i - addr;
	      if (!vin(v, addr2)) {
		vinsert(&v, addr2);
		pathinfo.cache++;
#ifdef DEBUG
		if (addr2 > maxi)
		  maxi = addr2;
#endif
	      }
	    }
	}
#ifdef DEBUG
	printf("paths ");
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next)
	  printf("%d ", pptr2->path->pathnum);

	printf("addr");
	to = 0;
	while (to <= maxi) {
	  for (from = to; from <= maxi && !vin(v, from); from++);
	  for (to = from; to <= maxi && vin(v, to); to++);
	  if (from <= maxi)
	    printf(" %d-%d", (from+addr)<<2, (to-1+addr)<<2);
	}
	printf("\n");
#endif
	vfree(v);
	fetchinfo.mm += pathinfo.mm;
	fetchinfo.cache += pathinfo.cache - pathinfo.mm;
      }
#else /*!MERGE_STATES*/
      pinfo = pptr->path;
      pathinfo.mm = 0;
      for (fmptr = pinfo->mmptr; fmptr; fmptr = fmptr->next)
	pathinfo.mm++;
      fetchinfo.mm += pathinfo.mm;
      fetchinfo.cache += pinfo->numinst - pathinfo.mm;
#endif /*!MERGE_STATES*/
    }

  refs = fetchinfo.mm + fetchinfo.cache;
#ifdef DEBUG
  printf("STATIC CACHE PREDICTION for call graph (DAG):\n");
  printf("\t%2.2f%%\t%d\t memory fetches\n",
         100.0 * (float) fetchinfo.mm    / (float) refs, fetchinfo.mm);
  printf("\t%2.2f%%\t%d\t cache fetches\n",
	 100.0 * (float) fetchinfo.cache / (float) refs, fetchinfo.cache);
#else /*!DEBUG*/
#ifdef BIT_ENCODE
  printf(" %6.2f%% %6.2f%%",
         100.0 * (float) fetchinfo.mm    / (float) refs,
	 100.0 * (float) fetchinfo.cache / (float) refs);
#endif /*BIT_ENCODE*/
#endif /*!DEBUG*/
}

/*
 * merge_equiv_prediction - add memory references of overlapping lines
 * of other paths into current path
 */
void merge_equiv_prediction()
{
  struct funclist *fptr;
  struct pathlist *pptr;
  struct fmline_t *fmptr;
  struct pathlist *pptr2;
  struct instinfo *iptr;
  vector v;

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      v = vinit();

      for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next)
	for (fmptr = pptr2->path->mmptr; fmptr; fmptr = fmptr->next)
	  for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	    if (iptr->firstline <= fmptr->line &&
		fmptr->line <= iptr->lastline &&
		!infmlist(pptr->path->mmptr, fmptr->line) &&
		!vin(v, fmptr->line)) {
	      vinsert(&v, fmptr->line);
	      addfmlist(&pptr->path->mmptr, fmptr->line);
	      break;
	    }
      vfree(v);
    }
}

/*
 * get_cost - return current cost of a path instance
 */
int get_cost(pinfo)
struct pathinfo *pinfo;
{
  struct callerlist *cptr;
  struct fmline_t *fmptr;
  struct costlist_t *costptr;
  int c, i, k;
  vector v;

  i = 0;
  v = vinit();  
  for (k = 0, cptr = pinfo->curfunc->caller; cptr; k++, cptr = cptr->next)
    for (fmptr = pinfo->fmptr[k]; fmptr; fmptr = fmptr->next)
      if (!infmlist(pinfo->mmptr, fmptr->line) && !vin(v, fmptr->line)) {
	vinsert(&v, fmptr->line);
	i++;
      }
  vfree(v);

  for (fmptr = pinfo->mmptr; fmptr; fmptr = fmptr->next)
    i++;

  c = 0;
  for (costptr = pinfo->costptr; costptr; costptr = costptr->next)
    c += costptr->cost;

  return ((pinfo->numinst - i) * HITCOST + i * MISSCOST + c
	  - pinfo->numinst * MMCOST);
}

#define NON_REC  0
#define BACKWARD 1
#define FORWARD  2

/*
 * move_cost - move the cost to neighboring paths, return TRUE if successful
 */
int move_cost(pinfo, head, mvcost, rec)
struct pathinfo *pinfo;
struct pathlist *head;
int mvcost, rec;
{
  struct pathlist *pptr;

  if (!head)
    return(FALSE);

  for (pptr = head; pptr; pptr = pptr->next)
    if (get_cost(pptr->path) + mvcost > 0) {
      if (!head->next) /* only one way to go */
	switch (rec) {
	case BACKWARD:
	  if (head->path->preds)
	    return(move_cost(pinfo, head->path->preds, mvcost, rec));
	case FORWARD:
	  if (head->path->succs)
	    return(move_cost(pinfo, head->path->succs, mvcost, rec));
	default:
	  ;
	}
      return(FALSE);
    }

  for (pptr = head; pptr; pptr = pptr->next)
    addcostlist(&pptr->path->costptr, pinfo, mvcost);

  addcostlist(&pinfo->costptr, pinfo, -mvcost);

  return(TRUE);
}

/*
 * check_savings - check if bit encoding resulted in savings over non-caching
 * for all paths
 */
void check_savings()
{
  struct funclist *fptr;
  struct pathlist *pptr;
  int mvcost, found;

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      /*
       * count # misses: MM fetches +
       *                 firstmisses which are not in MM fetch list
       */
      
      mvcost = get_cost(pptr->path);
      
      /*
       * move the remaining cost into preds/succs paths
       * NEEDS WORK
       */
      if (mvcost > 0) {
	found = move_cost(pptr->path, pptr->path->preds, mvcost, NON_REC);
	if (!found)
	  found = move_cost(pptr->path, pptr->path->succs, mvcost, NON_REC);
	if (!found)
	  found = move_cost(pptr->path, pptr->path->preds, mvcost, BACKWARD);
	if (!found)
	  found = move_cost(pptr->path, pptr->path->succs, mvcost, FORWARD);
	if (!found)
	  fprintf(stderr, "uncached cost %d < %d bit-cached for path%d%s\n",
		  pptr->path->numinst * MMCOST,
		  mvcost + pptr->path->numinst * MMCOST,
		  pptr->path->pathnum, fptr->func->funcname);
      }
    }
}

/*
 * output_static_savings - print out the static savings
 * for all paths
 */
void output_static_savings()
{
  struct funclist *fptr;
  struct pathlist *pptr;
  long min, max, avg, n, mvcost;

  avg = n = 0;
  min = max = -get_cost(funcs->func->paths->path); /* first path */
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      mvcost = -get_cost(pptr->path);
      avg += mvcost;
      min = MIN(min, mvcost);
      max = MAX(max, mvcost);
      n++;
    }

#ifdef DEBUG
  printf("STATIC PATH COST ANALYSIS\n");
  printf("avg: %2.2f cycle savings\n", (float) avg / (float) n);
  printf("min: %d cycle savings\n", min);
  printf("max: %d cycle savings\n", max);
#else
  printf(" %2.2f", (float) avg / (float) n);
#endif
}
#endif /*BIT_ENCODE*/

