/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: dead.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Dead code elimination
 */

#include "scs.h"

/*
 * dead_code_elim - eliminate unreachable code
 * This should be extended to eliminate paths with 0 instructions
 * and fix the control flow to skip these paths.
 * The paths then become dead code as well.
 */
void dead_code_elim()
{
  struct funclist *fptr;
  struct pathlist *pptr, *ppptr, *npptr;
  struct pathinfo *pinfo;
  struct instinfo *iptr, *niptr;
  struct pathlistlist *lptr;
  int i, j ,k, change;

  for (fptr = funcs; fptr; fptr = fptr->next) {
    change = TRUE;

    while (change) {
      change = FALSE;
      
      for (pptr = fptr->func->paths; pptr, npptr = pptr->next; pptr = npptr)
	if ((!pptr->path->preds &&
	     !inpathlist(fptr->func->entrypaths, pptr->path)) ||
	    pptr->path->numinst == 0) {
	  change = TRUE;
	  for (iptr = pptr->path->insts; iptr; iptr = niptr) {
	    i = iptr->startindex;
	    k = iptr->tag;
	    for (j = 0; j < iptr->numlines; j++) {
	      for (lptr = setpath[i]; lptr; lptr = lptr->next)
		if (lptr->tag == k)
		  break;
	      
	      del_pathlist(&lptr->list, pptr->path);
	      
	      if (((i + 1) % numsets) == 0) {
		k++;
		i = 0;
	      }
	      else
		i++;
	    }	  
	    niptr = iptr->next;
	    FREE(iptr);
	  }
	  if (pptr->path->numinst == 0)
	    if (inpathlist(fptr->func->entrypaths, pptr->path))
	      del_pathlist(&fptr->func->entrypaths, pptr->path);

	  if (inpathlist(fptr->func->exitpaths, pptr->path))
	    del_pathlist(&fptr->func->exitpaths, pptr->path);

	  /*
	   * skip over this path
	   */
	  if (pptr->path->numinst == 0) {
	    for (ppptr = pptr->path->preds; ppptr; ppptr = ppptr->next) {
	      del_pathlist(&ppptr->path->succs, pptr->path);
	      for (npptr = pptr->path->succs; npptr; npptr = npptr->next)
		add_pathlist(&ppptr->path->succs, npptr->path);
	    }
	    for (ppptr = pptr->path->succs; ppptr; ppptr = ppptr->next) {
	      del_pathlist(&ppptr->path->preds, pptr->path);
	      for (npptr = pptr->path->preds; npptr; npptr = npptr->next)
		add_pathlist(&ppptr->path->preds, npptr->path);
	    }
	    for (ppptr = pptr->path->preds; ppptr; ppptr = npptr) {
	      npptr = ppptr->next;
	      FREE(ppptr);
	    }
	    for (ppptr = pptr->path->succs; ppptr; ppptr = npptr) {
	      npptr = ppptr->next;
	      FREE(ppptr);
	    }
	  }
	  /*
	   * only successors, delete their backlink (to this preds)
	   */
	  else
	    for (ppptr = pptr->path->succs; ppptr; ppptr = npptr) {
	      del_pathlist(&ppptr->path->preds, pptr->path);
	      npptr = ppptr->next;
	      FREE(ppptr);
	    }

	  npptr = pptr->next;
	  del_pathlist(&fptr->func->paths, pinfo = pptr->path);
	  add_pathlist(&dead_paths, pinfo);
	}
    }
  }
}

