/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: equiv.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Instruction equivalence classes
 */

#include "scs.h"

#ifdef MERGE_STATES
#ifdef MERGE_INST_OVERLAP
/*
 * build_equivs - build equivalence lists for path with overlapping instructions
 */
void build_equivs()
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2, *pptr3;
  struct instinfo *iptr, *iptr2;
  int change, searching;
  vector v;

  void add_pathlist();

  for (fptr = funcs; fptr; fptr = fptr->next)
    do {
      change = FALSE;
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	if (!pptr->path->equivs) {
	  add_pathlist(&pptr->path->equivs, pptr->path);
	  change = TRUE;
	}
	for (pptr2 = fptr->func->paths; pptr2; pptr2 = pptr2->next)
	  if (!inpathlist(pptr->path->equivs, pptr2->path)) {
	    searching = TRUE;
	    for (iptr = pptr->path->insts; iptr && searching; iptr = iptr->next)
	      for (iptr2 = pptr2->path->insts; iptr2; iptr2 = iptr2->next) {
		if (iptr->addr + iptr->size <= iptr2->addr || /* e1 before s2 */
		    iptr2->addr + iptr2->size <= iptr->addr) /* e2 before s1 */
		  continue;
		/*
		 * merge equivalences of path2 into those of path1
		 * by adding path2 as second elements of equivs of path1
		 */
		if (pptr2->path->equivs)
		  for (pptr3 = pptr2->path->equivs; pptr3; pptr3 = pptr3->next){
		    add_pathlist(&pptr->path->equivs->next, pptr3->path);
		    pptr3->path->equivs = pptr->path->equivs;
		  }
		else {
		  add_pathlist(&pptr->path->equivs->next, pptr2->path);
		  pptr2->path->equivs = pptr->path->equivs;
		}
		searching = FALSE;
		break;
	      }
	  }
      }
    } while (change);
}
#endif /*MERGE_INST_OVERLAP*/

#ifdef DEBUG
/*
 * output_equivs - print out equivalence classes of paths
 */
void output_equivs()
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;

  for (fptr = funcs; fptr; fptr = fptr->next) {
    printf("func %s\n", fptr->func->funcname);
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      printf(" path %d equivs", pptr->path->pathnum);
      for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next)
	if (pptr2->path->curfunc != fptr->func)
	  printf(" %s.%d",
		 pptr2->path->curfunc->funcname, pptr2->path->pathnum);
	else
	  printf(" %d", pptr2->path->pathnum);
      printf("\n");
    }
  }
}
#endif /*DEBUG*/
#endif /*MERGE_STATES*/

