/************************************************************************/
/*                                                                      */
/* Copyright (C) 2000 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: flow.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Optimized abstract data-flow analysis
 * (1) reduce control flow
 * (2) perform flow analysis for a subset
 *
 * Interface:
 * (a) bool item_in_node(int subset);
 * (b) void split_op(struct pathlist *preds, struct pathlist *succs);
 * (c) void join_op(struct pathlist *preds, struct pathlist *succs);
 * (d) direction: forward/backward
 */

typedef enum direction_e {
  forward, backward
} direction_t;

struct flowinfo_s {
  int *(*item_in_node)();
  void *(*split_op)();
  void *(*join_op)();
  direction_t direction;
} flowinfo;

typedef struct nodelist_s {
  struct node_s *node;      /* this node */
  nodelist_t *next;         /* next item */
} *nodelist_t;

typedef struct node_s {
  int num;                   /* id number */
  void *equiv;               /* equivalence node */
  nodelist_t *preds, *succs; /* predecessor / successor nodes */
} *node_t;

node_t start, stop;

void init_flow(item_in_node, split_op, join_op, direction)
     int *(*item_in_node)();
     void *(*split_op)();
     void *(*join_op)();
     direction_t direction;
{
  flowinfo.item_in_node = item_in_node;
  flowinfo.split_op = split_op;
  flowinfo.join_op = jion_op;
  flowinfo.direction = direction;
  start = (node_t) calloc(sizeof(node_t*));
  stop = (node_t) calloc(sizeof(node_t*));
  start->next = stop;
}

void reduce_graph() {
  node_t n, p, s;
  nodelist_t preds, succs;

  n = start;
  
