/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: loops.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Loop identification
 */

#include "scs.h"

/*#define DEBUG*/

/*
 * find_loops_body - determine for each loop which program lines belong to the loop,
 * even transitively across calls. The algorithm works bottom-up
 * starting with leaf routines, working its way up to main.
 */
void find_loops_body(set)
     int set;
{
  struct callerlist *cptr, *cptr2;
  struct funclist *fptr;
  struct funcinfo *func;
  struct pathlist *pptr, *pptr2, *pptr3;
  struct pathinfo *pinfo;
  int change, maxloop, prevloop, i, k, l, s;
  vector v, v2, w, out, diff[MAXSETSIZE];
  vector *w_list;

  do {
    change = FALSE;
    for (fptr = funcs; fptr; fptr = fptr->next)
      if (!fptr->func->entrypaths->path->loopheader) {
	maxloop = 0;
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	  pptr->path->curloop = pptr->path->loopno;
	  if (pptr->path->loopno > maxloop)
	    maxloop = pptr->path->loopno;
	  if (pptr->path->func && !pptr->path->calltoexit &&
	      !pptr->path->func->entrypaths->path->loopheader)
	    break;
	}

	if (!pptr) {
	  change = TRUE;

	  for (i = maxloop; i >= 0; i--) {
	    /*
	     * sanity check: try to find at least one path with this loopno
	     */
	    for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	      if (pptr->path->loopno == i)
		break;
	    if (!pptr) {
	      if (warning)
		fprintf(stderr,
			"WARNING: func %s does not have a path with loopno %d\n",
			fptr->func->funcname, i);
	      continue;
	    }

	    if (i == 0) {
	      pinfo = fptr->func->entrypaths->path;
	      prevloop = 0;
	    }
	    else {
	      pinfo = (struct pathinfo *) NULL;
	      prevloop = -1;
	    }
	    v = vinit();
	    pptr3 = (struct pathlist *) NULL;

	    for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	      if (pptr->path->curloop == i) {
		cache_loop(pptr->path, &v, set);

		/*
		 * propagate vectors up from called functions
		 */
		if (pptr->path->func && !pptr->path->calltoexit)
		  vunion(&v, pptr->path->func->entrypaths->path->loopmask);

		/*
		 * find loop exits (store in pptr3)
		 */
		for (pptr2 = pptr->path->succs; pptr2; pptr2 = pptr2->next)
		  if (i > pptr2->path->curloop)
		    add_pathlist(&pptr3, pptr2->path);
		    
		/*
		 * find loop header
		 * explicitly search for closest enclosing loop, i.e.
		 * a the surrounding loop with the largest loopnum
		 */
		for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
		  if (i > pptr2->path->curloop &&
		      pptr2->path->curloop > prevloop) {
		    if (!pptr->path->loopheader)
		      pinfo = pptr->path;
		    prevloop = pptr2->path->curloop;
		    break;
		  }
		for (pptr2 = pptr->path->succs; pptr2; pptr2 = pptr2->next)
		  if (i > pptr2->path->curloop &&
		      pptr2->path->curloop > prevloop) {
		    if (!pptr->path->loopheader)
		      pinfo = pptr->path;
		    prevloop = pptr2->path->curloop;
		    break;
		  }
	      }

	    /*
	     * Still no loopheader? Just find any path only in this loop
	     * and make it the loop header
	     */
	    if (!pinfo)
	      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
		if (pptr->path->loopno == i) {
		  pinfo = pptr->path;
		  break;
		}

	    /*
	     * assign exit paths to header
	     */
	    pinfo->exits = pptr3;

	    /*
	     * Determine the union of the output state of the preheader(s)
	     */
	    for (k = 0, cptr = pinfo->curfunc->caller; cptr;
		 k++, cptr = cptr->next);
	    w_list = (vector *) MALLOC(sizeof(vector) * k);

	    for (k = 0, cptr = pinfo->curfunc->caller; cptr;
		 k++, cptr = cptr->next) {
	      w = vinit();

	      if (i == 0) /* function nesting level */
		if (fptr->func == mainfuncptr)
		  w = fptr->func->entrypaths->path->in[0][0].v;
		else
		  for (pptr = cptr->succs->path->preds; pptr;
		       pptr = pptr->next)  {
		    v2 = vinit();
		    init_set(diff);
		    for (s = 0; s < setsize; s++) {
		      out = vinit();
		      cache_path(pptr->path,
				 pptr->path->in[cptr->instance][s].v,
				 &out, diff, s, v2, set);
		      vunion(&v2, out);
		      vunion(&w, out);
		      vfree(out);
		      if (s + 1 < setsize)
			v2 = diff[s];
		    }
		    free_set(diff);
		  }
	      else { /* loop nesting level */
		for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
		  if (pptr->path->curloop >= i) /* inside loop */
		    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
		      if (pptr2->path->curloop < i) /* outside loop */
			/*
			 * union of callee's exit paths (output states)
			 */
			if (pptr2->path->func) {
			  for (l = 0, cptr2 = pptr2->path->func->caller; cptr2;
			       l++, cptr2 = cptr2->next)
			    if (cptr2->succs->path->pathlabel ==
				pptr2->path->succs->path->pathlabel &&
				cptr2->func == pptr2->path->curfunc &&
				cptr2->instance == k)
			      break;
			  
			  for (pptr3 = pptr2->path->func->exitpaths; pptr3;
			       pptr3 = pptr3->next) {
			    v2 = vinit();
			    init_set(diff);
			    for (s = 0; s < setsize; s++) {
			      out = vinit();
			      cache_path(pptr3->path, pptr3->path->in[l][s].v,
					 &out, diff, s, v2, set);
			      vunion(&v2, out);
			      vunion(&w, out);
			      vfree(out);
			      if (s + 1 < setsize)
				v2 = diff[s];
			    }
			    free_set(diff);
			  }
			}
		        /*
		        * just the output state of the predecessor
			*/ 
			else {
			  v2 = vinit();
			  init_set(diff);
			  for (s = 0; s < setsize; s++) {
			    out = vinit();
			    cache_path(pptr2->path, pptr2->path->in[k][s].v,
				     &out, diff, s, v2, set);
			    vunion(&v2, out);
			    vunion(&w, out);
			    vfree(out);
			    if (s + 1 < setsize)
			      v2 = diff[s];
			  }
			  free_set(diff);
			}
	      }
	      w_list[k] = w;
	    }

	    /*
	     * assign loop header to all blocks in current loop
	     * but skip blocks of inner loops (already assigned)
	     */
	    for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	      if (pptr->path->curloop == i) {
		if (!pptr->path->loopheader) {
		  pptr->path->loopheader = pinfo;
		  pptr->path->loopmask = v;
		  pptr->path->preheaders = w_list;
		}
		if (!pptr->path->prevheader && pptr->path->loopno > i)
		  pptr->path->prevheader = pinfo;
		pptr->path->curloop = prevloop;
	      }
	  }
	}
      }
  } while (change);

#ifdef DEBUG
  {
    char s[MAXSTRING];

    fprintf(stdout, "NESTING\n");
    for (fptr = funcs; fptr; fptr = fptr->next) {
      fprintf(stdout, "func %s\n", fptr->func->funcname);
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	if (pptr->path == pptr->path->loopheader) {
	  vget(pptr->path->loopmask, s);
	  fprintf(stdout, "block %d prev %d loop %s preheaders ",
		  pptr->path->pathnum,
		  (pptr->path->prevheader ? pptr->path->prevheader->pathnum:0),
		  s);

	  for (k = 0, cptr = pptr->path->curfunc->caller; cptr;
	       k++, cptr = cptr->next) {
	    vget(pptr->path->preheaders[k], s);
	    fprintf(stdout, "%s ", s);
	  }
	  fprintf(stdout, "\n");
	}
    }
  }
#endif
}

/*
 * find_loops - determine for each loop which program lines belong to the loop,
 * even transitively across calls. The algorithm works bottom-up
 * starting with leaf routines, working its way up to main.
 */
void find_loops()
{
  int set;

  for (set = 0; set < numsets; set++)
    find_loops_body(set);
}

/*
 * clear_loops -
 */
void clear_loops()
{
  struct callerlist *cptr;
  struct funclist *fptr;
  struct pathlist *pptr;
  int k;

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      pptr->path->prevheader = NULL;
      if (pptr->path == pptr->path->loopheader) {
	vfree(pptr->path->loopmask);
	pptr->path->loopmask = vinit();
	for (k = 0, cptr = pptr->path->curfunc->caller; cptr;
	     k++, cptr = cptr->next) {
	  /* memory leak but w's pointer may be copied instead of
	   * allocated again. Hence, commented out:
	   *   vfree(pptr->path->preheaders[k]);
	   */
	  pptr->path->preheaders[k] = vinit();
	}
	pptr->path->loopheader = NULL;
      }
    }
}

/*
 * is_backedge - return TRUE if pinfo2 is a backedge for pinfo in current loop
 */
int is_backedge(pinfo, pinfo2)
     struct pathinfo *pinfo, *pinfo2;
{
  return(pinfo == pinfo->loopheader && pinfo->loopno <= pinfo2->loopno);
}

