/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: myalloc.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Fast heap routines
 */

#include <stdio.h>

#define SBRKSIZE 524288    /* Allocation size -- larger than any request */
/*#define SBRKSIZE 1024*1024*8*/

/*#define DISPOSE*/
#define MALLOC_DEBUG

#ifdef MALLOC_DEBUG
struct malloc_check {
  void *ptr;
  struct malloc_check *next;
} *check = NULL;
#endif

char *mal_curr = (char *) NULL;
char *mal_next = (char *) NULL;
char *mal_end  = (char *) NULL;

#ifdef DISPOSE
/*
 * mymalloc - allocate memory - super-fast version - free does nothing!
 */
char *mymalloc(size)
register int size;
{
    register char *n;
    char *sbrk();
    void error();

    size = (size + 3) & (-4);
    n = mal_next + size;
    if (n <= mal_end) {
      mal_curr = mal_next;
      mal_next = n;
      }  
    else {
      if ((mal_curr = sbrk(SBRKSIZE)) == (char *) -1)
        if ((mal_curr = sbrk(size)) == (char *) -1)
          error("out of memory\n");
        else
          mal_end = mal_curr + size;
      else
        mal_end = mal_curr + SBRKSIZE;
      mal_next = mal_curr + size;
      }  
    return(mal_curr);
}

char *mycalloc(nelem, size)
register int nelem, size;
{
  register char *n, *ret;
  int i;

  ret = n = mymalloc(nelem * size);
  for (i = 0; i < nelem * size; i++)
    *(n++) = '\0';

  return(ret);
}

/*
 * myfree - can only free the last thing allocated.
 */
void myfree(ptr)
register char *ptr;
{
    if (ptr == mal_curr)
      mal_next = mal_curr;
}

/*
 * dispose - free all allocated memory (back to ptr)
 */
void dispose(ptr)
register char *ptr;
{
    mal_curr = mal_next = ptr;
}
#else /*!DISPOSE*/

char *mymalloc(size)
register int size;
{
  register char *n;
  void error();
  
  if (!(n = (char *) malloc(size)))
    error("out of memory\n");
  
#ifdef MALLOC_DEBUG
  {
    struct malloc_check *top;

    top = (struct malloc_check *) malloc(sizeof(struct malloc_check));
    top->ptr = n;
    top->next = check;
    check = top;
  }
#endif

#ifdef DEBUG
  fprintf(stderr, "malloc %x\n", n); fflush(stderr);
#endif
  return (n);
}

char *mycalloc(nelem, size)
register int nelem, size;
{
  register char *n;
  void error();
  
  if (!(n = (char *) calloc(nelem, size)))
    error("out of memory\n");
  
#ifdef MALLOC_DEBUG
  {
    struct malloc_check *top;

    top = (struct malloc_check *) malloc(sizeof(struct malloc_check));
    top->ptr = n;
    top->next = check;
    check = top;
  }
#endif

#ifdef DEBUG
  fprintf(stderr, "calloc %x\n", n); fflush(stderr);
#endif
  return (n);
}

void myfree(ptr)
register char *ptr;
{
#ifdef MALLOC_DEBUG
  if (ptr) {
    struct malloc_check *top, *prev;

    for (prev= NULL, top = check; top; prev = top, top = top->next)
      if (top->ptr == ptr) {
	if (prev)
	  prev->next = top->next;
	else
	  check = top->next;
	break;
      }
    if (top)
      free(top);
    else {
      fprintf(stderr, "ERROR: attempt to free unallocated pointer %x\n", ptr);
      fflush(stderr);
    }
  }
#endif

#ifdef DEBUG
  fprintf(stderr, "free %x\n", ptr); fflush(stderr);
#endif
  if (ptr)
    free(ptr);
}
#endif /*DISPOSE*/
