/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: pathfile.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Input of PATH file
 */

#include "scs.h"

/*
 * addpath - add a path to the list of paths if not already there
 */
struct pathlist *addpath(paths, paths_tail, pathnum, pathlabel, entrypath)
struct pathlist **paths, **paths_tail;
int pathnum, pathlabel, entrypath;
{
  register struct pathlist *pptr, *pptr2, *pptr3, *pptr4;

  /* attempt to see if path already exists */
  for (pptr = *paths; pptr && pptr->path->pathlabel != pathlabel;
       pptr = pptr->next)
    ;
  
  /* if does not exist or path label already seen */
  if (!pptr || (pathnum != -1 && pptr->path->pathnum != -1)) {
    numpaths++;
    pptr2 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
    pptr2->path = (struct pathinfo *) MALLOC(sizeof(struct pathinfo));
    pptr2->path->pathnum = pathnum;
    pptr2->path->pathlabel = pathlabel;
    pptr2->path->preds = (struct pathlist *) NULL;
    pptr2->path->succs = (struct pathlist *) NULL;
    pptr2->path->func = (struct funcinfo *) NULL;

    pptr2->path->in = (struct veclist **) NULL;
    pptr2->path->instance = 0;
    pptr2->path->conf = (struct conf_t **) NULL;
    pptr2->path->states = -1;
#ifdef MERGE_STATES
    pptr2->path->eqcfptr = (struct fmline_t *) NULL;
    pptr2->path->equivs = (struct pathlist *) NULL;
#endif /*MERGE_STATES*/

#ifdef BIT_ENCODE
    pptr2->path->mmptr = (struct fmline_t *) NULL;
    pptr2->path->missptr = (struct fmline_t *) NULL;
    pptr2->path->firstmissptr = (struct fmline_t *) NULL;
    pptr2->path->conflictptr = (struct fmline_t *) NULL;
    pptr2->path->costptr = (struct costlist_t *) NULL;
#endif /*BIT_ENCODE*/

    pptr2->path->loopheader =
      pptr2->path->prevheader = (struct pathinfo *) NULL;
    pptr2->path->exits = (struct pathlist *) NULL;

    /*
     * append to path list
     */
    pptr2->next = (struct pathlist *) NULL;
    if (*paths) {
      (*paths_tail)->next = pptr2;
      *paths_tail = pptr2;
    }
    else
      *paths = *paths_tail = pptr2;
    
    /* if a previous path with the same path label */
    if (pptr) {
      
      /* two paths with the same path label (starts at same basic block)
	 will always have the same set of predecessors and equivalences */
      pptr2->path->preds = pptr->path->preds;
#ifdef MERGE_STATES
#ifndef MERGE_INST_OVERLAP
      /*
       * merge states for any two paths with same label
       */
      pptr2->path->equivs = pptr->path->equivs;
#endif /*!MERGE_INST_OVERLAP*/
#endif /*MERGE_STATES*/
      
      /* but still need to add this path to the succ list of paths that
	 go to this path label */
      for (pptr3 = pptr->path->preds; pptr3; pptr3 = pptr3->next) {
	pptr4 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
	pptr4->path = pptr2->path;
	pptr4->next = pptr3->path->succs;
	pptr3->path->succs = pptr4;
      }

    }

#ifdef MERGE_STATES
#ifndef MERGE_INST_OVERLAP
    /*
     * add itself to the list of paths with the same label (equivalence)
     * add as second element to avoid changing the root. (If the root
     * was changed, all other equivalence paths had to get the root adjusted.)
     */
    pptr3 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
    pptr3->path = pptr2->path;
    if (pptr2->path->equivs) {
      pptr3->next = pptr2->path->equivs->next;
      pptr2->path->equivs->next = pptr3;
    }
    else {
      pptr3->next = pptr2->path->equivs;
      pptr2->path->equivs = pptr3;
    }
#endif /*!MERGE_INST_OVERLAP*/
#endif /*MERGE_STATES*/
    
    /* save the newly created path */
    pptr = pptr2;
  }
  
  /* else if exists only as a successor */
  else if (pptr->path->pathnum == -1)
    pptr->path->pathnum = pathnum;
  
  /* Place the path in the list of entry paths if the entry paths
     are still being read. */
  if (entrypath && pathnum != -1) {
    pptr3 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
    pptr3->path = pptr->path;
    pptr3->add.allhits = FALSE;
    pptr3->next = funcs_tail->func->entrypaths;
    funcs_tail->func->entrypaths = pptr3;
  }
  
  return(pptr);
}

/*
 * connectpaths - connect the current path with a successor path
 */
void connectpaths(paths, paths_tail, curpath, pathlabel)
struct pathlist **paths, **paths_tail;
struct pathinfo *curpath;
int pathlabel;
{
  register struct pathinfo *succpath;
  register struct pathlist *pptr, *pptr2;
  struct pathlist *addpath();
  
  /* Add successor path to path list if not already there */
  (void) addpath(paths, paths_tail, -1, pathlabel, FALSE);
  
  /* For each successor path with the pathlabel */
  for (pptr = *paths; pptr; pptr = pptr->next)
    if ((succpath = pptr->path)->pathlabel == pathlabel) {
      
      /* Add the successor path to the list of successors of the
	 current path. */
      pptr2 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
      pptr2->path = succpath;
      pptr2->next = curpath->succs;
      curpath->succs = pptr2;
      
      /* Add the current path to the list of predecessors of the
	 successor path. */
      pptr2 = (struct pathlist *) MALLOC(sizeof(struct pathlist));
      pptr2->path = curpath;
      pptr2->next = succpath->preds;
      succpath->preds = pptr2;
    }
}

/*
 * add_pathlist - add an element to a pathlist if not already present
 */
void add_pathlist(head, cptr)
struct pathlist **head;
struct pathinfo *cptr;
{
  struct pathlist *pptr;
  
  for (pptr = *head; pptr; pptr = pptr->next)
    if (pptr->path == cptr)
      return;

  pptr = (struct pathlist *) MALLOC(sizeof(struct pathlist));
  pptr->next = *head;
  pptr->path = cptr;
  *head = pptr;
}

/*
 * del_pathlist - delete an element in a pathlist and free heap space
 */
void del_pathlist(head, cptr)
struct pathlist **head;
struct pathinfo *cptr;
{
  struct pathlist *pptr, *nptr;

  pptr = (struct pathlist *) NULL;
  for (nptr = *head; nptr; pptr = nptr, nptr = nptr->next)
    if (nptr->path == cptr) {
      if (pptr)
	pptr->next = nptr->next;
      else
	*head = nptr->next;
      FREE(nptr);
      return;
    }
}

/*
 * free_pathlist - delete all elements in a pathlist and free heap space
 */
void free_pathlist(nptr)
struct pathlist *nptr;
{
  struct pathlist *pptr;

  if (!nptr)
    return;

  for (pptr = nptr, nptr = nptr->next; nptr; pptr = nptr, nptr = nptr->next)
    FREE(pptr);
  FREE(pptr);
}

/*
 * inpathlist - determine if some path is in a given pathlist
 */
int inpathlist(pptr, path)
struct pathlist *pptr;
struct pathinfo *path;
{
  while (pptr)
    if (pptr->path == path)
      return(TRUE);
    else
      pptr = pptr->next;

  return(FALSE);
}

/*
 * procpathfile - read in a path file and put the info into the path flow graph
 */
int procpathfile(fpath, address)
FILE *fpath;
int *address;
{
  int offset = *address;
  int is_main = FALSE;
  int i, j, k, status, pathnum, moreentrypaths, num, low, high, allhits;
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;
  struct pathlistlist *lptr;
  struct pathinfo *curpath;
  struct instinfo *iptr, *liptr;
  struct pathlist *addpath();
  struct pathlist *paths_tail = (struct pathlist *) NULL;
  void connectpaths();
  extern char *strcpy();
  
  highaddr = *address;
  
  /* A -1 should preceed a function name. */
  status = fscanf(fpath, "%d", &i);  
  if (i != -1)
    error("-1 expected at beginning of function information");
  
  /* Process information for a function. */
  while (status != EOF) {
    
    /* Allocate space for function information. */
    fptr = (struct funclist *) MALLOC(sizeof(struct funclist));
    fptr->func = (struct funcinfo *) MALLOC(sizeof(struct funcinfo));
    fptr->func->entrypaths = (struct pathlist *) NULL;
    fptr->func->exitpaths = (struct pathlist *) NULL;
    fptr->func->caller = (struct callerlist *) NULL;
    fptr->func->funcnum = ++numfuncs;
#ifdef DEBUG2
    fptr->func->numcalls = 0;
#endif
    /*
     * append to function list
     */
    fptr->next = (struct funclist *) NULL;
    if (funcs) {
      funcs_tail->next = fptr;
      funcs_tail = fptr;
    }
    else
      funcs = funcs_tail = fptr;
    
    /* Read the name of the function. */
    fscanf(fpath, "%s", funcs_tail->func->funcname);

    /* Following Line commented out by Sibin M on May 03, 2004.
     * To avoid unnecessary prints.
     */

     printf("func = %s\n",funcs_tail->func->funcname);
    is_main |= (strcmp(funcs_tail->func->funcname, "_main") == 0);
    
    /* Prepare to process path information. */
    funcs_tail->func->paths = (struct pathlist *) NULL;
    pathnum = 0;
    moreentrypaths = TRUE;
    
    /* Read the first label for the first path. */
    fscanf(fpath, "%d", &i);

    /* Following Line commented out by Sibin M on May 03, 2004.
     * To avoid unnecessary prints.
     */

    printf(" b = %d\n",i);
    do {
      
      /* Determine if path has been referenced previously as a successor.
	 If not, add it to the list of paths. */
      pathnum++;
      curpath =
	(addpath(&funcs_tail->func->paths, &paths_tail,
		 pathnum, i, moreentrypaths))->path;
      curpath->curfunc = funcs_tail->func;
      
      /*
       * Read in loop info (loop#, frequency)
       */
      fscanf(fpath, "%d %d", &i, &k);
      curpath->loopno = i;
      curpath->freq = k;

      /* Read in information about the location and number of insts
	 within the path.  Note we are assuming that each inst will
	 require a constant number of bytes. */
      curpath->insts = (struct instinfo *) NULL;
      low = 10000000;
      high = -1;
      num = 0;
      fscanf(fpath, "%d", &i);
      while (i != -1) {
	iptr = (struct instinfo *) MALLOC(sizeof(struct instinfo));
	iptr->addr = i + offset;
	iptr->firstline = iptr->addr / linesize;
	iptr->startindex = iptr->firstline % numsets;
	iptr->tag = iptr->firstline / numsets;
	if (iptr->addr < low)
	  low = iptr->addr;
	fscanf(fpath, "%d", &i);
	iptr->num = i;
	num += i;
	iptr->size = i*INSTSIZE;
	iptr->lastline = (iptr->addr + (i-1)*INSTSIZE) / linesize;
	iptr->numlines = iptr->lastline - iptr->firstline + 1;
	
	/*
	 * update cache and determine conflicts with other paths
	 */
	i = iptr->startindex;
	k = iptr->tag;
	for (j = 0; j < iptr->numlines; j++) {
	  /*
	   * new line, cache current path and add to all previous conflicts
	   */
	  if (currtag[i] != k) {
	    /*
	     * search for same tag in list of pathlists, look at first
	     * path in each pathlist only since the tags should be the
	     * same for all paths in this particular pathlist data structure
	     */
	    for (lptr = setpath[i]; lptr; lptr = lptr->next)
	      if (lptr->tag == k)
		break;

	    /*
	     * allocate new pathlistlist entry if tag not yet in list
	     */
	    if (!lptr) {
	      lptr =
		(struct pathlistlist *) MALLOC (sizeof(struct pathlistlist));
	      lptr->tag = currtag[i] = k;
	      lptr->list = (struct pathlist *) NULL;
	      lptr->next = setpath[i];
	      setpath[i] = lptr;
	    }
	  }
	  else
	    lptr = setpath[i];

	  /*
	   * add current path to conflict list
	   */
	  if (!inpathlist(lptr->list, curpath)) {
	    pptr = (struct pathlist *) MALLOC(sizeof(struct pathlist));
	    pptr->path = curpath;
	    pptr->next = lptr->list;
	    lptr->list = pptr;
	  }
	  
	  if (((i + 1) % numsets) == 0) {
	    k++;
	    i = 0;
	  }
	  else
	    i++;
	}
	
	iptr->next = (struct instinfo *) NULL;
	if (iptr->addr + iptr->size > high) {
	  high = iptr->addr + iptr->size;
	  if (high > highaddr)
	    highaddr = high;
	}
	liptr = curpath->insts;
	if (liptr) {
	  for (; liptr->next; liptr = liptr->next)
	    ;
	  liptr->next = iptr;
	}
	else
	  curpath->insts = iptr;
	fscanf(fpath, "%d", &i);
      }
      curpath->numinst = num;
      curpath->fitsincache = (high - low <= cachesize);
      curpath->traversal = (struct traversalinfo *) NULL;
      
      /* Read in information about the successors to the path. */
      fscanf(fpath, "%d", &i);

    /* Following Line commented out by Sibin M on May 03, 2004.
     * To avoid unnecessary prints.
     */

      printf(" i = %d\n",i);
      curpath->succtype = i;
      if (i == NOCALL) {
	
	/* Path does not end with a call.  Read in a list of successors
	   filling the predecessors as well.  The successor list is
	   terminated by a -1. */
	fscanf(fpath, "%d", &i);
	if (i == -1)
	  add_pathlist(&fptr->func->exitpaths, curpath);
	else
	  do {
	    connectpaths(&funcs_tail->func->paths, &paths_tail, curpath, i);
	    fscanf(fpath, "%d", &i);
	  } while (i != -1);
      }
      else if (i == CALL) {
	char s[MAXSTRING];

	fscanf(fpath, "%s", s);
	curpath->funcname = strcpy((char *) MALLOC(strlen(s) + 1), s);
	curpath->calltoexit = (strcmp(curpath->funcname, "_exit") == 0 ||
			       strcmp(curpath->funcname, "__exit") == 0);
	fscanf(fpath, "%d %d", &curpath->callnum, &i);
	if (i != -1)
	  connectpaths(&funcs_tail->func->paths, &paths_tail, curpath, i);
      }
      else {
	printf("error cause of %d\n\n",i);
	error("unexpected code for transitions");

      }      

      /* Read the label for the next path.  If it is a -1 and currently
	 processing entry paths, then indicate that there are no more
	 entry paths and continue processing the next path. */
      status = fscanf(fpath, "%d", &i);
      if (i == -1 && moreentrypaths) {
	moreentrypaths = FALSE;
	status = fscanf(fpath, "%d", &i);
      }
    }
    while (status != EOF && i != -1);

    if (status != EOF && i == -1) {
      status = fscanf(fpath, "%d", &fptr->func->base_on_stack);
      if (status != EOF)
	status = fscanf(fpath, "%d", &i);
    }
    
#ifdef DEBUG
    if (highaddr/linesize > MAXPROGLINES)
      error("too many insts");
#endif
    /* set all program lines to false */
    for (i = 0; i < highaddr/linesize; i++)
      proglines[i] = FALSE;
    
    /*
     * for each path in the function, determine if each successor to
     * the path references a subset of the lines that the path references
     * if the original path doesn't contain a call (other than exit())
     */
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
      if (pptr->path->succtype == NOCALL || pptr->path->calltoexit) {
	for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	  for (i = iptr->firstline; i <= iptr->lastline; i++)
	    proglines[i] = TRUE;
	for (pptr2 = pptr->path->succs; pptr2; pptr2 = pptr2->next)
	  if (pptr2->path->fitsincache) {
	    allhits = TRUE;
	    for (iptr = pptr2->path->insts; iptr && allhits;
		 iptr = iptr->next)
	      for (i = iptr->firstline; i <= iptr->lastline; i++)
		if (!proglines[i]) {
		  allhits = FALSE;
		  break;
		}
	    pptr2->add.allhits = allhits;
	  }
	  else
	    pptr2->add.allhits = FALSE;
	for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	  for (i = iptr->firstline; i <= iptr->lastline; i++)
	    proglines[i] = FALSE;
      }

    fptr->func->numpaths = pathnum;
  }
  *address = highaddr;
  return(TRUE);
}
