/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: scs.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * The main part of the simulator
 */

#include "scs.h"

#include <sys/time.h>

#define	MACRO_BEGIN	do {
#define	MACRO_END	} while (0)

#define MINUS_TIME(dst, src1, src2) \
  MACRO_BEGIN \
    if (src2.tv_usec > src1.tv_usec) { \
      dst.tv_sec = src1.tv_sec - src2.tv_sec - 1; \
      dst.tv_usec = (src1.tv_usec - src2.tv_usec) + 1000000; \
    } \
    else { \
      dst.tv_sec = src1.tv_sec - src2.tv_sec; \
      dst.tv_usec = src1.tv_usec - src2.tv_usec; \
    } \
  MACRO_END

int predict = FALSE;
int profile = FALSE;
int verbose = FALSE;
int warning = TRUE;
FILE *fmacros, *ftables, *fprofile;

struct funclist *funcs =         /* ptr to list of funcs in the program */
  (struct funclist *) NULL;
struct funclist *funcs_tail =    /* ptr to end of list of funcs in the pgm */
  (struct funclist *) NULL;
struct pathlist *dead_paths =    /* ptr to list of dead code paths */
  (struct pathlist *) NULL;

int proglines[MAXPROGLINES];     /* indicates if prog lines referenced */
int numpaths = 0;                /* number of unique paths in program */
int numfuncs = 0;                /* number of functions in program */
int highaddr;

struct funcinfo *mainfuncptr =   /* pointer to func info for main func */
   (struct funcinfo *) NULL;

/*
 * This array represents the maximum cache to simulate.
 */
unsigned int currtag[MAXNUMSETS];  /* contains the set of tags in cache */

struct pathlistlist *setpath[MAXNUMSETS];
                                 /* paths containing line of cached set */

FILE *output_file;               /* Cache simulation output file     */
int numsets;                     /* Number of sets in the cache      */
int setsize;                     /* Number of lines per set in cache */
int instonline;                  /* Instructions per line            */
int linesize;                    /* Size of cache line in bytes      */
int cachesize;                   /* Size of cache in bytes           */
int numtagwords;                 /* Number of words to hold all tags */
int cline_shift;                 /* Log2 of cache line size in bytes */
int vecbits;                     /* number of bits in a bit vector   */
vector vallptr;                  /* ptr to vall() vec, alloc. once!  */

/*
 * simple_conflicts_body - 
 */
#ifndef TRASH
void simple_conflicts_body(func, set)
     struct funcinfo *func;
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr;
  struct callerlist *cptr;
  register int i, s;

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	pptr->path->in[i][0].changed = TRUE;
	pptr->path->rout[i].changed = TRUE;
	pptr->path->lin[i][0].changed = TRUE;
	for (s = 0; s < setsize; s++) {
	  if (pptr->path->in[i][s].v != vallptr) {
	    vfree(pptr->path->in[i][s].v);
	    pptr->path->in[i][s].v = vinit();
	  }
	}
	if (pptr->path->rout[i].v != vallptr) {
	  vfree(pptr->path->rout[i].v);
	  pptr->path->rout[i].v = vinit();
	}
	for (s = 0; s < setsize; s++) {
	  if (pptr->path->lin[i][s].v != vallptr) {
	    vfree(pptr->path->lin[i][s].v);
	    pptr->path->lin[i][s].v = vinit();
	  }
	}
	if (pptr->path->pout[i].v != vallptr) {
	  vfree(pptr->path->pout[i].v);
	  pptr->path->pout[i].v = vallptr;
	}
	if (pptr->path->din[i].v != vallptr) {
	  vfree(pptr->path->din[i].v);
	  pptr->path->din[i].v = vallptr;
	}
      }

  invalidate_union_cache(func->entrypaths, numsets, set);
  simple_union_body(set);
/*
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_vec(fptr->func, i);
  }
*/

  clear_loops();
  find_loops_body(set);

  invalidate_reaching_cache(func->exitpaths, numsets, set);
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	if (pptr->path->calltoexit && set == 0)
	  pptr->path->rout[i].changed = TRUE;
  simple_reaching_body(set);
/*
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_rvec(fptr->func, i);
  }
*/

  invalidate_post_dominator_cache(func->exitpaths, numsets, set);
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	if (pptr->path->calltoexit) {
	  pptr->path->pout[i].v = vinit();
	  pptr->path->pout[i].changed = TRUE;
	}
  simple_post_dominator_body(set);
/*
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_pvec(fptr->func, i);
  }
*/

  invalidate_linear_cache(func->entrypaths, numsets, set);
  simple_linear_body(set);
/*
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_lvec(fptr->func, i);
  }
*/

  invalidate_dominator_cache(func->entrypaths, numsets, set);
  simple_dominator_body(set);
/*
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_dvec(fptr->func, i);
  }
*/

  output_simple_conflicts_body(set);
#ifdef BIT_ENCODE
  output_dag_prediction();
  merge_equiv_prediction();
  check_savings();
  output_static_savings();
#endif
}
#else
void simple_conflicts_body(func)
     struct funcinfo *func;
{
  struct funclist *fptr;
  struct pathlist *pptr;
  struct callerlist *cptr;
  register int i;

  invalidate_union_cache(func->entrypaths, numsets, 0);
  simple_union();
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_vec(fptr->func, i);
  }

  find_loops();

  invalidate_reaching_cache(func->exitpaths, numsets, 0);
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	if (pptr->path->calltoexit)
	  pptr->path->rout[i].changed = TRUE;
  simple_reaching();
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_rvec(fptr->func, i);
  }

  invalidate_post_dominator_cache(func->exitpaths, numsets, 0);
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	if (pptr->path->calltoexit) {
	  pptr->path->pout[i].v = vinit();
	  pptr->path->pout[i].changed = TRUE;
	}
  simple_post_dominator();
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_pvec(fptr->func, i);
  }

  invalidate_linear_cache(func->entrypaths, numsets, 0);
  simple_linear();
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_lvec(fptr->func, i);
  }

  invalidate_dominator_cache(func->entrypaths, numsets, 0);
  simple_dominator();
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    optimize_dvec(fptr->func, i);
  }

  output_simple_conflicts();
#ifdef BIT_ENCODE
  output_dag_prediction();
  merge_equiv_prediction();
  check_savings();
  output_static_savings();
#endif
}
#endif

/*
 * simple_conflicts - long vector-based simple method to determine conflicts
 * Uses a bit vector with size equal to (# lines in program + # lines in cache).
 * Two bit vectors associated with each unique path.
 * Incoming vector of start paths in main function initialized to uncached.
 * Incoming vector = U Outgoing vectors of predecessors
 * Outgoing vector = Incoming vector - conflicting lines of path + lines in path
 * Determine cache behavior by examining incoming vector:
 * (1) if line in path set and no other conflicting bit set, then hit.
 * (2) if line in path not set, then miss.
 * (3) if line in path set and some other conflicting bit set, then state-dep.
 */
void simple_conflicts()
{
  struct funclist *fptr;
  struct funcinfo *func;
  struct pathlist *pptr;
  struct callerlist *cptr;
  struct instinfo *iptr;
  register int i, k, l, m;
  vector vall();

  /*
   * find last program line, assume last function to be last in linked list
   * bit vector length = value of the last program line + 1 + # cache lines
   */
  for (fptr = funcs; fptr && fptr->next; fptr = fptr->next);
  if (!fptr)
    fptr = funcs;
  for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
    for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
      if (iptr->lastline > vecbits)
	vecbits = iptr->lastline;
  /*
   * We need 0..numsets-1,numsets..lastline+numsets bits in each vector.
   * Hence, the number of bits in the vector is lastline+numsets+1.
   */
  vecbits += numsets + 1;
  veclen = (vecbits >> intscale) + 1;  
  veclen = (veclen / numsets) + 1;
  vallptr = vall();

  /*
   * initialize vectors in main
   */
  func = mainfuncptr;
  allocate_caller(func, (struct pathinfo *) NULL, 0); /* main called by UNIX */

  /*
   * traverse call graph to initialize calls to functions other than main
   */
  allocate_caller_dfs(func->paths, 0);

  /*
   * traverse all functions instances to determine max. vector size
   * needed to represent function instances.
   */

#ifdef TRASH
  /*
   * This offset[] vector doesn't seem to be used
   */

  /* determine max. # func insts */
  k = 0;
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    fptr->func->offset = (int *) MALLOC(sizeof(int) * i);
    k = MAX(k, i - 1);
  }

  /* determine offset for each func inst */
  m = 0;
  for (l = 0; l <= k; l++)
    for (fptr = funcs; fptr; fptr = fptr->next)
      for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next)
	if (i == l)
	  fptr->func->offset[i] = m++;

  if (m >= vecbits) {
    fprintf(stderr, "ERROR: bit vectors too small for func insts\n");
    exit(-1);
  }
#endif /*TRASH*/

  /*
   * traverse all functions to initialize vectors
   */
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (i = 0, cptr = fptr->func->caller; cptr; i++, cptr = cptr->next);
    allocate_vectors(fptr->func, i);
  }

  /*
   * initialize counters for dfa
   */
  init_counters();

#ifndef TRASH
  {
    int set;

    for (set = 0; set < numsets; set++)
      simple_conflicts_body(func, set);
  }
#else
  simple_conflicts_body(func);
#endif
}

/*
 * main - controls command line parameters to static simulator and invokes
 * all other functions
 */
int main(argc, argv)
     int argc;
     char *argv[];
{
  register int i;
  int address = 0;
  long StartTime, StopTime;
  float TotalTime;
  
  struct timeval start, stop, total;
  char filename[MAXNAME];
  char *basename;
  FILE *fpath;
  FILE *fopen();
  int procpathfile();
  void dead_code_elim();
  void buildcallgraph(), staticsim(),
  optstates(), dumpstates(), init_cacher();
#ifdef DEBUG
  void dumpproginfo();
#endif
  int fclose();
  
  gettimeofday(&start, (struct timezone *) NULL);
  init();
  
  /* initialize the cache */
  init_cacher();
  c_init();
  
  basename = (char *) NULL;

  /* Process each path file to build the path flow graph. */
  for (i = 1; i < argc; i++) {
    switch (argv[i][0]) {
    case '-':
      switch(argv[i][1]) {
      case 't':
	predict = TRUE;
	break;
#ifdef PROFILING
      case 'p':
        profile = TRUE;
        break;
#endif
      case 'v':
        verbose = TRUE;
        break;
      case 'w':
        warning = FALSE;
        break;
      default:
	fprintf(stderr, "use: staticsim [-p] [-t] [-v] [-w] {basename}*\n");
#ifdef PROFILING
	fprintf(stderr, "  -p for profiling\n");
#endif
	fprintf(stderr, "  -t for timing predictions\n");
	fprintf(stderr, "  -v for verbose output\n");
	fprintf(stderr, "  -w suppress warnings\n");
	exit(-1);
      }
      break;
    default:
      sprintf(filename, "%s.path", argv[i]);
      if (!(fpath = fopen(filename, "r")))
	error("cannot open path input file");
      if (procpathfile(fpath, &address) && !basename)
	basename = argv[i];
      (void) fclose(fpath);
    }
  }

  dead_code_elim();

#ifdef DEBUG
  output_conflicts();
#endif /*DEBUG*/
  
  /*
   * Build call graph (connect functions together according to calls).
   * set veclen (i.e. only call vector functions after this point!!!)
   */
  buildcallgraph();
  
#ifdef DEBUG2
  determine_numcalls();
  determine_relevance();
#endif
  
#ifdef MERGE_STATES
#ifdef MERGE_INST_OVERLAP
  build_equivs();
#endif /*MERGE_INST_OVERLAP*/
#ifdef DEBUG
  output_equivs();
#endif /*DEBUG*/
#endif /*MERGE_STATES*/
  if (predict) {
    sprintf(filename, "%s.ist", basename);
    if (!(output_file = fopen(filename, "w")))
      error("cannot open cache simulation output file");
  }
  simple_conflicts();
  if (predict)
    fclose(output_file);
#ifdef PROFILING
  else {
    sprintf(filename, "%s.macros.h", basename);
    if (!(fmacros = fopen(filename, "w")))
      error("cannot open macro output file");

    sprintf(filename, "%s.tables.h", basename);
    if (!(ftables = fopen(filename, "w")))
      error("cannot open table output file");

    if (profile) {
      sprintf(filename, "%s.icf", basename);
      if (!(fprofile = fopen(filename, "w")))
	error("cannot open table output file");
    }

    simple_states();
    fclose(fmacros);
    fclose(ftables);
    if (profile)
      fclose(fprofile);
  }
#endif

  gettimeofday(&stop, (struct timezone *) NULL);
  MINUS_TIME(total, stop, start);
  printf("simulation time was %3d.%06d seconds\n", total.tv_sec, total.tv_usec);
  exit(0);
}

